module ctaper_mod
use global_params_mod
use type_mod
implicit none 
contains
!===========================================
subroutine ctaper2d(tape,vgrid,nxpad,nypad,opt) 
        integer                                 ::nx,ny,ntap,nxpad,nypad,ix,iy,opt
        real                                    ::mx,my,dx,dy
        real, dimension(:,:)                    ::tape
        real, dimension(:),allocatable          ::tmp,coeff,cc
        real                                    ::pi
        real                                    ::dc,oc
        integer                                 ::ic,nc
        type(vel_grid)                          ::vgrid    
        
        pi= 3.14159265359
        
        nx=vgrid%ax%n; ny=vgrid%ay%n
        dx=vgrid%ax%d; dy=vgrid%ay%d
        tape=1.

        if (opt==1) then
        mx=1/(dx*ceiling(nx*taper_pct));
        my=1/(dy*ceiling(ny*taper_pct));
            do ix=1,ceiling(nx*taper_pct)
                tape(nxpad+ix+1,:)=mx*(ix*dx);
                tape(nxpad+nx-ix,:)=mx*(ix*dx);
            enddo
            tape(1,:)=0.001*tape(2,:); tape(nx,:)=0.001*tape(nx-1,:);
            if (ny>1)then
                    do iy=1,ceiling(ny*taper_pct)
                        tape(nypad+iy+1,:)=my*(iy*dy);
                        tape(nypad+ny-iy,:)=my*(iy*dy);
                    enddo
            endif
        elseif (opt==2)then
            ntap=ceiling(nx*taper_pct);       
            allocate(coeff(2*ntap)); allocate(cc(ntap))
            dc=pi/(2*ntap);oc=0.;nc=2*ntap
            do ic=1,nc
                coeff(ic)=abs(cos(oc+(ic-1)*dc));
            enddo
            cc=coeff(ntap+1:2*ntap);
            do ix=1,ceiling(nx*taper_pct)
                tape(   ix  ,:)=cc(ix);
                tape(nx-ix+1,:)=cc(ix);
            enddo
            if (ny>1)then
                    ntap=ceiling(ny*taper_pct);      
                    deallocate(coeff,cc)
                    allocate(coeff(2*ntap)); allocate(cc(ntap))
                    dc=pi/(2*ntap);oc=0.;nc=2*ntap
                    do ic=1,nc
                        coeff(ic)=abs(cos(oc+(ic-1)*dc));
                    enddo
                    cc=coeff(ntap+1:2*ntap);
                    do iy=1,ceiling(ny*taper_pct)
                        tape(:,   iy  )=cc(iy)*tape(:,   iy  );
                        tape(:,ny-iy+1)=cc(iy)*tape(:,ny-iy+1);
                    enddo
            endif
        else 
            call seperr('wrong taper param. triang=1 hanning=2.')
        endif
        if(allocated(coeff))deallocate(coeff)
        if(allocated(cc   ))deallocate(cc   )
end subroutine
!#######################
subroutine ctaper2d_marine(tape,pgrid,opt)
        integer                                 ::nx,ny,ntap,nxpad,nypad,ix,iy,opt
        real                                    ::mx,my,dx,dy
        real, dimension(:,:)                    ::tape
        real, dimension(:),allocatable          ::tmp,coeff,cc
        real                                    ::pi
        real                                    ::dc,oc
        integer                                 ::ic,nc
        type(pro_grid)                          ::pgrid    
        
        pi= 3.14159265359
        nx=pgrid%ax%n; ny=pgrid%ay%n
        dx=pgrid%ax%d; dy=pgrid%ay%d
        tape=1.
        if (opt==1) then
        mx=1/(dx*ceiling(nx*taper_pct));
        my=1/(dy*ceiling(ny*taper_pct));
            do ix=1,ceiling(nx*taper_pct)
                tape(pgrid%nxpad_beg+ix+1,:)=mx*(ix*dx);
                tape(pgrid%nxpad_end+nx-ix,:)=mx*(ix*dx);
            enddo
            tape(1,:)=0.001*tape(2,:); tape(nx,:)=0.001*tape(nx-1,:);
            if (ny>1)then
                    do iy=1,ceiling(ny*taper_pct)
                        tape(pgrid%nypad_beg+iy+1,:)=my*(iy*dy);
                        tape(pgrid%nypad_end+ny-iy,:)=my*(iy*dy);
                    enddo
            endif
        elseif (opt==2)then
            ntap=ceiling(nx*taper_pct);       
            if(allocated(coeff))deallocate(coeff);allocate(coeff(2*ntap)); 
            if(allocated(cc))   deallocate(cc);   allocate(cc(ntap));
            dc=pi/(2*ntap);oc=0.;nc=2*ntap
            do ic=1,nc
                coeff(ic)=abs(cos(oc+(ic-1)*dc));
            enddo
            cc=coeff(ntap+1:2*ntap);
            do ix=1,ceiling(nx*taper_pct)
                tape(ix     ,:)=cc(ix);
                tape(nx-ix+1,:)=cc(ix);
            enddo
            if (ny>1)then
                    ntap=ceiling(ny*taper_pct);       
                    if(allocated(coeff))deallocate(coeff);allocate(coeff(2*ntap)); 
                    if(allocated(cc))   deallocate(cc);   allocate(cc(ntap));
                    dc=pi/(2*ntap);oc=0.;nc=2*ntap
                    do ic=1,nc
                        coeff(ic)=abs(cos(oc+(ic-1)*dc));
                    enddo
                    cc=coeff(ntap+1:2*ntap);
                    do iy=1,ceiling(ny*taper_pct)
                        tape(:,iy     )=cc(iy);
                        tape(:,ny-iy+1)=cc(iy);
                    enddo
            endif
            deallocate(coeff)
            deallocate(cc)
        else 
            call seperr('wrong taper param. triang=1 hanning=2.')
        endif
end subroutine


!######
subroutine ctaper1d(taped,modl,dx,nx,nxpad,opt)
        integer                                 ::nx, ntap, nxpad, ix, opt
        real                                    ::m, dx
        complex, dimension(:),pointer           ::taped, modl
        real   , dimension(:),pointer           ::tmp,coeff,cc
        real                                    :: pi
        real                                    :: dc,oc
        integer                                 :: ic,nc
        pi= 3.14159265359
        
        
        m=1/(dx*ceiling(nx*taper_pct));
        taped=modl;
        
        if (opt==1) then
            do ix=1,ceiling(nx*taper_pct)
                taped(nxpad+ix+1)=modl(nxpad+ix+1)*m*(ix*dx);
                taped(nxpad+nx-ix)=modl(nx-ix-nxpad)*m*(ix*dx);
            enddo
            taped(1)=0.001*taped(2); taped(nx)=0.001*taped(nx-1);
        elseif (opt==2)then
            ntap=ceiling(nx*taper_pct);       
            allocate(coeff(2*ntap)); allocate(cc(ntap))
            dc=pi/(2*ntap);oc=0.;nc=2*ntap
            do ic=1,nc
                coeff(ic)=abs(cos(oc+(ic-1)*dc));
            enddo
            cc=coeff(ntap+1:2*ntap)   
            do ix=1,ceiling(nx*taper_pct)
                taped(nxpad+ix)=modl(nxpad+ix)*cc(ix);
                taped(nxpad+nx-ix+1)=modl(nxpad+nx-ix+1)*cc(ix);
            enddo
            taped(1:max(nxpad,1))=0; 
            taped(nxpad+nx:2*nxpad+nx)=0;
        else 
            call seperr('wrong taper param. triang=1 hanning=2.')
        endif
        
        if (associated(cc)) deallocate(cc);
        if (associated(coeff)) deallocate(coeff);
end subroutine

!================================================modified from yaxuns
  !----------------------------------------------
  ! 2D Hamming window taper ....one freq.
  !----------------------------------------------
  subroutine circular_taper(wfld,tape)
    complex, dimension(:,:)                      :: wfld
    real   , dimension(:,:)                      :: tape
    integer                            :: taplen
    integer                            :: nx, ny, nw, ix, iy, iw, minxy
    real                               :: x_orig, y_orig
    real                               :: taper_radius_min
    real                               :: taper_radius_max
    real                               :: taper, dist

    nx = size(wfld, 1) - pgrid%nxpad_beg - pgrid%nxpad_end
    ny = size(wfld, 2) - pgrid%nypad_beg - pgrid%nypad_end
    if(ny>1)then
        if(nx>1)then
             minxy=min(nx,ny)
        else
             minxy=ny
        endif        
    else
        minxy=nx
    endif
    taplen = taper_pct * minxy

    x_orig = max(1.,nx * 0.5)
    y_orig = max(1.,ny * 0.5)

    taper_radius_min = minxy * 0.5 - taplen
    taper_radius_max = minxy

    do iy=1, ny
       do ix=1, nx
          dist = sqrt((ix-x_orig)**2 + (iy-y_orig)**2)
          if (dist <= taper_radius_min ) then
             tape(ix,iy) = 1.
          else if (dist > taper_radius_min .and. dist <= taper_radius_max) then
             tape(ix,iy) = 0.54 + 0.46 * cos(PI*(taper_radius_min-dist)/(taper_radius_max-taper_radius_min))
          else
             tape(ix,iy) = 0.
          end if
       end do
    end do

  end subroutine



end module
