module encode_data_compute_mod
  use global_params_mod
  use encode_data_io
  use myfft_mod
  use adj_mod

  implicit none
        complex, allocatable, private   :: spk(:,:,:)
        complex, allocatable, private   :: green_ws (:,:,:,:,:)
        complex, allocatable, private   :: green_w  (:,:,:,:,:,:)
        complex, allocatable, private   :: green_w_s(:,:,:,:,:,:)
  contains
!===========================================================================================================================================================================
!#######
  integer function encode_compute_init1(nodes) result(stat)
    integer :: nodes
        stat = 1
            if(allocated(green_ws     )) deallocate(green_ws ); allocate(green_ws (ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n))
            if(allocated(green_w      )) deallocate(green_w  ); allocate(green_w  (ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, nodes))
            if(encode)then
                if(allocated(green_w_s)) deallocate(green_w_s); allocate(green_w_s(dgrid%asx%n    , dgrid%asy%n    , ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, nodes))
            else
                if(allocated(green_w_s)) deallocate(green_w_s); allocate(green_w_s(wgrid%asx%n    , wgrid%asy%n    , ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, nodes))
            endif
            if(allocated(spk    )) deallocate(spk    ); allocate(spk(size(csou_full,1),size(csou_full,2), nodes))
        stat = 0
  end function
!#######
  integer function encode_compute_init(iwd_in,inode) result(stat)
    integer :: iwd_in
    integer :: inode
        stat = 1
            !call data_io_greens(green_ws, iwd_in); 
            green_w(:,:,:,:,:,inode)=green_ws 
            spk(:,:,inode)   = csou_full(:,:,iwd_in); 
        stat = 0
  end function
!#######
  integer function encode_compute_init_blk(iiw,max_nw) result(stat)
    integer :: iiw,max_nw
        stat = 1
            call data_io_greens_blk(green_w,iiw,max_nw);
            spk=0.;
            spk(:,:,1:max_nw-iiw+1)=csou_full(:,:,iiw:max_nw);
       stat = 0
  end function
!#######
subroutine encode_compute_data(adj, modl, data, inode, iwd)
    logical               :: adj,  add
    real, dimension(size(ref,1),size(ref,2),size(ref,3)):: modl
    complex, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n):: data
    integer :: inode, iwd
        if (encode)then
            call encode_compute_data_lop2(adj, modl, data, inode, iwd)
        else
            call encode_shifts(adj,iwd)  
            call encode_compute_data_lop(adj, modl, data, inode, iwd)
        endif
end subroutine
!#######
! compute data from all model points for all shots and receivers encoded
  subroutine encode_compute_data_lop2(adj, modl, data, inode,iwd)
    logical               :: adj,  add
    integer               :: inode, iwd
    real, dimension(size(ref,1),size(ref,2),size(ref,3)):: modl
    complex, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n):: data
        integer               ::  iszd,  isxd, isyd
        integer               ::  igzd,  igxd, igyd
        integer               :: iiszd, iisxd, iisyd
        integer               :: iigzd, iigxd, iigyd
        integer               ::  ijzd,  ijxd, ijyd
        integer               ::   jzd,   jxd,  jyd
        real                  ::    zd,    xd, yd  
        real                  :: sxd , syd, gxd, gyd 

        green_w_s(1,1,:,:,:,iwd)=0.


        if(verb>5)write(0,'(A10,2I10,8E20.5)')'b4',inode,iwd,&
                      & sum(modl), sum(data), sum(green_w_s(:,:,:,:,:,iwd)), sum(green_w(:,:,:,:,:,iwd)) 
        if(adj)then 
            do jzd=data_imin_z, data_imax_z;     zd = vgrid%az%o  + (jzd-1)*vgrid%az%d;     ijzd = 1.5 + ( zd-ggrid%az_modl%o)/ggrid%az_modl%d 
            do jyd=data_imin_y, data_imax_y;     yd = vgrid%ay%o  + (jyd-1)*vgrid%ay%d;     ijyd = 1.5 + ( yd-ggrid%ay_modl%o)/ggrid%ay_modl%d
            do jxd=data_imin_x, data_imax_x;     xd = vgrid%ax%o  + (jxd-1)*vgrid%ax%d;     ijxd = 1.5 + ( xd-ggrid%ax_modl%o)/ggrid%ax_modl%d 
                do isyd=1, wgrid%asy%n;         syd = wgrid%asy%o + (isyd-1)*wgrid%asy%d;  iisyd = 1.5 + (syd-ggrid%ay_surf%o)/ggrid%ay_surf%d
                do isxd=1, wgrid%asx%n;         sxd = wgrid%asx%o + (isxd-1)*wgrid%asx%d;  iisxd = 1.5 + (sxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                    green_w_s(1,1,jxd,jyd,jzd,iwd)=green_w_s(1,1,jxd,jyd,jzd,iwd)+(green_w(iisxd,iisyd,jxd,jyd,jzd,iwd))*(spk(isxd,isyd,iwd))
                end do; 
                end do; 
                    do igyd=1, dgrid%agy%n;     gyd = dgrid%agy%o + (igyd-1)*dgrid%agy%d;  iigyd = 1.5 + (gyd-ggrid%ay_surf%o)/ggrid%ay_surf%d
                    do igxd=1, dgrid%agx%n;     gxd = dgrid%agx%o + (igxd-1)*dgrid%agx%d;  iigxd = 1.5 + (gxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                        modl( jxd, jyd, jzd     )=modl( jxd, jyd, jzd     )+conjg(green_w(iigxd,iigyd,ijxd,ijyd,ijzd,iwd)*green_w_s(1,1,ijxd,ijyd,ijzd,iwd))*data(igxd,igyd,1,1); 
                    end do;
                    end do;
            end do;  
            end do;    
            end do;   
        else
            do jzd=data_imin_z, data_imax_z;     zd = vgrid%az%o  + (jzd-1)*vgrid%az%d;     ijzd = 1.5 + ( zd-ggrid%az_modl%o)/ggrid%az_modl%d
            do jyd=data_imin_y, data_imax_y;     yd = vgrid%ay%o  + (jyd-1)*vgrid%ay%d;     ijyd = 1.5 + ( yd-ggrid%ay_modl%o)/ggrid%ay_modl%d
            do jxd=data_imin_x, data_imax_x;     xd = vgrid%ax%o  + (jxd-1)*vgrid%ax%d;     ijxd = 1.5 + ( xd-ggrid%ax_modl%o)/ggrid%ax_modl%d
                do isyd=1, wgrid%asy%n;         syd = wgrid%asy%o + (isyd-1)*wgrid%asy%d;  iisyd = 1.5 + (syd-ggrid%ay_surf%o)/ggrid%ay_surf%d
                do isxd=1, wgrid%asx%n;         sxd = wgrid%asx%o + (isxd-1)*wgrid%asx%d;  iisxd = 1.5 + (sxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                    green_w_s(1,1,jxd,jyd,jzd,iwd)=green_w_s(1,1,jxd,jyd,jzd,iwd)+(green_w(iisxd,iisyd,jxd,jyd,jzd,iwd))*(spk(isxd,isyd,iwd))
                end do; 
                end do; 
                    do igyd=1, dgrid%agy%n;     gyd = dgrid%agy%o + (igyd-1)*dgrid%agy%d;  iigyd = 1.5 + (gyd-ggrid%ay_surf%o)/ggrid%ay_surf%d
                    do igxd=1, dgrid%agx%n;     gxd = dgrid%agx%o + (igxd-1)*dgrid%agx%d;  iigxd = 1.5 + (gxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                        data(igxd,igyd,1,1)=data(igxd,igyd,1,1)+     (green_w(iigxd,iigyd,ijxd,ijyd,ijzd,iwd)*green_w_s(1,1,ijxd,ijyd,ijzd,iwd))*modl( jxd, jyd, jzd     ); 
                    end do;   
                    end do; 
            end do;
            end do;  
            end do;   
        end if;
        if(verb>5)write(0,'(A10,2I10,8E20.5)')'af',inode,iwd,&
                      & sum(modl), sum(data), sum(green_w_s(:,:,:,:,:,iwd)), sum(green_w(:,:,:,:,:,iwd))

  end subroutine
!####### 
!#######
! compute data from all model points for all shots and receivers encoded
  subroutine encode_compute_data_lop(adj, modl, data, inode,iwd)
    logical               :: adj,  add
    integer               :: inode, iwd
    real, dimension(size(ref,1),size(ref,2),size(ref,3)):: modl
    complex, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n):: data
        integer               ::  iszd,  isxd, isyd
        integer               ::  igzd,  igxd, igyd
        integer               :: iiszd, iisxd, iisyd
        integer               :: iigzd, iigxd, iigyd
        integer               ::  ijzd,  ijxd, ijyd
        integer               ::   jzd,   jxd,  jyd
        real                  ::    zd,    xd, yd  
        real                  :: sxd , syd, gxd, gyd 

        if(verb>5)write(0,'(A10,2I10,8E20.5)')'b4',inode,iwd,&
                      & sum(modl), sum(data), sum(green_w_s(:,:,:,:,:,iwd)), sum(green_w(:,:,:,:,:,iwd)) 

        do jzd=data_imin_z, data_imax_z; zd = vgrid%az%o + (jzd-1)*vgrid%az%d;  ijzd = 1.5 + (zd-ggrid%az_modl%o)/ggrid%az_modl%d
        if(adj)then 
            do jyd=data_imin_y, data_imax_y;    yd = vgrid%ay%o + (jyd-1)*vgrid%ay%d;  ijyd = 1.5 + (yd-ggrid%ay_modl%o)/ggrid%ay_modl%d
            do jxd=data_imin_x, data_imax_x;    xd = vgrid%ax%o + (jxd-1)*vgrid%ax%d;  ijxd = 1.5 + (xd-ggrid%ax_modl%o)/ggrid%ax_modl%d
                do isyd=1, dgrid%asy%n;         
                do isxd=1, dgrid%asx%n;         
                        do igyd=1, dgrid%agy%n; gyd = dgrid%agy%o + (igyd-1)*dgrid%agy%d;  iigyd = 1.5 + (gyd-ggrid%ay_surf%o)/ggrid%ay_surf%d
                        do igxd=1, dgrid%agx%n; gxd = dgrid%agx%o + (igxd-1)*dgrid%agx%d;  iigxd = 1.5 + (gxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                                modl( jxd, jyd, jzd     )=modl( jxd, jyd, jzd     )+conjg(green_w(iigxd,iigyd,ijxd,ijyd,ijzd,iwd)*green_w_s(isxd,isyd,ijxd,ijyd,ijzd,iwd))*data(igxd,igyd,isxd,isyd); 
                        end do;
                        end do;
                end do; 
                end do; 
            end do;  
            end do;   
        else
            do jyd=data_imin_y, data_imax_y;    yd = vgrid%ay%o + (jyd-1)*vgrid%ay%d;  ijyd = 1.5 + (yd-ggrid%ay_modl%o)/ggrid%ay_modl%d
            do jxd=data_imin_x, data_imax_x;    xd = vgrid%ax%o + (jxd-1)*vgrid%ax%d;  ijxd = 1.5 + (xd-ggrid%ax_modl%o)/ggrid%ax_modl%d
                do isyd=1, dgrid%asy%n;         
                do isxd=1, dgrid%asx%n;         
                        do igyd=1, dgrid%agy%n; gyd = dgrid%agy%o + (igyd-1)*dgrid%agy%d;  iigyd = 1.5 + (gyd-ggrid%ay_surf%o)/ggrid%ay_surf%d
                        do igxd=1, dgrid%agx%n; gxd = dgrid%agx%o + (igxd-1)*dgrid%agx%d;  iigxd = 1.5 + (gxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                                data(igxd,igyd,isxd,isyd)=data(igxd,igyd,isxd,isyd)+     (green_w(iigxd,iigyd,ijxd,ijyd,ijzd,iwd)*green_w_s(isxd,isyd,ijxd,ijyd,ijzd,iwd))*modl( jxd, jyd, jzd     ); 
                        end do;   
                        end do; 
                end do;
                end do;
            end do;
            end do;  
        end if;
        end do;
        if(verb>5)write(0,'(A10,2I10,8E20.5)')'af',inode,iwd,&
                      & sum(modl), sum(data), sum(green_w_s(:,:,:,:,:,iwd)), sum(green_w(:,:,:,:,:,iwd))

  end subroutine
!####### 
  subroutine encode_shifts(adj,iwd)   
    logical                   :: adj,  add
        integer               ::   inode
        integer               ::   jzd,   jxd,  jyd
        integer               ::  iszd,  isxd, isyd, iwd 
        integer               ::  igzd,  igxd, igyd
        integer               ::  ijzd,  ijxd, ijyd
        integer               :: iiszd, iisxd, iisyd
        integer               :: iigzd, iigxd, iigyd
        real                  :: zd, xd, yd   
        real                  :: sxd , syd, gxd, gyd 

        green_w_s(:,:,:,:,:,iwd)=0.

        if(encode)then
            do jzd=data_imin_z, data_imax_z;    
            do jyd=data_imin_y, data_imax_y;    
            do jxd=data_imin_x, data_imax_x;    
                do isyd=1, wgrid%asy%n;        syd = wgrid%asy%o + (isyd-1)*wgrid%asy%d;  iisyd = 1.5 + (syd-ggrid%ay_surf%o)/ggrid%ay_surf%d 
                do isxd=1, wgrid%asx%n;        sxd = wgrid%asx%o + (isxd-1)*wgrid%asx%d;  iisxd = 1.5 + (sxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                    green_w_s(   1,   1,jxd,jyd,jzd,iwd)=green_w_s(   1,   1,jxd,jyd,jzd,iwd)+(green_w(iisxd,iisyd,jxd,jyd,jzd,iwd))*(spk(isxd,isyd,iwd))
                end do;
                end do;
            end do;
            end do;  
            end do;
                    else
            do jzd=data_imin_z, data_imax_z;    
            do jyd=data_imin_y, data_imax_y;    
            do jxd=data_imin_x, data_imax_x;    
                do isyd=1, wgrid%asy%n;        syd = wgrid%asy%o + (isyd-1)*wgrid%asy%d;  iisyd = 1.5 + (syd-ggrid%ay_surf%o)/ggrid%ay_surf%d 
                do isxd=1, wgrid%asx%n;        sxd = wgrid%asx%o + (isxd-1)*wgrid%asx%d;  iisxd = 1.5 + (sxd-ggrid%ax_surf%o)/ggrid%ax_surf%d
                    green_w_s(isxd,isyd,jxd,jyd,jzd,iwd)=green_w_s(isxd,isyd,jxd,jyd,jzd,iwd)+(green_w(iisxd,iisyd,jxd,jyd,jzd,iwd))*(spk(isxd,isyd,iwd))
                end do;
                end do;
            end do;
            end do;  
            end do;  
        endif
  end subroutine


end module


