module encode_data_mod
  use encode_data_history
  use encode_data_io
  use encode_data_array
  use encode_data_param 
  use encode_data_wavelet
  use encode_data_grid
  use encode_data_compute_mod
  use cgstep_module
  use solver_sep_mod

 
  implicit none

contains
!=======================================================================================
  integer function data_init() result(stat)
    call green_input_history()
    call green_params()                                     !initialize data modeling parameters 
    !call data_params()                                      !initialize data modeling parameters 
    call green_grid()                                       !get greens grid 
    call velocity_grid()                                    !get velocity/refl grid      
    call data_wavelet_s()                                    !get wavelet
    call data_params()                                      !initialize data modeling parameters 
    call data_array_alloc()                                 !allocate memory 
    call data_velocity()                                    !get velocity/refl models
    if (wfld_d_out) then;                                   !save down wavefield in time
        call wfld_output_history(wfld_d_); 
        call wfld_io_init(wfld_d_); 
    endif; 
    call data_grid_born();
    call data_output_history(); 
    call data_io_init();
    call inv_param()
    !if(encode) call data_io_green_s();  
    stat = 0
  end function
!=======================================================================================
  integer function data_clos() result(stat); call data_array_dealloc(); stat = 0; end function
!!=======================================================================================
  integer function data_oper() result(stat)
    if (adj.or.inv)call data_io_a('reed', fdata);
    stat = encode_init_data(dat1)
    stat = encode_compute_init1(nodes)
    if(inv) then
        call solver_tiny(m=mod1, d=dat1, Fop=data_lop, stepper=cgstepper, niter=niter, verb=T,resd=norm, timer=T)
    else 
        stat=data_lop(adj,F,mod1,dat1);
    end if 
    stat = encode_io(mod1,dat1)
  end function
!!=======================================================================================
  integer function data_lop(adj, add, modl, data) result(stat)
    logical,intent(in)    :: adj,  add
    real, dimension(:)    :: modl, data
        stat = 1
        call adjnull(adj,F,modl,data)
        stat = encode_compute(adj,modl,data)
  end function
!!=======================================================================================
  integer function encode_io(modl, data) result(stat)
    real, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n,dgrid%aw%n):: data
    real, dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n):: modl
        stat = 1
        if (adj.or.inv)then
            call data_io_img(modl)  
        else
            call data_io_a('rite', fdata);
        endif
        stat = 0
  end function
!!=======================================================================================
  integer function encode_compute(adj,  modl, data) result(stat)
    logical,intent(in) :: adj
    real, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n,dgrid%aw%n):: data 
    real, dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n):: modl
    real, dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n,ggrid%aw%n ):: modl2
    integer         :: iiw, iiiw, nb_nw, max_nw
            stat = encode_beg(adj,data,modl)
            do inode=1,ggrid%aw%n ; modl2(:,:,:,inode)=modl; enddo

write(0,*)
write(0,*)'+++++'
write(0,'(A10,10F25.5)')'min',minval(data),minval(data),minval(data),minval(modl),minval(modl),minval(modl)
write(0,'(A10,10F25.5)')'max',maxval(data),maxval(data),maxval(data),maxval(modl),maxval(modl),maxval(modl)
write(0,*)'++++++'
write(0,*)

            nb_nw=nint(1.*ggrid%aw%n/nodes)  
            do iiw= 1,ggrid%aw%n,nodes
                max_nw=min(iiw+nodes-1,ggrid%aw%n)
                stat = encode_compute_init_blk(iiw,max_nw)
!$OMP parallel do private(inode, iw, iiiw, w, index_aw) schedule(dynamic,1) 
                do iw=iiw, max_nw                                ! freq.  blocks
                    inode=omp_get_thread_num()+1;  
                    iiiw=iw-iiw+1
                    w=ggrid%aw%o+(iw-1)*ggrid%aw%d; index_aw=1.5+(w-dgrid%aw%o)/dgrid%aw%d
                    if(sum(cabs(csou_full(:,:,index_aw)))<dgrid%asx%n*dgrid%asy%n*mcsou) cycle
                    call encode_compute_data(adj, modl2(:,:,:,index_aw), fdata(:,:,:,:,index_aw),inode,iiiw)
                    !write(0,'(6I10,4F40.5)')inode,iw,iiw,iiiw,max_nw,index_aw, sum(fdata(:,:,:,:,index_aw))/maxval(cabs(fdata(:,:,:,:,index_aw)))
                    stat = encode_report(index_aw,w,modl2(:,:,:,index_aw),adj,inode)
                end do
!$OMP end parallel do  
            end do 
            if(adj)modl=sum(modl2,4)
        stat = encode_end(adj,data,modl)
  end function
!!=======================================================================================
  integer function  encode_beg(adj,data,modl) result(stat)
    logical,intent(in) :: adj
    real, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n,dgrid%aw%n):: data
    real, dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n):: modl
        stat=1
        if(adj.or.inv)then
            fdata=data; 
            stat=encode_transf(1)
        else
            modl=ref
        endif
        stat=0
  end function
!!=======================================================================================
  integer function  encode_end(adj,data,modl) result(stat)
    logical,intent(in) :: adj
    real, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n,dgrid%aw%n):: data
    real, dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n):: modl
        stat=1
        if(adj .and. .not. inv)then
            call data_io_img(modl)
        elseif (.not. (adj .or. inv))then
            call data_io_a('rite', fdata);
            data=fdata
        elseif(inv)then
            stat=encode_transf(-1)
            data=fdata
        endif 
        stat=0
    end function
!!=======================================================================================
  integer function encode_transf(dir) result(stat)
        integer :: dir
        stat=1
        if(dir==1)then    
                do isy=1, dgrid%asy%n;      do isx=1, dgrid%asx%n;  do igy=1, dgrid%agy%n;      do igx=1, dgrid%agx%n; 
                    call FFT1DT(fdata(igx, igy, isx, isy, :), 1);
                end do;                     end do;                 end do;                     end do;   
        elseif(dir==-1)then
                do isy=1, dgrid%asy%n;      do isx=1, dgrid%asx%n;  do igy=1, dgrid%agy%n;      do igx=1, dgrid%agx%n; 
                    call FFT1DT(fdata(igx, igy, isx, isy, :),-1);
                end do;                     end do;                 end do;                     end do;   
        else
            call seperr('Error dir encode')
        endif
        stat=0
  end function
!!=======================================================================================
  integer function  encode_init_data(data) result(stat)
    real, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n,dgrid%aw%n):: data
        stat=1
        stat=encode_transf(-1); data=fdata
        stat=0
  end function
!!=======================================================================================
  integer function lop(adj, add, modl, data) result(stat) !used for dottest
    logical,intent(in)    :: adj,  add
    real, dimension(:)    :: modl, data
    complex, dimension(size(data)):: ddata
        stat = 1
        call adjnull(adj,add,modl,data)
        ddata=data
        inode=omp_get_thread_num()+1;  
        call encode_compute_data(adj,modl,ddata,inode,1)
        data=ddata
  end function
!!=======================================================================================
  integer function  encode_report(index_aw,w,modl,adj,inode) result(stat)
    real, dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n):: modl
    integer                 :: index_aw, inode
    real                    ::  w
    logical,intent(in)      :: adj
    stat=1
    if(verb>2)then
        if(inv)then
            if(verb>3)then
                if(adj)then
                    write(0,'(A10,I5,A3,F7.3,A5,2F10.4, A6,I5, A15, F20.5, A7, I3)')'done iw: ', index_aw, '|', w,'=> (',sum(csou_full(:,:,index_aw))/size(csou_full(:,:,index_aw)),') of ',ggrid%aw%n, ' |  sum mig:', sum(modl), ' @ node', inode; 
                else
                    write(0,'(A10,I5,A3,F7.3,A5,2F10.4, A6,I5, A7, I3)')'done iw: ', index_aw, '|', w,'=> (',sum(csou_full(:,:,index_aw))/size(csou_full(:,:,index_aw)),') of ',ggrid%aw%n, ' @ node', inode; 
                endif
            endif
        else
            if(adj)then
                write(0,'(A10,I5,A3,F7.3,A5,2F10.4, A6,I5, A15, F20.5, A7, I3)')'done iw: ', index_aw, '|', w,'=> (',sum(csou_full(:,:,index_aw))/size(csou_full(:,:,index_aw)),') of ',ggrid%aw%n, ' |  sum mig:', sum(modl), ' @ node', inode; 
            else
                write(0,'(A10,I5,A3,F7.3,A5,2F10.4, A6,I5, A7, I3)')'done iw: ', index_aw, '|', w,'=> (',sum(csou_full(:,:,index_aw))/size(csou_full(:,:,index_aw)),') of ',ggrid%aw%n, ' @ node', inode; 
            endif
        endif
    endif
    stat=0
  end function
!===========================================================================================================================================================================
end module


