module gauss_mod
  use sep
  implicit none
  integer,private               :: n1,n2,n3,nh1,nh2,nh3,ierr,panel
  integer,private               :: sreed,srite,center,node,inode,i1,i2,i3
  integer, allocatable,private  ::  b1(:), b2(:), b3(:)
  integer, allocatable,private  ::  e1(:), e2(:), e3(:)
  integer, allocatable,private  :: sz1(:),sz2(:),sz3(:)
  real,allocatable,private      :: tri(:,:,:)
  real,private                  :: theta,std1,std2,std3
contains  
! get parameters
	subroutine gauss_init (n1_in,n2_in,n3_in,nh1_in,nh2_in,nh3_in,node_in,theta_in,std1_in,std2_in,std3_in)
                integer         :: n1_in,n2_in,n3_in,nh1_in,nh2_in,nh3_in,node_in
                real            :: theta_in,std1_in,std2_in,std3_in
                optional        :: theta_in,std1_in,std2_in,std3_in
                real                    :: r(2,2),u(2)
                        n1=n1_in; nh1=nh1_in
                        n2=n2_in; nh2=nh2_in
                        n3=n3_in; nh3=nh3_in
                        if(present(theta_in))then; theta=theta_in; else; theta=0.  ; endif
                        if(present(std1_in ))then; std1=std1_in  ; else; std1=nh1; endif
                        if(present(std2_in ))then; std2=std2_in  ; else; std2=nh2; endif
                        if(present(std3_in ))then; std3=std3_in  ; else; std3=nh3; endif
                        node=node_in
                        allocate(b1(node),b2(node),b3(node),e1(node),e2(node),e3(node),sz1(node),sz2(node),sz3(node))
                        allocate(tri(nh1*2+1,nh2*2+1,nh3*2+1))

                        r(1,1)=cos(theta); r(1,2)=-sin(theta);
                        r(2,1)=sin(theta); r(2,2)= cos(theta);
                        do i2 = 1, 2*nh2+1
                                do i1 = 1, 2*nh1+1
                                        u(1) = r(1,1) * (i2-(size(tri,2)+1)/2) + r(1,2) * (i1-(size(tri,1)+1)/2);
                                        u(2) = r(2,1) * (i2-(size(tri,2)+1)/2) + r(2,2) * (i1-(size(tri,1)+1)/2);
                                        tri(i1,i2,:) = gauss1_lop(u(1),std1)*gauss1_lop(u(2),std2);
                                end do
                        end do
        
                        do i3 = 1, 2*nh3+1
                                u(1) = r(1,1) * (i3-(size(tri,3)+1)/2)
                                tri(:,:,i3) = tri(:,:,i3)*gauss1_lop(u(1),std3)
                        end do
                        tri=tri/maxval(tri)
                        tri=tri/sum(tri**2)
	end subroutine

	function gauss_lop ( adj, add, xx, yy) result(stat)
		integer            :: stat
		logical,intent(in) :: adj,add
		real,dimension(:)  :: xx,yy
			call gauss_lop3(adj,add,xx,yy ); stat=0
	end function

	subroutine gauss_lop3(adj,add,in,out)
        	real,dimension(n1,n2,n3):: in,out
        	logical                 :: add,adj
                integer, external       :: omp_get_thread_num
                        call omp_set_num_threads(node)
                        !$OMP parallel private(i1,i2,i3,center,inode)
                       inode=omp_get_thread_num()+1;
                        !$OMP do schedule (static)
                        do i3=1,n3
                                b3(inode)=max( 1,i3-nh3)
                                e3(inode)=min(n3,i3+nh3)
                                sz3(inode)=e3(inode)-b3(inode)+1
                                do i2=1,n2 
                                        b2(inode)=max(1 ,i2-nh2)
                                        e2(inode)=min(n2,i2+nh2)
                                        sz2(inode)=e2(inode)-b2(inode)+1
                                        do i1=1,n1
                                                b1(inode)=max(1 ,i1-nh1)
                                                e1(inode)=min(n1,i1+nh1)
                                                sz1(inode)=e1(inode)-b1(inode)+1
!                                                out(i1,i2,i3)=sum(in(b1(inode):e1(inode),b2(inode):e2(inode),b3(inode):e3(inode))*tri)
                                                out(i1,i2,i3)=sum(in(b1(inode):e1(inode),b2(inode):e2(inode),b3(inode):e3(inode)))/(sz1(inode)*sz2(inode)*sz3(inode))
                                        end do
                                end do
                        end do 
                        !$OMP end do
                        !$OMP end parallel
        end subroutine


	subroutine gauss_clos()
                deallocate(b1,b2,b3,e1,e2,e3,sz1,sz2,sz3,tri)
	end subroutine
	
	
        real function gauss1_lop(x,std)
                real :: x, std, pi
                pi=3.14159265358979323846
                gauss1_lop = exp(-x**2/(2*std**2)) / (std*sqrt(2*pi))
                return ;
        end function
	
        
end module
