module grid_mod
  use global_params_mod
  implicit none

contains
!!!!!!!!!!!!!!!!  ###########
subroutine propagation_grid()
    real    :: pgrid_xmin, pgrid_xmax, pgrid_ymin, pgrid_ymax, pgrid_zmin, pgrid_zmax

    pgrid%ax%d  = vgrid%ax%d        ; pgrid%ay%d  = vgrid%ay%d        ; pgrid%az%d  = vgrid%az%d

    pgrid%gxmin = ggrid%surface_xmin; pgrid%gymin = ggrid%surface_ymin; 
    pgrid%gxmax = ggrid%surface_xmax; pgrid%gymax = ggrid%surface_ymax;

    !################################################
    pgrid_xmin = pgrid%gxmin; pgrid_xmax = pgrid%gxmax; pgrid_ymin = pgrid%gymin; pgrid_ymax = pgrid%gymax
    pgrid_zmin = max(min_z,  vgrid%az%o  ) ;  
    pgrid_zmax = min(max_z, (vgrid%az%n-1)*vgrid%az%d)

    !!.. add the padding on both x and y direction
    pgrid_xmin = pgrid_xmin - pgrid%nxpad_beg*pgrid%ax%d; pgrid_xmax = pgrid_xmax + pgrid%nxpad_end*pgrid%ax%d
    pgrid_ymin = pgrid_ymin - pgrid%nypad_beg*pgrid%ay%d; pgrid_ymax = pgrid_ymax + pgrid%nypad_end*pgrid%ay%d

    !!.. make sure the propagation grid is within the velocity model 
    pgrid_xmin = max(pgrid_xmin, vgrid%ax%o); pgrid_xmax = min(pgrid_xmax, vgrid%ax%o + (vgrid%ax%n-1)*vgrid%ax%d)
    pgrid_ymin = max(pgrid_ymin, vgrid%ay%o); pgrid_ymax = min(pgrid_ymax, vgrid%ay%o + (vgrid%ay%n-1)*vgrid%ay%d)
    !################################################
    !!.. propagation grid 
    pgrid%ax%o = pgrid_xmin; pgrid%ax%n = 1.5 + (pgrid_xmax - pgrid_xmin)/pgrid%ax%d
    pgrid%ay%o = pgrid_ymin; pgrid%ay%n = 1.5 + (pgrid_ymax - pgrid_ymin)/pgrid%ay%d
    pgrid%az%n = min(1.5 + (pgrid_zmax - pgrid_zmin)/pgrid%az%d, 1.*vgrid%az%n)

    write(*,*)'=================================='
    write(*,*)'prid%ax', pgrid%ax; write(*,*)'prid%ay', pgrid%ay; write(*,*)'prid%az', pgrid%az
    write(*,*)'=================================='

    !!.. the first and last index of surface x and y on the propagation grid (for output use)
    ggrid%index_surpro_xmin  = 1.5 + (ggrid%surface_xmin - pgrid%ax%o) / pgrid%ax%d; ggrid%index_surpro_xmax  = 1.5 + (ggrid%surface_xmax - pgrid%ax%o) / pgrid%ax%d
    ggrid%index_surpro_ymin  = 1.5 + (ggrid%surface_ymin - pgrid%ay%o) / pgrid%ay%d; ggrid%index_surpro_ymax  = 1.5 + (ggrid%surface_ymax - pgrid%ay%o) / pgrid%ay%d
    !!....first and last index of geoph. and source
    pgrid%index_datpro_gxmin = 1.5 + (pgrid%gxmin        - pgrid_xmin )/ pgrid%ax%d; pgrid%index_datpro_gymin = 1.5 + (pgrid%gymin        - pgrid_ymin ) / pgrid%ay%d  
    pgrid%index_datpro_gxmax = 1.5 + (pgrid%gxmax        - pgrid_xmin )/ pgrid%ax%d; pgrid%index_datpro_gymax = 1.5 + (pgrid%gymax        - pgrid_ymin ) / pgrid%ay%d 
    !!....first and last index of propagation grid in the velocity model
    pgrid%index_xmin         = 1.5 + (pgrid_xmin         - vgrid%ax%o )/ pgrid%ax%d; pgrid%index_ymin         = 1.5 + (pgrid_ymin         - vgrid%ay%o ) / pgrid%ay%d  
    pgrid%index_xmax         = 1.5 + (pgrid_xmax         - vgrid%ax%o )/ pgrid%ax%d; pgrid%index_ymax         = 1.5 + (pgrid_ymax         - vgrid%ay%o ) / pgrid%ay%d  
    !!...first index of shots in propagation grid
    pgrid%index_datpro_sx    = 1.5 + (dgrid%asx%o        - pgrid_xmin )/ pgrid%ax%d; pgrid%index_datpro_sy    = 1.5 + (dgrid%asy%o        - pgrid_ymin ) / pgrid%ay%d  
  
  end subroutine

!######################################################################
  subroutine green_grid()
        logical :: center_sou
    call from_param("center_sou", center_sou,F)
    if (max_z==-99999.) max_z= vgrid%az%o+vgrid%az%d*(vgrid%az%n-1); imax_z=1.5+(max_z-vgrid%az%o)/vgrid%az%d;  
    if (min_z==-99999.) min_z= vgrid%az%o                          ; imin_z=1.5+(min_z-vgrid%az%o)/vgrid%az%d;  
    if(wgrid%asx%n>1)ggrid%ax_surf=wgrid%asx; if(encode)ggrid%ax_surf%n=1
    if(wgrid%asy%n>1)ggrid%ay_surf=wgrid%asy; if(encode)ggrid%ay_surf%n=1
	
	if (ggrid%ax_modl%n==-99999)then;ggrid%ax_modl=vgrid%ax;pgrid%nxpad_beg=0.;endif
	if (ggrid%ay_modl%n==-99999)then;ggrid%ay_modl=vgrid%ay;pgrid%nypad_beg=0.;endif
	if (ggrid%az_modl%n==-99999)then;ggrid%az_modl=vgrid%az;                  ;endif 

	if (ggrid%ax_surf%n==-99999)ggrid%ax_surf%n=1; if (ggrid%ax_surf%d==-99999)ggrid%ax_surf%d=vgrid%ax%d;if (ggrid%ax_surf%o==-99999)ggrid%ax_surf%o=vgrid%ax%o+(nint(vgrid%ax%n/2.)-1)*vgrid%ax%d;
	if (ggrid%ay_surf%n==-99999)ggrid%ay_surf%n=1; if (ggrid%ay_surf%d==-99999)ggrid%ay_surf%d=vgrid%ay%d;if (ggrid%ay_surf%o==-99999)ggrid%ay_surf%o=vgrid%ay%o+(nint(vgrid%ay%n/2.)-1)*vgrid%ay%d;
    
    if(ldata)then
        ggrid%ax_surf=dgrid%agx
        ggrid%ay_surf=dgrid%agy
    endif

    ggrid%index_mod_minx=max(1.,1.5+abs(ggrid%ax_modl%o-vgrid%ax%o)/vgrid%ax%d); ggrid%index_mod_maxx=min(ggrid%ax_modl%n, vgrid%ax%n) 
    ggrid%index_mod_miny=max(1.,1.5+abs(ggrid%ay_modl%o-vgrid%ay%o)/vgrid%ay%d); ggrid%index_mod_maxy=min(ggrid%ay_modl%n, vgrid%ay%n) 
    ggrid%index_mod_minz=max(1.,1.5+abs(ggrid%az_modl%o-vgrid%az%o)/vgrid%az%d); ggrid%index_mod_maxz=min(ggrid%az_modl%n, vgrid%az%n) 
    if (center_sou)then
        ggrid%ax_surf%n=1; ggrid%ax_surf%o=vgrid%ax%o+(vgrid%ax%n/2.-1)*vgrid%ax%d
        ggrid%ay_surf%n=1; ggrid%ay_surf%o=vgrid%ay%o+(vgrid%ay%n/2.-1)*vgrid%ay%d
    endif
    write(0,*)ggrid%ax_surf%o,vgrid%ax%o,(vgrid%ax%n/2.-1),vgrid%ax%d
    write(0,*)ggrid%ay_surf%o,vgrid%ay%o,(vgrid%ay%n/2.-1),vgrid%ay%d


    write(0,'(A65)')'========================Summary greens============================'
    write(0,'(A2,A15,A2,2(F10.3,2X),A2)') '|','mimz/maxz      ','|',min_z,max_z,'|'
    write(0,'(A2,A15,A2,2(I10  ,2X),A2)') '|','iminz/imaxz    ','|',imin_z,imax_z,'|'
    write(0,'(A2,A15,A2,2(I10,2X),A2)') '|','ix, min/max      ','|',ggrid%index_mod_minx, ggrid%index_mod_maxx,'|'
    write(0,'(A2,A15,A2,2(I10,2X),A2)') '|','iy, min/max      ','|',ggrid%index_mod_miny, ggrid%index_mod_maxy,'|'
    write(0,'(A2,A15,A2,2(I10,2X),A2)') '|','iz, min/max      ','|',ggrid%index_mod_minz, ggrid%index_mod_maxz,'|'
    write(0,'(A65)')'------------------------------------------------------------------'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','ggrid%ax_surf  ','|',ggrid%ax_surf  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','ggrid%ay_surf  ','|',ggrid%ay_surf  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','ggrid%ax_modl  ','|',ggrid%ax_modl  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','ggrid%ay_modl  ','|',ggrid%ay_modl  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','ggrid%az_modl  ','|',ggrid%az_modl  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','ggrid%aw       ','|',ggrid%aw       , '|'
    write(0,'(A65)')'------------------------------------------------------------------'
    if (wfld_d_out)then
        write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','wfld%at       ','|',dgrid%at       , '|'
    endif
    write(0,'(A65)')'------------------------------------------------------------------'
    write(0,*) '|', 'spike ===>>',spike
    write(0,'(A65)')'===================================================================='
  end subroutine

!!!!!!!!!!!!!!!!  ###########
subroutine green_pro_grid(igx,igy,progrid)
    type(pro_grid) :: progrid   
    real           :: progrid_xmin, progrid_xmax, progrid_ymin, progrid_ymax, progrid_zmin, progrid_zmax
    integer        :: igx,igy    

    progrid%ax%d  = vgrid%ax%d        ; progrid%ay%d  = vgrid%ay%d        ; progrid%az%d  = vgrid%az%d

    call from_param("nxpad_beg", progrid%nxpad_beg, 0); call from_param("nxpad_end", progrid%nxpad_end, 0); nxpad=(pgrid%nxpad_beg+pgrid%nxpad_end)/2.
    call from_param("nypad_beg", progrid%nypad_beg, 0); call from_param("nypad_end", progrid%nypad_end, 0); nypad=(pgrid%nypad_beg+pgrid%nypad_end)/2.

!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 
!    write(0,*)'==========================================================>'
!    write(0,*)'geoph. @ ixg/iyg in vel. model:',igx,igy
    progrid_xmin = min(ggrid%ax_surf%o, ggrid%ax_modl%o); progrid_xmax = max(ggrid%ax_surf%o + (ggrid%ax_surf%n-1)*ggrid%ax_surf%d, ggrid%ax_modl%o + (ggrid%ax_modl%n-1)*ggrid%ax_modl%d)
    progrid_ymin = min(ggrid%ay_surf%o, ggrid%ay_modl%o); progrid_ymax = max(ggrid%ay_surf%o + (ggrid%ay_surf%n-1)*ggrid%ay_surf%d, ggrid%ay_modl%o + (ggrid%ay_modl%n-1)*ggrid%ay_modl%d)
    progrid_zmin = max(min_z          , vgrid%az%o                     ); progrid_zmax = min(max_z                                                ,(vgrid%az%n       -1)*vgrid%az%d)
!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 

    !!.. add the padding on both x and y direction
    progrid_xmin = progrid_xmin - progrid%nxpad_beg*progrid%ax%d; progrid_xmax = progrid_xmax + progrid%nxpad_end*progrid%ax%d
    progrid_ymin = progrid_ymin - progrid%nypad_beg*progrid%ay%d; progrid_ymax = progrid_ymax + progrid%nypad_end*progrid%ay%d
!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 

    !!.. make sure the propagation grid is within the velocity model 
    progrid_xmin = max(progrid_xmin, vgrid%ax%o); progrid_xmax = min(progrid_xmax, vgrid%ax%o + (vgrid%ax%n-1)*vgrid%ax%d)
    progrid_ymin = max(progrid_ymin, vgrid%ay%o); progrid_ymax = min(progrid_ymax, vgrid%ay%o + (vgrid%ay%n-1)*vgrid%ay%d)
!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 
    !################################################
    !!.. propagation grid 
    progrid%ax%o = progrid_xmin; progrid%ax%n =     1.5 + (progrid_xmax - progrid_xmin)/progrid%ax%d
    progrid%ay%o = progrid_ymin; progrid%ay%n =     1.5 + (progrid_ymax - progrid_ymin)/progrid%ay%d
    progrid%az%o = progrid_zmin; progrid%az%n = min(1.5 + (progrid_zmax - progrid_zmin)/progrid%az%d, 1.*vgrid%az%n)
                                 progrid%az%n = min(progrid%az%n,int(1.5+(ggrid%az_modl%o+(ggrid%az_modl%n-1)*ggrid%az_modl%d - progrid%az%o)/progrid%az%d))

    !!....first and last index of propagation grid in the velocity model
    progrid%index_xmin         = 1.5 + (progrid_xmin - vgrid%ax%o )/ progrid%ax%d; progrid%index_ymin         = 1.5 + (progrid_ymin - vgrid%ay%o) / progrid%ay%d  
    progrid%index_xmax         = 1.5 + (progrid_xmax - vgrid%ax%o )/ progrid%ax%d; progrid%index_ymax         = 1.5 + (progrid_ymax - vgrid%ay%o) / progrid%ay%d  
    !!.. index of the geoph. in the propagation grid
    progrid%index_datpro_sx  = max(1,igx - progrid%index_xmin+1)  
    progrid%index_datpro_sy  = max(1,igy - progrid%index_ymin+1)
    !!....index of 1st and last point of greens grid in propagation grid
    ggrid%index_modpro_minx   = 1.5 + (ggrid%ax_modl%o-progrid_xmin)/progrid%ax%d; ggrid%index_modpro_maxx=1.5 + (ggrid%ax_modl%o+(ggrid%ax_modl%n-1)*ggrid%ax_modl%d - progrid%ax%o)/progrid%ax%d
    ggrid%index_modpro_miny   = 1.5 + (ggrid%ay_modl%o-progrid_ymin)/progrid%ay%d; ggrid%index_modpro_maxy=1.5 + (ggrid%ay_modl%o+(ggrid%ay_modl%n-1)*ggrid%ay_modl%d - progrid%ay%o)/progrid%ay%d

!    write(0,*)'----------------------------------------------------------!'
!    write(0,*)'prid%ax   ', progrid%ax        ; write(0,*)'prid%ay   ', progrid%ay; write(0,*)'prid%az   ', progrid%az
!    write(0,*)'----------------------------------------------------------'
!    write(0,*)'progrid%xmin/max           ',  progrid_xmin, progrid_xmax ; 
!    write(0,*)'progrid%index_x            ', progrid%index_xmin, progrid%index_xmax ; 
!    write(0,*)'progrid%index_y            ', progrid%index_ymin, progrid%index_ymax ; 
!    write(0,*)'ggrid%index_pminx/maxx     ', ggrid%index_modpro_minx, ggrid%index_modpro_maxx  
!    write(0,*)'ggrid%index_pminx/maxy     ', ggrid%index_modpro_miny, ggrid%index_modpro_maxy  
!    write(0,*)'progrid%index_sx/sy        ', progrid%index_datpro_sx, progrid%index_datpro_sy
 end subroutine
!!!!!!!!!!!!!!!!  ###########
subroutine mig_pro_grid(isx, isy, progrid,ooogrid)
    type(pro_grid) :: progrid   
    type(out_grid) :: ooogrid   
    real           :: progrid_xmin, progrid_xmax, progrid_ymin, progrid_ymax, progrid_zmin, progrid_zmax
    integer        :: isx,isy    

    progrid%ax%d  = vgrid%ax%d        ; progrid%ay%d   = vgrid%ay%d      ; progrid%az%d  = vgrid%az%d
    ooogrid%ax%d  = vgrid%ax%d        ; ooogrid%ay%d   = vgrid%ay%d      ; ooogrid%az%d  = vgrid%az%d
    ooogrid%ahx%d = igrid%ahx%d       ; ooogrid%ahy%d  = igrid%ahy%d       
    ooogrid%ahx%o = igrid%ahx%o       ; ooogrid%ahy%o  = igrid%ahy%o       
    ooogrid%ahx%n = igrid%ahx%n       ; ooogrid%ahy%n  = igrid%ahy%n       

    ! write(0,*)'==========================================================>'
    ! write(0,*)'shot. @ isx/isy in vel. model:',isx, isy
    progrid_xmin = min(dgrid%agx%o, vgrid%ax%o + (isx-1)*vgrid%ax%d); progrid_xmax = max(dgrid%agx%o + (dgrid%agx%n-1)*dgrid%agx%d, vgrid%ax%o + (isx-1)*vgrid%ax%d)
    progrid_ymin = min(dgrid%agy%o, vgrid%ay%o + (isy-1)*vgrid%ay%d); progrid_ymax = max(dgrid%agy%o + (dgrid%agy%n-1)*dgrid%agy%d, vgrid%ay%o + (isy-1)*vgrid%ay%d)
    if (min_z == -99999.) min_z=vgrid%az%o                            ; progrid_zmin = min_z 
    if (max_z == -99999.) max_z=vgrid%az%o + (vgrid%az%n-1)*vgrid%az%d; progrid_zmax = max_z 
!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 
!    write(0,*)'1', dgrid%agx%o,dgrid%agy%o,dgrid%agx%o + (dgrid%agx%n-1)*dgrid%agx%d, vgrid%ax%o + (isx-1)*vgrid%ax%d, isx
 

    ooogrid%ax%o=progrid_xmin; ooogrid%ax%n=1.5+(progrid_xmax-progrid_xmin)/ooogrid%ax%d
    ooogrid%ay%o=progrid_ymin; ooogrid%ay%n=1.5+(progrid_ymax-progrid_ymin)/ooogrid%ay%d
    ooogrid%az%o=progrid_zmin; ooogrid%az%n=1.5+(progrid_zmax-progrid_zmin)/ooogrid%az%d

    !!.. add the padding on both x and y direction
    call from_param("nxpad_beg", progrid%nxpad_beg, 0); call from_param("nxpad_end", progrid%nxpad_end, 0); nxpad=(pgrid%nxpad_beg+pgrid%nxpad_end)/2.
    call from_param("nypad_beg", progrid%nypad_beg, 0); call from_param("nypad_end", progrid%nypad_end, 0); nypad=(pgrid%nypad_beg+pgrid%nypad_end)/2.
    progrid_xmin = progrid_xmin - progrid%nxpad_beg*progrid%ax%d; progrid_xmax = progrid_xmax + progrid%nxpad_end*progrid%ax%d
    progrid_ymin = progrid_ymin - progrid%nypad_beg*progrid%ay%d; progrid_ymax = progrid_ymax + progrid%nypad_end*progrid%ay%d
!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 

    !!.. make sure the propagation grid is within the velocity model 
    progrid_xmin = max(progrid_xmin, vgrid%ax%o); progrid_xmax = min(progrid_xmax, vgrid%ax%o + (vgrid%ax%n-1)*vgrid%ax%d)
    progrid_ymin = max(progrid_ymin, vgrid%ay%o); progrid_ymax = min(progrid_ymax, vgrid%ay%o + (vgrid%ay%n-1)*vgrid%ay%d)
!    write(0,*)'prid_xmin/max', progrid_xmin,progrid_xmax; write(0,*)'prid_ymin/max', progrid_ymin,progrid_ymax;  write(0,*)'prid_zmin/max', progrid_zmin,progrid_zmax; 
    !################################################
    !!.. propagation grid 
    progrid%ax%o = progrid_xmin; progrid%ax%n =     1.5 + (progrid_xmax - progrid_xmin)/progrid%ax%d
    progrid%ay%o = progrid_ymin; progrid%ay%n =     1.5 + (progrid_ymax - progrid_ymin)/progrid%ay%d
    progrid%az%o = progrid_zmin; progrid%az%n = min(1.5 + (progrid_zmax - progrid_zmin)/progrid%az%d, 1.*vgrid%az%n)

    !!....first and last index of propagation grid in the velocity model
    progrid%index_xmin         = 1.5 + (progrid_xmin - vgrid%ax%o )/ progrid%ax%d; progrid%index_ymin         = 1.5 + (progrid_ymin - vgrid%ay%o) / progrid%ay%d  
    progrid%index_xmax         = 1.5 + (progrid_xmax - vgrid%ax%o )/ progrid%ax%d; progrid%index_ymax         = 1.5 + (progrid_ymax - vgrid%ay%o) / progrid%ay%d  
    !!.. index of the source. in the propagation grid
    progrid%index_datpro_sx  = max(1,isx - progrid%index_xmin+1)  
    progrid%index_datpro_sy  = max(1,isy - progrid%index_ymin+1)
    !!.. index of 1st and last geoph. in the velocity model
    progrid%gxmin  = max(vgrid%ax%o, dgrid%agx%o )                                                       ; progrid%index_gxmin  = 1.5 + (progrid%gxmin-vgrid%ax%o)/vgrid%ax%d  
    progrid%gymin  = max(vgrid%ay%o, dgrid%agy%o )                                                       ; progrid%index_gymin  = 1.5 + (progrid%gymin-vgrid%ay%o)/vgrid%ay%d  
    progrid%gxmax  = min(vgrid%ax%o+(vgrid%ax%n-1)*vgrid%ax%d, dgrid%agx%o + (dgrid%agx%n-1)*dgrid%agx%d); progrid%index_gxmax  = 1.5 + (progrid%gxmax-vgrid%ax%o)/vgrid%ax%d  
    progrid%gymax  = min(vgrid%ay%o+(vgrid%ay%n-1)*vgrid%ay%d, dgrid%agy%o + (dgrid%agy%n-1)*dgrid%agy%d); progrid%index_gymax  = 1.5 + (progrid%gymax-vgrid%ay%o)/vgrid%ay%d 
    !!.. index of 1st and last geoph. in the propagation grid
    progrid%index_datpro_gxmin  = max(1.,progrid%index_gxmin - progrid%index_xmin+1.);  progrid%index_datpro_gymin  = max(1.,progrid%gymin - progrid%index_ymin+1.)
    progrid%index_datpro_gxmax  = max(1.,progrid%index_gxmax - progrid%index_xmin+1.);  progrid%index_datpro_gymax  = max(1.,progrid%gymax - progrid%index_ymin+1.)
    !!....index of 1st and last point of output grid in propagation grid
    ooogrid%index_pro_xmin   = 1.5 + (ooogrid%ax%o-progrid_xmin)/progrid%ax%d; ooogrid%index_pro_xmax=1.5 + (ooogrid%ax%o+(ooogrid%ax%n-1)*ooogrid%ax%d - progrid%ax%o)/progrid%ax%d
    ooogrid%index_pro_ymin   = 1.5 + (ooogrid%ay%o-progrid_ymin)/progrid%ay%d; ooogrid%index_pro_ymax=1.5 + (ooogrid%ay%o+(ooogrid%ay%n-1)*ooogrid%ay%d - progrid%ay%o)/progrid%ay%d
    ooogrid%index_pro_zmin   = 1.5 + (ooogrid%az%o-progrid_zmin)/progrid%az%d; ooogrid%index_pro_zmax=1.5 + (ooogrid%az%o+(ooogrid%az%n-1)*ooogrid%az%d - progrid%az%o)/progrid%az%d
    !!....index of 1st and last point of output grid in velocity grid
    ooogrid%index_img_xmin   = 1.5 + (ooogrid%ax%o-vgrid%ax%o)/progrid%ax%d; ooogrid%index_img_xmax=1.5 + (ooogrid%ax%o+(ooogrid%ax%n-1)*ooogrid%ax%d - vgrid%ax%o)/vgrid%ax%d
    ooogrid%index_img_ymin   = 1.5 + (ooogrid%ay%o-vgrid%ay%o)/progrid%ay%d; ooogrid%index_img_ymax=1.5 + (ooogrid%ay%o+(ooogrid%ay%n-1)*ooogrid%ay%d - vgrid%ay%o)/vgrid%ay%d
    ooogrid%index_img_zmin   = 1.5 + (ooogrid%az%o-vgrid%az%o)/progrid%az%d; ooogrid%index_img_zmax=1.5 + (ooogrid%ay%o+(ooogrid%az%n-1)*ooogrid%az%d - vgrid%az%o)/vgrid%az%d

!     write(0,*)'----------------------------------------------------------!'
!     write(0,*)'prid%ax   ', progrid%ax        ; write(0,*)'prid%ay   ', progrid%ay; write(0,*)'prid%az   ', progrid%az
!     write(0,*)'orid%ax   ', ooogrid%ax          ; write(0,*)'orid%ay   ', ooogrid%ay  ; write(0,*)'orid%az   ', ooogrid%az
!     write(0,*)'orid%ahx  ', ooogrid%ahx         ; write(0,*)'orid%ahy  ', ooogrid%ahy  
!     write(0,*)'----------------------------------------------------------'
!     write(0,*)'progrid%_x                 ',  progrid_xmin, progrid_xmax ; 
!     write(0,*)'progrid%index_x            ', progrid%index_xmin, progrid%index_xmax ; 
!     write(0,*)'progrid%index_y            ', progrid%index_ymin, progrid%index_ymax ; 
!     write(0,*)'ogrid%index_pminx/maxx     ', ooogrid%index_pro_xmin, ooogrid%index_pro_xmax  
!     write(0,*)'ogrid%index_pminy/maxy     ', ooogrid%index_pro_ymin, ooogrid%index_pro_ymax  
!     write(0,*)'ogrid%index_iminx/maxx     ', ooogrid%index_img_xmin, ooogrid%index_img_xmax  
!     write(0,*)'ogrid%index_iminy/maxy     ', ooogrid%index_img_ymin, ooogrid%index_img_ymax  
!     write(0,*)'progrid%index_sx/sy        ', progrid%index_datpro_sx, progrid%index_datpro_sy  
!     write(0,*)'progrid%index_pgxmin/pgxmax', progrid%index_datpro_gxmin, progrid%index_datpro_gxmax  
!     write(0,*)'progrid%index_igxmin/igymax  ', progrid%index_gxmin, progrid%index_gxmax  
!     write(0,*)'----------------------------------------------------------'
  end subroutine
!######################################################################
  subroutine velocity_grid()
    call auxin(vel_); call ainaxis(vel_,vgrid%ax,1); call ainaxis(vel_,vgrid%ay,2); call ainaxis(vel_,vgrid%az,3)
    if (max_z==-99999.) max_z= vgrid%az%o+vgrid%az%d*(vgrid%az%n-1); imax_z=1.5+(max_z-vgrid%az%o)/vgrid%az%d;  
    if (min_z==-99999.) min_z= vgrid%az%o                          ; imin_z=1.5+(min_z-vgrid%az%o)/vgrid%az%d;  

    write(0,'(A65)')'=========================Summary velocity==============================='
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|', 'x             ','|',vgrid%ax, '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|', 'y             ','|',vgrid%ay, '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|', 'z             ','|',vgrid%az, '|'
    write(0,'(A65)')'===================================================================='
  end subroutine
!######################################################################
 subroutine migration_grid()
    call auxinout(mig_)
    igrid%ax%o=max(min_x,vgrid%ax%o); igrid%ay%o=max(min_y,vgrid%ay%o); igrid%az%o=max(min_z,vgrid%az%o)
    igrid%ax%n=vgrid%ax%n           ; igrid%ay%n=vgrid%ay%n           ; igrid%az%n=vgrid%az%n
    igrid%ax%d=vgrid%ax%d           ; igrid%ay%d=vgrid%ay%d           ; igrid%az%d=vgrid%az%d           ;igrid%ahx%d=igrid%ax%d; igrid%ahy%d=igrid%ay%d
    if (max_hx ==-99999. .and. min_hx==-99999. ) then;  
        igrid%ahx%n=1; igrid%ahx%o=0.; 
    elseif (min_hx==-99999. .and. max_hx /= -99999.) then; 
        min_hx=-max_hx; igrid%ahx%n=1.5 + (max_hx - min_hx)/igrid%ax%d; igrid%ahx%o=min_hx;  
    elseif (min_hx/=-99999. .and. max_hx == -99999.) then; 
        max_hx=-min_hx; igrid%ahx%n=1.5 + (max_hx - min_hx)/igrid%ax%d; igrid%ahx%o=min_hx;
    else
                        igrid%ahx%n=1.5 + (max_hx - min_hx)/igrid%ax%d; igrid%ahx%o=min_hx;
    endif
    if (max_hy ==-99999. .and. min_hy==-99999. ) then;  
        igrid%ahy%n=1; igrid%ahy%o=0.; 
    elseif (min_hy==-99999. .and. max_hy /= -99999.) then; 
        min_hy=-max_hy; igrid%ahy%n=1.5 + (max_hy - min_hy)/igrid%ay%d; igrid%ahy%o=min_hy;  
    elseif (min_hy/=-99999. .and. max_hy == -99999.) then; 
        max_hy=-min_hy; igrid%ahy%n=1.5 + (max_hy - min_hy)/igrid%ay%d; igrid%ahy%o=min_hy; 
    else
                        igrid%ahy%n=1.5 + (max_hy - min_hy)/igrid%ay%d; igrid%ahy%o=min_hy;
    endif

    write(0,'(A65)')'========================Summary migration============================'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','hx  ','|',igrid%ahx  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','hy  ','|',igrid%ahy  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','x   ','|',igrid%ax   , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','y   ','|',igrid%ay   , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','z   ','|',igrid%az   , '|'
    write(0,'(A65)')'===================================================================='
  end subroutine
!=============================================
  subroutine migration_grid_in()
    call auxin(mig_); call ainaxis(mig_,igrid%ahx,1); call ainaxis(mig_,igrid%ahy,2); call ainaxis(mig_,igrid%ax,3); call ainaxis(mig_,igrid%ay,4); call ainaxis(mig_,igrid%az,5); 

    write(0,*)'Summary migration======================================'
    write(0,*) 'hx             : ',igrid%ahx
    write(0,*) 'hy             : ',igrid%ahy
    write(0,*) 'x              : ',igrid%ax
    write(0,*) 'y              : ',igrid%ay
    write(0,*) 'z              : ',igrid%az
    write(0,*)'======================================================='
  end subroutine
!######################################################################
  subroutine data_grid()
    call auxin(data_)
    call ainaxis(data_,dgrid%agx,1); call ainaxis(data_,dgrid%agy,2); call ainaxis(data_,dgrid%asx,4); call ainaxis(data_,dgrid%asy,5); 
    !call ainaxis(data_,dgrid%aw,3); 

    dgrid%nsh = dgrid%asx%n * dgrid%asy%n               !num  of sources..
    dgrid%ntr = dgrid%agx%n * dgrid%agy%n               !num. of traces/recievers

    write(0,'(A65)')'========================Summary data ================================='
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','shot_x ','|',dgrid%asx  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','shot_y ','|',dgrid%asy  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','geoph_x','|',dgrid%agx  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','geoph_y','|',dgrid%agy  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','freq.  ','|',dgrid%aw   , '|'
    write(0,'(A65)')'===================================================================='

    write(0,'(A65)')'========================Summary hole ================================='
        if(.not. hole)then 
                write(0,*)'No hole in data..good..No need to worry...'
        else
                ih_sx1=1.5+(hole_sx1-dgrid%asx%o)/dgrid%asx%d;ih_sx2=1.5+(hole_sx2-dgrid%asx%o)/dgrid%asx%d
                ih_sy1=1.5+(hole_sy1-dgrid%asy%o)/dgrid%asy%d;ih_sy2=1.5+(hole_sy2-dgrid%asy%o)/dgrid%asy%d
                ih_gx1=1.5+(hole_gx1-dgrid%agx%o)/dgrid%agx%d;ih_gx2=1.5+(hole_gx2-dgrid%agx%o)/dgrid%agx%d
                ih_gy1=1.5+(hole_gy1-dgrid%agy%o)/dgrid%agy%d;ih_gy2=1.5+(hole_gy2-dgrid%agy%o)/dgrid%agy%d
                if(hole_sx1<=hole_sx2.and.ih_sx1>1.and.ih_sx2<dgrid%asx%n) write(0,*)'hole sx           : ', ih_sx1, ih_sx2
                if(hole_sy1<=hole_sy2.and.ih_sy1>1.and.ih_sx2<dgrid%asx%n) write(0,*)'hole sy           : ', ih_sy1, ih_sy2
                if(hole_gx1<=hole_gx2.and.ih_gx1>1.and.ih_gx2<dgrid%agx%n) write(0,*)'hole gx           : ', ih_gx1, ih_gx2
                if(hole_gy1<=hole_gy2.and.ih_gy1>1.and.ih_gy2<dgrid%agy%n) write(0,*)'hole gy           : ', ih_gy1, ih_gy2
        endif
    write(0,'(A65)')'===================================================================='
  end subroutine
!######################################################################
  subroutine data_grid_born()
    call from_param("data_min_x"  , data_min_x   ,  ggrid%ax_modl%o ); call from_param("data_max_x"  , data_max_x   ,  ggrid%ax_modl%o+(ggrid%ax_modl%n-1)*ggrid%ax_modl%d );
    call from_param("data_min_y"  , data_min_y   ,  ggrid%ay_modl%o ); call from_param("data_max_y"  , data_max_y   ,  ggrid%ay_modl%o+(ggrid%ay_modl%n-1)*ggrid%ay_modl%d );

    data_imin_z=max(1.,1.5+(data_min_z-ggrid%az_modl%o)/ggrid%az_modl%d); data_imax_z=min(ggrid%az_modl%n,nint(1.5+(data_max_z-ggrid%az_modl%o)/ggrid%az_modl%d))
    data_imin_x=max(1.,1.5+(data_min_x-ggrid%ax_modl%o)/ggrid%ax_modl%d); data_imax_x=min(ggrid%ax_modl%n,nint(1.5+(data_max_x-ggrid%ax_modl%o)/ggrid%ax_modl%d))
    data_imin_y=max(1.,1.5+(data_min_y-ggrid%ay_modl%o)/ggrid%ay_modl%d); data_imax_y=min(ggrid%ay_modl%n,nint(1.5+(data_max_y-ggrid%ay_modl%o)/ggrid%ay_modl%d))

    write(0,'(A65)')'=========================Summary data==================================='
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','dgrid%asx  ','|',dgrid%asx  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','dgrid%asy  ','|',dgrid%asy  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','dgrid%aw   ','|',dgrid%aw   , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','dgrid%at   ','|',dgrid%at   , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','dgrid%agx  ','|',dgrid%agx  , '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','dgrid%agy  ','|',dgrid%agy  , '|'
    write(0,'(A65)')'===================================================================='
  end subroutine


  end module

 subroutine table1_int()
 end subroutine
 subroutine table3_int()
 end subroutine
 subroutine table2_int()
 end subroutine

 

