module hessian_convolve_mod
  use hessian_convolve_init
  use hessian_convolve_module1
  use hessian_convolve_module2
  use hessian_convolve_module1_2d
  use hessian_convolve_module2_2d
!  use hessian_convolve_module1_2d
  use hessian_convolve_module2_3d
  use hessian_convolve_precond
  use hessian_convolve_axis
  use hessian_convolve_array
  use hessian_convolve_adjnull
  use hessian_convolve_add
  use hessian_convolve_regs
  use hessian_convolve_nhlx
  use hessian_convolve_vnhlx
  use hessian_convolve_vnhlx3d
  use chain_mod
  implicit none

contains
!#######
  integer function conv_init() result(stat) 
        stat = 1
        call hessian_conv_params_init()
        call hessian_conv_aux_init()
        call hessian_conv_axis_init()
        call hessian_conv_array_alloc()
        
        if(is2d .and. small) call hessian_init()
        if (adj .or. inv) then
            if (inv) write(*,*)'inversion............................'
            if (adj) write(*,*)'adjoint only.........................'
            call hessian_conv_inout_init()
        else
            write(*,*)'modeling only.................................'
            call hessian_conv_inout_model_init()
        endif
        call hessian_conv_summary()
        stat = 0.       
  end function
!#######
  integer function conv_clos() result(stat)
        stat = 1
        call  hessian_conv_array_clos() 
        stat = 0
  end function
!#######
  integer function conv_oper(adj, add, modl, data) result(stat)
    logical, intent(in):: adj, add
    real, dimension(:) :: modl, data
        stat = 1
       call hessian_conv_adjnull(adj,    modl,data) 
        if (is2d)then
                if(method==1)then
                        call hessian_convolve1_2d(adj,add)
                else
                        call hessian_convolve2_2d(adj,add)
                endif
        else
                if(method==1)then
                        !call hessian_convolve1_3d(adj,add)
                else
                        call hessian_convolve2_3d(adj,add)
                endif
        endif                
        call hessian_conv_add    (adj,add,modl,data)
        if (ldz_fact) call hessian_fact_dz(adj)
        if (timing2)then
                if (iter==0)then
                        call date_and_time(values=time_array_5)
                        start_t1 = time_array_5 (5) * 3600 + time_array_5 (6) * 60  + time_array_5 (7) + 0.001 * time_array_5 (8)
                        start_t0=start_t1
                        finish_t0=start_t0
                else
                        call date_and_time(values=time_array_6)
                        finish_t1 = time_array_6 (5) * 3600 + time_array_6 (6) * 60  + time_array_6 (7) + 0.001 * time_array_6 (8)
                        int_t = finish_t1-finish_t0
                        tot_t = finish_t1-start_t0
                        finish_t0=finish_t1
                        write(0,*)'Time: Interv | Spent | Remain :',int_t,tot_t, int_t*(niter-iter-1)
                endif
        endif
        riter=iter/niter; 
        iter=iter+1
        stat = 0
  end function
!#######
  integer function conv_precon_oper(adj, add, modl, data) result(stat)
    logical, intent(in):: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,    modl,data)
    call hessian_precond     (adj,add          )
    call hessian_conv_add    (adj,add,modl,data)
    stat = 0
  end function
!####### 
  integer function conv_grad_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    stat = 1
    call hessian_conv_adjnull  (adj,     modl,data)
    call hessian_conv_gradient (adj               )
    call hessian_conv_add      (adj, add,modl,data)
    stat = 0
  end function
!#######
  integer function conv_temp_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    stat = 1
    call hessian_conv_adjnull  (adj,    modl,data)
    call hessian_conv_grad_temp(adj              )
    call hessian_conv_add      (adj,add,modl,data)
    stat = 0
  end function
!#######
  integer function conv_grad_temp_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    real, dimension(nmod*nsurv) :: tmp1
    stat = 1
    call chain2 (oper1=conv_grad_oper, oper2=conv_temp_oper, adj=adj, add=.true., mod=modl, dat=data, tmp1=tmp1)
    stat = 0
  end function
!#######
  integer function conv_temp_grad_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    real, dimension(nmod*nsurv) :: tmp1
    stat = 1
    call chain2 (oper1=conv_temp_oper, oper2=conv_grad_oper, adj=adj, add=.true., mod=modl, dat=data, tmp1=tmp1)
    stat = 0
  end function
!#######
!####### 
  integer function conv_grad_oper_link(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    stat = 1
    call hessian_conv_adjnull  	   (adj,     modl,data)
    call hessian_conv_adjnull_link (         modl,data)
    call hessian_conv_gradient_link(adj               )
    call hessian_conv_add          (adj, add,modl,data)
    stat = 0
  end function
!####### 
!#######
  integer function conv_weight_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,modl,data)
    call hessian_conv_weight (adj,modl,data)
    stat = 0
  end function
!#######
  integer function conv_mask_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,modl,data)
    call hessian_conv_mask   (adj,modl,data)
    stat = 0
  end function
!####### 
  integer function conv_damp_oper(adj, add, modl, data) result(stat)
    logical, intent (in) :: adj, add
    real, dimension(:)  :: modl, data
        call adjnull (adj,add,modl,data ); 
        stat=1
        if(adj)then
                modl=modl+data
        else    
                data=data+modl
        endif
        stat=0
  end function
!#####################################
  subroutine hessian_conv_weight(adj, modl, data)
    logical, intent (in) :: adj
    real, dimension(:)  :: modl, data
    if ( adj) then
                modl = modl + data*wght
    else
                data = data + modl*wght
    end if
  end subroutine
!#####################################
  subroutine hessian_conv_mask(adj, modl, data)
    logical, intent (in) :: adj
    real, dimension(:)   :: modl, data
    if ( adj) then
        where( mask==1)
                modl = modl + data
        end where
    else
        where( mask==1)
                data = data + modl
        end where
    end if
  end subroutine

!######################################spatiotemporal
  integer function spatio_lop( adj, add, mod, dat) result(stat) 
        integer                   :: stat1,stat2 
        logical,intent(in)        :: adj,add 
        real,dimension(:)         :: mod,dat
                !write(0,*)'++++++++++++++++++',adj,add,lmd1
                !write(0,*)sum(mod),sum(dat(1:size(mod))),sum(dat(1+size(mod):)),size(mod),size(dat(1+size(mod):)),'b4'
                        call hessian_convolve_vdips_init(1,1)
                        stat= hessian_convolve_dips_oper(adj, add, mod, dat(1:size(mod) ))
                !write(0,*)sum(mod),sum(dat(1:size(mod))),sum(dat(1+size(mod):)),size(mod),size(dat(1+size(mod):)),'mid'
                        call           hessian_conv_temp(adj, add, mod, dat(1+size(mod):))
                !write(0,*)sum(mod),sum(dat(1:size(mod))),sum(dat(1+size(mod):)),size(mod),size(dat(1+size(mod):)),'af'
                !write(0,*)'++++++++++++++++++',adj,add,lmd1
  end function
!######################################spatiotemporal
  integer function spatio_int_lop( adj, add, mod, dat) result(stat)
        integer                   :: stat1,stat2 
        logical,intent(in)        :: adj,add
        real,dimension(:)         :: mod,dat
                call adjnull (adj,add,modl,data ); 
!                write(0,*)'++++++++++++++++++',adj,add,lmd1,lmd2
!                write(0,'(4g15.5,2(I8),A4)')sum(mod),sum(dat(1+0*size(mod):1*size(mod))),sum(dat(1+1*size(mod):2*size(mod))),sum(dat(1+2*size(mod):3*size(mod))),size(mod),size(dat(1+size(mod):)),'b4'
                        call hessian_convolve_vdips_init(1,1)
                        stat= hessian_convolve_dips_oper(adj, add, mod, dat(1+0*size(mod):1*size(mod)))
!                write(0,'(4g15.5,2(I8),A4)')sum(mod),sum(dat(1+0*size(mod):1*size(mod))),sum(dat(1+1*size(mod):2*size(mod))),sum(dat(1+2*size(mod):3*size(mod))),size(mod),size(dat(1+size(mod):)),'mid1'
                        call           hessian_conv_temp(adj, add, mod, dat(1+1*size(mod):2*size(mod)))
!                write(0,'(4g15.5,2(I8),A4)')sum(mod),sum(dat(1+0*size(mod):1*size(mod))),sum(dat(1+1*size(mod):2*size(mod))),sum(dat(1+2*size(mod):3*size(mod))),size(mod),size(dat(1+size(mod):)),'mid2'
                        stat= hessian_convolve_leak_oper(adj, add, mod, dat(1+2*size(mod):3*size(mod)))
!                write(0,'(4g15.5,2(I8),A4)')sum(mod),sum(dat(1+0*size(mod):1*size(mod))),sum(dat(1+1*size(mod):2*size(mod))),sum(dat(1+2*size(mod):3*size(mod))),size(mod),size(dat(1+size(mod):)),'af'
!                write(0,*)'++++++++++++++++++',adj,add,lmd1
  end function
!##################################################### leak 
  integer function spatio_grad_lop( adj, add, mod, dat) result(stat)
        integer                   :: stat1,stat2
        logical,intent(in)        :: adj,add
        real,dimension(:)         :: mod,dat
                call adjnull (adj,add,modl,data ); 
                        call hessian_convolve_vdips_init(1,1)
                        stat= hessian_convolve_dips_oper(adj, add, mod, dat(1+0*size(mod):1*size(mod)))
                        call           hessian_conv_temp(adj, add, mod, dat(1+1*size(mod):2*size(mod)))
                        stat= hessian_convolve_grad_oper(adj, add, mod, dat(1+2*size(mod):3*size(mod)))
  end function
!##################################################### leak 
!######################################spatiotemporal
  integer function spatio_lop1( adj, add, mod, dat) result(stat) 
        integer                   :: stat1,stat2 
        logical,intent(in)        :: adj,add 
        real,dimension(:)         :: mod,dat
                        call hessian_convolve_vdips_init(1,1)
                        stat= hessian_convolve_dips_oper(adj, add, mod, dat(1:size(mod) ))
                        call          hessian_conv_temp1(adj, add, mod, dat(1+size(mod):))
  end function
!######################################spatiotemporal
  integer function spatio_int_lop1( adj, add, mod, dat) result(stat)
        integer                   :: stat1,stat2 
        logical,intent(in)        :: adj,add
        real,dimension(:)         :: mod,dat
                call adjnull (adj,add,modl,data ); 
                        call hessian_convolve_vdips_init(1,1)
                        stat= hessian_convolve_dips_oper(adj, add, mod, dat(1+0*size(mod):1*size(mod)))
                        call          hessian_conv_temp1(adj, add, mod, dat(1+1*size(mod):2*size(mod)))
                        stat= hessian_convolve_leak_oper(adj, add, mod, dat(1+2*size(mod):3*size(mod)))
  end function
!######################################spatiotemporal
  integer function spatio_grad_lop1( adj, add, mod, dat) result(stat) 
        integer                   :: stat1,stat2 
        logical,intent(in)        :: adj,add
        real,dimension(:)         :: mod,dat
                call adjnull (adj,add,modl,data );
                        call hessian_convolve_vdips_init(1,1)
                        stat= hessian_convolve_dips_oper(adj, add, mod, dat(1+0*size(mod):1*size(mod)))
                        call          hessian_conv_temp1(adj, add, mod, dat(1+1*size(mod):2*size(mod)))
                        stat= hessian_convolve_grad_oper(adj, add, mod, dat(1+2*size(mod):3*size(mod)))
  end function
!##################################################### leak 
        integer function hessian_convolve_leak_oper(adj,add,xx,yy)  result(stat)
                real, dimension (:)             :: xx,yy
                integer                         :: isurv
                logical , intent(in)            :: adj,add
                stat=1
                        call adjnull (adj,add,xx,yy);
                        do isurv=1,nsurv; 
                                call leak_z (adj,add,xx((isurv-1)*nmod+1:isurv*nmod),yy((isurv-1)*nmod+1:isurv*nmod));
                        enddo
                stat=0
        end function
!##################################################### leak 
        integer function hessian_convolve_grad_oper(adj,add,xx,yy)  result(stat)
                real, dimension (:)             :: xx,yy
                integer                         :: isurv
                logical , intent(in)            :: adj,add
                stat=1
                        call adjnull (adj,add,xx,yy);
                        do isurv=1,nsurv; 
                                call grad_z (adj,xx((isurv-1)*nmod+1:isurv*nmod),yy((isurv-1)*nmod+1:isurv*nmod));
                        enddo
                stat=0
        end function
 !######################################
        integer function hessian_convolve_impedance_oper(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                stat=1; call chain0(conv_oper, hessian_convolve_grad_oper,adj,add,mod,dat,tmp); stat=0
        end function
 !######################################
        integer function spatial_impedance_oper(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                stat=1; call chain0(hessian_convolve_dips_oper,hessian_convolve_leak_oper,adj,add,mod,dat,tmp); stat=0
        end function
 !######################################
  integer function hessian_convolve_temp_oper(adj, add, modl, data) result(stat)
            logical, intent (in) :: adj, add
            real, dimension(:)  :: modl, data 
            stat = 1
            call adjnull (adj,add,modl,data);
            !write(0,*)adj,sum(data),sum(modl),'b4'
            call hessian_conv_temp(adj,add,modl,data)
            !write(0,*)adj,sum(data),sum(modl),'af'
            stat = 0
  end function
 !######################################
  integer function hessian_convolve_temp_oper1(adj, add, modl, data) result(stat)
            logical, intent (in) :: adj, add
            real, dimension(:)  :: modl, data
            stat = 1
            call adjnull (adj,add,modl,data);
            call hessian_conv_temp1(adj,add,modl,data)
            stat = 0
  end function
 !######################################
  integer function hessian_convolve_temp_oper2(adj, add, modl, data) result(stat)
            logical, intent (in) :: adj, add
            real, dimension(:)  :: modl, data
            stat = 1
            call adjnull (adj,add,modl,data);
            call hessian_conv_temp2(adj,add,modl,data)
            stat = 0
  end function
 !######################################
  integer function hessian_convolve_temp_oper3(adj, add, modl, data) result(stat)
            logical, intent (in) :: adj, add
            real, dimension(:)  :: modl, data
            stat = 1
            call adjnull (adj,add,modl,data);
            call hessian_conv_temp3(adj,add,modl,data)
            stat = 0
  end function
 !######################################
        integer function spatial_diff_oper(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                call hessian_convolve_dips_init(1,1)
                stat=1; call chain0(hessian_convolve_dips_oper,hessian_convolve_temp_oper1,adj,add,mod,dat,tmp); stat=0
        end function
 !######################################
        integer function vspatial_diff_oper(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
                call hessian_convolve_vdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_oper,hessian_convolve_temp_oper,adj,add,mod,dat,tmp); stat=0
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper1(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
                call hessian_convolve_vdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_oper,hessian_convolve_temp_oper1,adj,add,mod,dat,tmp); stat=0
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper2(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
                call hessian_convolve_vdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_oper,hessian_convolve_temp_oper2,adj,add,mod,dat,tmp); stat=0
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper3(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
                call hessian_convolve_vdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_oper,hessian_convolve_temp_oper3,adj,add,mod,dat,tmp); stat=0
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper_smth(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
                call hessian_convolve_vdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_smth,hessian_convolve_temp_oper,adj,add,mod,dat,tmp); stat=0
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper3_smth(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
                call hessian_convolve_vdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_smth,hessian_convolve_temp_oper3,adj,add,mod,dat,tmp); stat=0
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper_1var(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                stat= hessian_convolve_vdips_oper_1var_init(1)
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
                stat=1; call chain0(hessian_convolve_vdips_oper_1var,hessian_convolve_temp_oper,adj,add,mod,dat,tmp); stat=0
                stat= hessian_convolve_vdips_oper_1var_init(1)
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper2_1var(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                stat= hessian_convolve_vdips_oper_1var_init(1)
!                write(0,*)sum(dat),sum(mod),'dat,mod-b',adj,add
                stat=1; call chain0(hessian_convolve_vdips_oper_1var,hessian_convolve_temp_oper2,adj,add,mod,dat,tmp); stat=0
                stat= hessian_convolve_vdips_oper_1var_init(1)
!                write(0,*)sum(dat),sum(mod),'dat,mod-a',adj,add
        end function
 !######################################
        integer function vspatial_diff_oper_4dsmth(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                call hessian_convolve_4dvdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_smth,hessian_convolve_temp_oper,adj,add,mod,dat,tmp); stat=0
        end function
 !######################################
        integer function vspatial_diff_oper_4dsmth1(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                call hessian_convolve_4dvdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_smth,hessian_convolve_temp_oper1,adj,add,mod,dat,tmp); stat=0
        end function
 !######################################
        integer function vspatial_diff_oper_4dsmth2(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                call hessian_convolve_4dvdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_smth,hessian_convolve_temp_oper2,adj,add,mod,dat,tmp); stat=0
        end function
 !######################################
        integer function vspatial_diff_oper_4dsmth3(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                call adjnull (adj,add,mod,dat);
                call hessian_convolve_4dvdips_init(1,1)
                stat=1; call chain0(hessian_convolve_vdips_smth,hessian_convolve_temp_oper3,adj,add,mod,dat,tmp); stat=0
        end function
!#######

end module

