module hessian_convolve_precond
  use  hessian_convolve_params
  use sep_timers_mod

  implicit none

contains
!#######
  subroutine  hessian_precond(adj, add)
  logical            :: adj, add
  if(method==1)then
         if (nsurv > 5) then
                call hess_precond6(adj, modl1, modl2, modl3, modl4, modl5, modl6, data1, data2, data3, data4, data5, data6 )
         elseif (nsurv > 4) then
                call hess_precond5(adj, modl1, modl2, modl3, modl4, modl5       , data1, data2, data3, data4, data5        )
         elseif (nsurv > 3) then
                call hess_precond4(adj, modl1, modl2, modl3, modl4              , data1, data2, data3, data4               )
         elseif (nsurv > 2) then
                call hess_precond3(adj, modl1, modl2, modl3                     , data1, data2, data3                      )
         elseif (nsurv > 1) then
                call hess_precond2(adj, modl1, modl2                            , data1, data2                             )
         elseif (nsurv > 0) then 
                call hess_precond1(adj, modl1                                   , data1                                    )
         endif
     else
         if (nsurv > 5) then
                call hess_precond_6(adj, modl1, modl2, modl3, modl4, modl5, modl6, data1, data2, data3, data4, data5, data6 )
         elseif (nsurv > 4) then
                call hess_precond_5(adj, modl1, modl2, modl3, modl4, modl5       , data1, data2, data3, data4, data5        )
         elseif (nsurv > 3) then
                call hess_precond_4(adj, modl1, modl2, modl3, modl4              , data1, data2, data3, data4               )
         elseif (nsurv > 2) then
                call hess_precond_3(adj, modl1, modl2, modl3                     , data1, data2, data3                      )
         elseif (nsurv > 1) then
                call hess_precond_2(adj, modl1, modl2                            , data1, data2                             )
         elseif (nsurv > 0) then
                call hess_precond_1(adj, modl1                                   , data1                                    )
         endif
     endif
   end subroutine

! ##############################################
!/ | HH | 
  subroutine hess_precond1(adj, mod1, dat1)
    logical                       :: adj
    real         ,dimension(nmod) :: mod1
    real         ,dimension(nmod) :: dat1

              if (adj) then
                      mod1 = mod1 + hesp1*dat1
              else 
                      dat1 = dat1 + hesp1*mod1
              end if
    return
  end subroutine
!#####################################
!/ | H1 H1 |    
!/ | H1 H12|    
  subroutine hess_precond2(adj, mod1, mod2, dat1, dat2)
    logical                  :: adj
    real    ,dimension(nmod) :: mod1, mod2
    real    ,dimension(nmod) :: dat1, dat2
              if (adj) then
                      mod1 = mod1 + hesp1*dat1 -         hesp1*dat2 
                      mod2 = mod2 - hesp1*dat1 + (hesp1+hesp2)*dat2
              else 
                      dat1 = dat1 + hesp1*mod1 -         hesp1*mod2 
                      dat2 = dat2 - hesp1*mod1 + (hesp1+hesp2)*mod2 
              end if
    return
  end subroutine
!#####################################
!/ | H1 H1    0|    
!/ | H1 H12  H2|    
!/ | 0  H2  H23|    
 subroutine hess_precond3(adj, mod1, mod2, mod3, dat1, dat2, dat3)
    logical                  :: adj
    real    ,dimension(nmod) :: mod1, mod2, mod3
    real    ,dimension(nmod) :: dat1, dat2, dat3

              if (adj) then
                      mod1 = mod1 + hesp1*dat1                      - hesp1*dat2 
                      mod2 = mod2 - hesp1*dat1 + (hesp1+hesp2)*dat2 - hesp2*dat3 
                      mod3 = mod3 - hesp2*dat2 + (hesp2+hesp3)*dat3 
              else 
                      dat1 = dat1 + hesp1*mod1                      - hesp1*mod2 
                      dat2 = dat2 - hesp1*mod1 + (hesp1+hesp2)*mod2 - hesp2*mod3 
                      dat3 = dat3 - hesp2*mod2 + (hesp2+hesp3)*mod3 
              end if
    return
  end subroutine
!#####################################
!/ | H14 H24 H34 H4|    
!/ | H24 H24 H34 H4|    
!/ | H34 H34 H34 H4|    
!/ |  H4  H4  H4 H4|    
  subroutine hess_precond4(adj, mod1, mod2, mod3, mod4, dat1, dat2, dat3, dat4)
    logical                  :: adj
    real    ,dimension(nmod) :: mod1, mod2, mod3, mod4
    real    ,dimension(nmod) :: dat1, dat2, dat3, dat4

              if (adj) then
                      mod1 = mod1 + hesp1*dat1                      - hesp1*dat2 
                      mod2 = mod2 - hesp1*dat1 + (hesp1+hesp2)*dat2 - hesp2*dat3 
                      mod3 = mod3 - hesp2*dat2 + (hesp2+hesp3)*dat3 - hesp3*dat4 
                      mod4 = mod4 - hesp3*dat3 + (hesp3+hesp4)*dat4 
              else 
                      dat1 = dat1 + hesp1*mod1                      - hesp1*mod2 
                      dat2 = dat2 - hesp1*mod1 + (hesp1+hesp2)*mod2 - hesp2*mod3 
                      dat3 = dat3 - hesp2*mod2 + (hesp2+hesp3)*mod3 - hesp3*mod4 
                      dat4 = dat4 - hesp3*mod3 + (hesp3+hesp4)*mod4 
              end if
    return
  end subroutine
!#####################################
  subroutine hess_precond5(adj, mod1, mod2, mod3, mod4, mod5, dat1, dat2, dat3, dat4, dat5)
    logical               :: adj
    real, dimension(nmod) :: mod1, mod2, mod3, mod4, mod5
    real, dimension(nmod) :: dat1, dat2, dat3, dat4, dat5

              if (adj) then
                      mod1 = mod1 + hesp1*dat1                      - hesp1*dat2 
                      mod2 = mod2 - hesp1*dat1 + (hesp1+hesp2)*dat2 - hesp2*dat3 
                      mod3 = mod3 - hesp2*dat2 + (hesp2+hesp3)*dat3 - hesp3*dat4 
                      mod4 = mod4 - hesp3*dat3 + (hesp3+hesp4)*dat4 - hesp4*dat5 
                      mod5 = mod5 - hesp4*dat4 + (hesp4+hesp5)*dat5 
              else 
                      dat1 = dat1 + hesp1*mod1                      - hesp1*mod2 
                      dat2 = dat2 - hesp1*mod1 + (hesp1+hesp2)*mod2 - hesp2*mod3 
                      dat3 = dat3 - hesp2*mod2 + (hesp2+hesp3)*mod3 - hesp3*mod4 
                      dat4 = dat4 - hesp3*mod3 + (hesp3+hesp4)*mod4 - hesp4*mod5 
                      dat5 = dat5 - hesp4*mod4 + (hesp4+hesp5)*mod5 
              end if
    return
  end subroutine
!#####################################
  subroutine hess_precond6(adj, mod1, mod2, mod3, mod4, mod5, mod6, dat1, dat2, dat3, dat4, dat5, dat6)
    logical               :: adj
    real, dimension(nmod) :: mod1, mod2, mod3, mod4, mod5, mod6
    real, dimension(nmod) :: dat1, dat2, dat3, dat4, dat5, dat6
!    write(0,*)'prc',sum(mod1),sum(mod2),sum(mod3),sum(mod4),sum(mod5),sum(mod6)
!    write(0,*)'prc',sum(dat1),sum(dat2),sum(dat3),sum(dat4),sum(dat5),sum(dat6)
!    write(0,*)'modl',mod1(1:300)
              if (adj) then
                      mod1 = mod1 + hesp1*dat1                      - hesp1*dat2 
                      mod2 = mod2 - hesp1*dat1 + (hesp1+hesp2)*dat2 - hesp2*dat3 
                      mod3 = mod3 - hesp2*dat2 + (hesp2+hesp3)*dat3 - hesp3*dat4 
                      mod4 = mod4 - hesp3*dat3 + (hesp3+hesp4)*dat4 - hesp4*dat5 
                      mod5 = mod5 - hesp4*dat4 + (hesp4+hesp5)*dat5 - hesp5*dat6 
                      mod6 = mod6 - hesp5*dat5 + (hesp5+hesp6)*dat6 
              else 
                      dat1 = dat1 + hesp1*mod1                      - hesp1*mod2 
                      dat2 = dat2 - hesp1*mod1 + (hesp1+hesp2)*mod2 - hesp2*mod3 
                      dat3 = dat3 - hesp2*mod2 + (hesp2+hesp3)*mod3 - hesp3*mod4 
                      dat4 = dat4 - hesp3*mod3 + (hesp3+hesp4)*mod4 - hesp4*mod5 
                      dat5 = dat5 - hesp4*mod4 + (hesp4+hesp5)*mod5 - hesp5*mod6 
                      dat6 = dat6 - hesp5*mod5 + (hesp5+hesp6)*mod6 
              end if
!    write(0,*)'datl',dat1(1:300)
    return
  end subroutine
!#####################################
!##############################################
!##############################################
!##############################################
!##############################################
!##############################################
!/ | HH | 
  subroutine hess_precond_1(adj, mod1, dat1)
    logical                       :: adj
    real         ,dimension(nmod) :: mod1
    real         ,dimension(nmod) :: dat1

              if (adj) then
                      mod1 = mod1 + hesp1*dat1
              else 
                      dat1 = dat1 + hesp1*mod1 
              end if
    return
  end subroutine
!#####################################
!/ | H1 H1 |    
!/ | H1 H12|    
  subroutine hess_precond_2(adj, mod1, mod2, dat1, dat2)
    logical                  :: adj
    real    ,dimension(nmod) :: mod1, mod2
    real    ,dimension(nmod) :: dat1, dat2
              if (adj) then
                      mod1 = mod1 + hesp1*dat1
                      mod2 = mod2 + hesp2*dat2 
              else 
                      dat1 = dat1 + hesp1*mod1 
                      dat2 = dat2 + hesp2*mod2 
              end if
    return
  end subroutine
!#####################################
!/ | H1 H1    0|    
!/ | H1 H12  H2|    
!/ | 0  H2  H23|    
 subroutine hess_precond_3(adj, mod1, mod2, mod3, dat1, dat2, dat3)
    logical                  :: adj
    real    ,dimension(nmod) :: mod1, mod2, mod3
    real    ,dimension(nmod) :: dat1, dat2, dat3

              if (adj) then
                      mod1 = mod1 + hesp1*dat1
                      mod2 = mod2 + hesp2*dat2 
                      mod3 = mod3 + hesp3*dat3 
              else 
                      dat1 = dat1 + hesp1*mod1 
                      dat2 = dat2 + hesp2*mod2 
                      dat3 = dat3 + hesp3*mod3 
              end if

    return
  end subroutine
!#####################################
  subroutine hess_precond_4(adj, mod1, mod2, mod3, mod4, dat1, dat2, dat3, dat4)
    logical                  :: adj
    real    ,dimension(nmod) :: mod1, mod2, mod3, mod4
    real    ,dimension(nmod) :: dat1, dat2, dat3, dat4
              if (adj) then
                      mod1 = mod1 + hesp1*dat1
                      mod2 = mod2 + hesp2*dat2 
                      mod3 = mod3 + hesp3*dat3 
                      mod4 = mod4 + hesp4*dat4 
              else 
                      dat1 = dat1 + hesp1*mod1 
                      dat2 = dat2 + hesp2*mod2 
                      dat3 = dat3 + hesp3*mod3 
                      dat4 = dat4 + hesp4*mod4 
              end if
    return
  end subroutine
!#####################################
  subroutine hess_precond_5(adj, mod1, mod2, mod3, mod4, mod5, dat1, dat2, dat3, dat4, dat5)
    logical               :: adj
    real, dimension(nmod) :: mod1, mod2, mod3, mod4, mod5
    real, dimension(nmod) :: dat1, dat2, dat3, dat4, dat5

              if (adj) then
                      mod1 = mod1 + hesp1*dat1
                      mod2 = mod2 + hesp2*dat2 
                      mod3 = mod3 + hesp3*dat3 
                      mod4 = mod4 + hesp4*dat4 
                      mod5 = mod5 + hesp5*dat5 
              else 
                      dat1 = dat1 + hesp1*mod1 
                      dat2 = dat2 + hesp2*mod2 
                      dat3 = dat3 + hesp3*mod3 
                      dat4 = dat4 + hesp4*mod4 
                      dat5 = dat5 + hesp5*mod5 
              end if
    return
  end subroutine
!#####################################
  subroutine hess_precond_6(adj, mod1, mod2, mod3, mod4, mod5, mod6, dat1, dat2, dat3, dat4, dat5, dat6)
    logical               :: adj
    real, dimension(nmod) :: mod1, mod2, mod3, mod4, mod5, mod6
    real, dimension(nmod) :: dat1, dat2, dat3, dat4, dat5, dat6
              if (adj) then
                      mod1 = mod1 + hesp1*dat1
                      mod2 = mod2 + hesp2*dat2 
                      mod3 = mod3 + hesp3*dat3 
                      mod4 = mod4 + hesp4*dat4 
                      mod5 = mod5 + hesp5*dat5 
                      mod6 = mod6 + hesp6*dat6 
              else 
                      dat1 = dat1 + hesp1*mod1 
                      dat2 = dat2 + hesp2*mod2 
                      dat3 = dat3 + hesp3*mod3 
                      dat4 = dat4 + hesp4*mod4 
                      dat5 = dat5 + hesp5*mod5 
                      dat6 = dat6 + hesp6*mod6 
              end if
    return
  end subroutine
!#####################################
end module
