!hessian compute
module hessian_mod
  use type_mod
  use global_defi_mod
  use ctaper_mod
  use util_
  use sep

  implicit none
  real   , dimension(:,:)      , pointer, private :: dgrid_header_rec_save
  real                                  , private :: dgrid_gxmin_save
  real                                  , private :: dgrid_gymin_save

  integer :: index_null_gxmin, index_null_gxmax
  real    :: null_gxmin, null_gxmax

    complex, dimension(:,:,:,:,:          ), allocatable :: greens1
    real,    dimension(:,:                ), allocatable :: hessian0
    real,    dimension(:,:,:,:,:          ), allocatable :: hessian1
    real,    dimension(:,:,:,:,:          ), allocatable :: hessian2
!   real,    dimension(:,:,:,:,:,:,:,:,:,:), allocatable :: hessian3
!   real,    dimension(:,:,:,:,:,:,:,:,:,:), allocatable :: hessian4
contains
!=======================================================================================
  integer function hess_init() result(stat)
    call greens_input_history()  !read greens history
    call param_init()             !initialize oneway parameters 
    call hess_output_history()    !output the oneway history
    call hess_array_alloc()       !allocate oneway memory space for the data
    call hess_io_init()  
    
    call from_param("null_gxmin", null_gxmin, 0.) !min. geoph.
    call from_param("null_gxmax", null_gxmax, 0.) !max. geoph.

    index_null_gxmin = 1.5 + (null_gxmin - ggrid%ax_surf%o) / ggrid%ax_surf%d
    index_null_gxmax = 1.5 + (null_gxmax - ggrid%ax_surf%o) / ggrid%ax_surf%d

    write(*,*)'========================='
    write(0,*)"num shts.", dgrid%nsh!nsh num. of shots ......ntr num. of traces
    write(0,*)"num recv.", dgrid%ntr !nsh num. of shots ......ntr num. of traces
    write(*,*)'========================='
    
    stat = 0
  end function
!=======================================================================================
  integer function hess_clos() result(stat)
    call hess_array_dealloc()
    close(10)
    stat = 0
  end function
!=======================================================================================
  integer function hess_oper() result(stat)
    real                :: sx, sy, gx, gy
    real                :: maxx, maxy, maxz
    real                :: w
    integer             :: index_aw
    integer             ::        x_modl,        y_modl,        z_modl,        hx_modl,       hy_modl,         z_offd,        y_offd,        x_offd,        hy_offd,        hx_offd
    integer             ::       ix_modl,       iy_modl,       iz_modl,       ihx_modl,      ihy_modl,        iz_offd,       iy_offd,       ix_offd,       ihy_offd,       ihx_offd
    integer             ::  index_x_modl,  index_y_modl,  index_z_modl,  index_hx_modl, index_hy_modl,   index_z_offd,  index_y_offd,  index_x_offd,  index_hy_offd,  index_hx_offd 
    integer             ::      jxd_modl,      jyd_modl,                                                                    jyd_offd,      jxd_offd 
    integer             ::      jxu_modl,      jyu_modl,                                                                    jyu_offd,      jxu_offd 
    integer             ::       jx_modl,       jy_modl,       jz_modl,       jhx_modl,      jhy_modl,        jz_offd,       jy_offd,       jx_offd,       jhy_offd,       jhx_offd
    
    integer             :: index_sx, index_sy                            !!.. relative index on Green's grid
    integer             :: igx, igy, index_gx, index_gy, isx,isy 
    complex             :: greens_rec, greens_so, cw

    write(0,*)"num. freq.:", ggrid%aw
    write(*,*)'========================='

    maxz=ggrid%az_modl%o + (ggrid%az_modl%n-1)*ggrid%az_modl%d 
    maxx=ggrid%ax_modl%o + (ggrid%ax_modl%n-1)*ggrid%ax_modl%d
    maxy=ggrid%ay_modl%o + (ggrid%ay_modl%n-1)*ggrid%ay_modl%d 
    open(unit=10, file=report_, status='unknown')
    
    hessian1=0.;   
    hessian2=0.;
    nnode =omp_get_num_procs(); thread=omp_get_thread_num(); nodes=min(nodes,nnode-1) 
    call omp_set_num_threads(nodes); write(*,*) 'using', nodes,'of', nnode
!$OMP parallel do private(iw,inode,w,index_aw,cw,iz_modl,iy_modl,ix_modl,ihy_modl,ihx_modl,iz_offd,iy_offd,ix_offd,ihy_offd,ihx_offd,z_modl,y_modl,x_modl,hy_modl,hx_modl,z_offd,y_offd,x_offd,hy_offd,hx_offd,index_z_modl,index_y_modl,index_x_modl,index_hy_modl,index_hx_modl,index_z_offd,index_y_offd,index_x_offd,index_hy_offd,index_hx_offd,jhy_modl,jhx_modl,jz_offd,jy_offd,jx_offd,jhy_offd,jhx_offd,jyd_offd,jxd_offd,jyu_offd,jxu_offd,jyd_modl,jxd_modl,isx,sx,index_sx,igx,gx,index_gx,isy,sy,index_sy,igy,gy,index_gy,greens_so,greens_rec,greens1,hessian1,hessian2) schedule(dynamic,1)   

do iw=1, ggrid%aw%n                                                            ! freq. count.
        inode=omp_get_thread_num(); 
        w=ggrid%aw%o+(iw-1)*ggrid%aw%d
        index_aw= 1.5+(w-dgrid%aw%o)/dgrid%aw%d
        if (.not. source) csou(index_aw)=1
        cw=cabs(csou(index_aw)**2.)
        write(* ,*)'start iw: ', index_aw, 'of ',ggrid%aw%n ,'@ inode: ', inode;
        if (.not. allocated(greens1 ))then; allocate(greens1 (ggrid%ax_surf%n , ggrid%ay_surf%n,ggrid%ax_modl%n, ggrid%ay_modl%n,ggrid%az_modl%n)); greens1=0.;endif
        if (.not. allocated(hessian1))then; allocate(hessian1(hgrid%ahx_offd%n,hgrid%ahy_offd%n,hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n)); hessian1=0.;endif
        if (.not. allocated(hessian2))then; allocate(hessian2(hgrid%ahx_offd%n,hgrid%ahy_offd%n,hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n)); hessian2=0.;endif
        !$OMP CRITICAL 
          call hess_io_greens(greens1, index_aw)
        !$OMP END CRITICAL
do iz_modl= 1,hgrid%az_modl%n                                                !  z loop 
        z_modl = hgrid%az_modl%o + (iz_modl-1)*hgrid%az_modl%d; index_z_modl = 1.5 + (z_modl - ggrid%az_modl%o)/ggrid%az_modl%d;
do iy_modl= 1,hgrid%ay_modl%n                                                !  y loop
        y_modl = hgrid%ay_modl%o + (iy_modl-1)*hgrid%ay_modl%d; index_y_modl = 1.5 + (y_modl - ggrid%ay_modl%o)/ggrid%ay_modl%d;
do ix_modl= 1,hgrid%ax_modl%n                                                !  x loop
        x_modl = hgrid%ax_modl%o + (ix_modl-1)*hgrid%ax_modl%d; index_x_modl = 1.5 + (x_modl - ggrid%ax_modl%o)/ggrid%ax_modl%d;
        do ihy_modl= 1,hgrid%ahy_modl%n                                      ! hy loop                                
                 hy_modl = hgrid%ahy_modl%o + (ihy_modl-1)*hgrid%ahy_modl%d; jhy_modl = nint(hy_modl / ggrid%ay_modl%d)
                    jyd_modl = index_y_modl - jhy_modl;               jyu_modl = index_y_modl + jhy_modl
                if (jyd_modl < 1 .or. jyd_modl > ggrid%ay_modl%n .or. jyu_modl < 1 .or. jyu_modl > ggrid%ay_modl%n) cycle
        do ihx_modl= 1,hgrid%ahx_modl%n                                      ! hx loop          
                 hx_modl = hgrid%ahx_modl%o + (ihx_modl-1)*hgrid%ahx_modl%d; jhx_modl = nint(hx_modl / ggrid%ax_modl%d)
                    jxd_modl = index_x_modl - jhx_modl;               jxu_modl = index_x_modl + jhx_modl
                if (jxd_modl < 1 .or. jxd_modl > ggrid%ax_modl%n .or. jxu_modl < 1 .or. jxu_modl > ggrid%ax_modl%n) cycle
                hessian2=0.
                do iz_offd=1, hgrid%az_offd%n                                ! zo loop            
                         z_offd = hgrid%az_offd%o + (iz_offd-1)*hgrid%az_offd%d; jz_offd = nint(z_offd / ggrid%az_modl%d)      
                        index_z_offd = index_z_modl + jz_offd        
                        if (index_z_offd < 1 .or. index_z_offd > ggrid%az_modl%n) cycle   
                do iy_offd=1, hgrid%ay_offd%n                                ! yo loop  
                        y_offd = hgrid%ay_offd%o + (iy_offd-1)*hgrid%ay_offd%d; jy_offd = nint(y_offd / ggrid%ay_modl%d)
                do ix_offd=1, hgrid%ax_offd%n                                ! xo loop 
                        x_offd = hgrid%ax_offd%o + (ix_offd-1)*hgrid%ax_offd%d; jx_offd = nint(x_offd / ggrid%ax_modl%d)
                        do ihy_offd=1, hgrid%ahy_offd%n                      !hyo loop     
                                hy_offd  = hgrid%ahy_offd%o + (ihy_offd-1)*hgrid%ahy_offd%d; jhy_offd  = nint(hy_offd / ggrid%ay_modl%d)
                                jyd_offd = jyd_modl + jy_offd - jhy_offd; jyu_offd = jyu_modl + jy_offd + jhy_offd     
                                if (jyd_offd < 1 .or. jyd_offd > ggrid%ay_modl%n .or. jyu_offd < 1 .or. jyu_offd > ggrid%ay_modl%n) cycle
                        do ihx_offd=1, hgrid%ahx_offd%n                      !hxo loop  
                                hx_offd  = hgrid%ahx_offd%o + (ihx_offd-1)*hgrid%ahx_offd%d; jhx_offd  = nint(hx_offd / ggrid%ax_modl%d)
                                jxd_offd = jxd_modl + jx_offd - jhx_offd; jxu_offd = jxu_modl + jx_offd + jhx_offd     
                                if (jxd_offd < 1 .or. jxd_offd > ggrid%ax_modl%n .or. jxu_offd < 1 .or. jxu_offd > ggrid%ax_modl%n) cycle

                                greens_rec=0.
                                do igy= 1,dgrid%agy%n
                                        gy = dgrid%agy%o + (igy-1)*dgrid%agy%d; index_gy = 1.5 + (gy - ggrid%ay_surf%o)/ggrid%ay_surf%d;
                                do igx= 1,dgrid%agx%n
                                        gx = dgrid%agx%o + (igx-1)*dgrid%agx%d; index_gx = 1.5 + (gx - ggrid%ax_surf%o)/ggrid%ax_surf%d;
                                        greens_rec = greens_rec+ greens1(index_gx, index_gy, jxu_modl, jyu_modl,index_z_modl)*conjg(greens1(index_gx, index_gy, jxu_offd, jyu_offd, index_z_offd))
                                        
                                enddo 
                                enddo !recv
                                greens_so=0.
                                do isy= 1,dgrid%asy%n
                                        sy = dgrid%asy%o + (isy-1)*dgrid%asy%d; index_sy = 1.5 + (sy - ggrid%ay_surf%o)/ggrid%ay_surf%d;
                                do isx= 1,dgrid%asx%n
                                        sx = dgrid%asx%o + (isx-1)*dgrid%asx%d; index_sx = 1.5 + (sx - ggrid%ax_surf%o)/ggrid%ax_surf%d;
                                        greens_so = greens_so+  greens1(index_sx, index_sy, jxd_modl, jyd_modl,index_z_modl)*conjg(greens1(index_sx, index_sy, jxd_offd, jyd_offd, index_z_offd))
                                enddo
                                enddo  !shot 
                                hessian1(ihx_offd,ihy_offd,ix_offd,iy_offd,iz_offd)=real(greens_so*greens_rec)
                        enddo   !hxoffd
                        enddo   !hyoffd
                enddo           !xoffd
                enddo           !yoffd
                enddo           !zoffd
                if (nostack)then
                        !$OMP CRITICAL 
                         hessian2=         hessian1*cw; call hess_io_hess(hessian2,iw,ihx_modl,ihy_modl, ix_modl,iy_modl,iz_modl, 'rite'); 
                        !$OMP END CRITICAL
                else
                        !$OMP CRITICAL 
                        !write(0,*)sum(hessian2)
                         hessian2=hessian2+hessian1*cw; call hess_io_hess(hessian2,1 ,ihx_modl,ihy_modl, ix_modl,iy_modl,iz_modl, 'rite'); 
                        !$OMP END CRITICAL
                endif
        enddo                   !hxmodl
        enddo                   !hymodl
enddo                           !xmodl
enddo                           !ymodl
enddo                           !zmodl
write(* ,*)' done iw: ', index_aw, 'of ',ggrid%aw%n ,'@ inode: ', inode;
enddo
!$OMP end parallel do 
stat = 0
end function

!=========================================================================================
  subroutine hess_array_alloc()
    integer :: memory
    allocate(hessian0(hgrid%ahx_offd%n,hgrid%ahy_offd%n                                                ));hessian0=0.
    allocate(hessian1(hgrid%ahx_offd%n,hgrid%ahy_offd%n,hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n));hessian1=0.
    allocate(hessian2(hgrid%ahx_offd%n,hgrid%ahy_offd%n,hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n));hessian2=0.
    allocate( greens1( ggrid%ax_surf%n, ggrid%ay_surf%n,ggrid%ax_modl%n,ggrid%ay_modl%n,ggrid%az_modl%n));greens1=0.
    allocate(wavelet(1))

    memory = 4 * (size(hessian0) + size(hessian1) +  size(hessian2)) + 8 * (size(greens1) + size(wavelet))  
    memory = memory / 1024 / 1024
    write(0,*)"Memory allocated: ", memory, "Mb"

  end subroutine
!=================================================================================================
subroutine hess_array_dealloc()
  if( allocated(greens1 )) deallocate(greens1 )
  if( allocated(hessian0)) deallocate(hessian0)
  if( allocated(hessian1)) deallocate(hessian1)
  if( allocated(hessian2)) deallocate(hessian2)
  if( allocated(wavelet )) deallocate(wavelet )
  if( allocated(greens  )) deallocate(greens  )
  end subroutine
!==========================================================================================
  subroutine param_init()
    real    :: xmin_modl_default, xmax_modl_default
    real    :: ymin_modl_default, ymax_modl_default
    real    :: zmin_modl_default, zmax_modl_default
    real    :: surface_xmin_default, surface_xmax_default
    real    :: surface_ymin_default, surface_ymax_default
    integer :: ixn, iyn, izn

    !============================================================
    call from_param("nsx", dgrid%asx%n, ggrid%ax_surf%n); call from_param("dsx", dgrid%asx%d,  ggrid%ax_surf%d); call from_param("osx", dgrid%asx%o,  ggrid%ax_surf%o) 
    call from_param("nsy", dgrid%asy%n, ggrid%ay_surf%n); call from_param("dsy", dgrid%asy%d,  ggrid%ay_surf%d); call from_param("osy", dgrid%asy%o,  ggrid%ay_surf%o) 
    dgrid%nsh = dgrid%asx%n * dgrid%asy%n     !num  of sources..
    call from_param("ngx", dgrid%agx%n, ggrid%ax_surf%n); call from_param("dgx", dgrid%agx%d,  ggrid%ax_surf%d); call from_param("ogx", dgrid%agx%o,  ggrid%ax_surf%o)
    call from_param("ngy", dgrid%agy%n, ggrid%ay_surf%n); call from_param("dgy", dgrid%agy%d,  ggrid%ay_surf%d); call from_param("ogy", dgrid%agy%o,  ggrid%ay_surf%o)
    dgrid%ntr = dgrid%agx%n * dgrid%agy%n   !num. of traces/recievers
    write(*,*)'sx:n_d_o',dgrid%asx
    write(*,*)'sy:n_d_o',dgrid%asy
    write(*,*)'gx:n_d_o',dgrid%agx
    write(*,*)'gy:n_d_o',dgrid%agy 
   !============================================================
    !sub-offset off-diags
    call from_param("nhx_offd", hgrid%ahx_offd%n, 1 ); call from_param("dhx_offd", hgrid%ahx_offd%d, ggrid%ax_modl%d); call from_param("ohx_offd", hgrid%ahx_offd%o, 0.)
    call from_param("nhy_offd", hgrid%ahy_offd%n, 1 ); call from_param("dhy_offd", hgrid%ahy_offd%d, ggrid%ay_modl%d); call from_param("ohy_offd", hgrid%ahy_offd%o, 0.)
    !off-diags
    call from_param(" nx_offd",  hgrid%ax_offd%n, 1 ); call from_param(" dx_offd",  hgrid%ax_offd%d, ggrid%ax_modl%d); call from_param(" ox_offd",  hgrid%ax_offd%o, 0.)               
    call from_param(" ny_offd",  hgrid%ay_offd%n, 1 ); call from_param(" dy_offd",  hgrid%ay_offd%d, ggrid%ay_modl%d); call from_param(" oy_offd",  hgrid%ay_offd%o, 0.)
    call from_param(" nz_offd",  hgrid%az_offd%n, 1 ); call from_param(" dz_offd",  hgrid%az_offd%d, ggrid%az_modl%d); call from_param(" oz_offd",  hgrid%az_offd%o, 0.)
    !sub-offset
    call from_param("nhx_modl", hgrid%ahx_modl%n, 1 ); call from_param("dhx_modl", hgrid%ahx_modl%d, ggrid%ax_modl%d); call from_param("ohx_modl", hgrid%ahx_modl%o, 0.)
    call from_param("nhy_modl", hgrid%ahy_modl%n, 1 ); call from_param("dhy_modl", hgrid%ahy_modl%d, ggrid%ay_modl%d); call from_param("ohy_modl", hgrid%ahy_modl%o, 0.)
    !============================================================
    !main-model
    xmin_modl_default = ggrid%ax_modl%o; xmax_modl_default = ggrid%ax_modl%o + (ggrid%ax_modl%n-1)*ggrid%ax_modl%d
    ymin_modl_default = ggrid%ay_modl%o; ymax_modl_default = ggrid%ay_modl%o + (ggrid%ay_modl%n-1)*ggrid%ay_modl%d
    zmin_modl_default = ggrid%az_modl%o; zmax_modl_default = ggrid%az_modl%o + (ggrid%az_modl%n-1)*ggrid%az_modl%d

    call from_param("xmin_modl", hgrid%xmin_modl, xmin_modl_default);call from_param("ymin_modl", hgrid%ymin_modl, ymin_modl_default);
    call from_param("xmax_modl", hgrid%xmax_modl, xmax_modl_default);call from_param("ymax_modl", hgrid%ymax_modl, ymax_modl_default);
    call from_param("zmin_modl", hgrid%zmin_modl, zmin_modl_default);call from_param("zmax_modl", hgrid%zmax_modl, zmax_modl_default)
    ixn =   1.5 + (hgrid%xmax_modl-hgrid%xmin_modl)/ggrid%ax_modl%d
    iyn =   1.5 + (hgrid%ymax_modl-hgrid%ymin_modl)/ggrid%ay_modl%d
    izn =   1.5 + (hgrid%zmax_modl-hgrid%zmin_modl)/ggrid%az_modl%d

    call from_param(" nx_modl",  hgrid%ax_modl%n, ixn); call from_param(" dx_modl",  hgrid%ax_modl%d, ggrid%ax_modl%d); call from_param(" ox_modl",  hgrid%ax_modl%o, hgrid%xmin_modl)
    call from_param(" ny_modl",  hgrid%ay_modl%n, iyn); call from_param(" dy_modl",  hgrid%ay_modl%d, ggrid%ay_modl%d); call from_param(" oy_modl",  hgrid%ay_modl%o, hgrid%ymin_modl)
    call from_param(" nz_modl",  hgrid%az_modl%n, izn); call from_param(" dz_modl",  hgrid%az_modl%d, ggrid%az_modl%d); call from_param(" oz_modl",  hgrid%az_modl%o, hgrid%zmin_modl)
      !============================================================
    write(*,*)'========================='
    write(*,*)'Summary Hessian    ======'
    write(*,*)'========================='
    write(*,*)'ahx_offd:n_d_o',hgrid%ahx_offd; write(*,*)'ahy_offd:n_d_o',hgrid%ahy_offd; write(*,*)' ax_offd:n_d_o',hgrid%ax_offd; write(*,*)' ay_offd:n_d_o',hgrid%ay_offd; write(*,*)' az_offd:n_d_o',hgrid%az_offd
    write(*,*)'========================='
    write(*,*)'ahx_modl:n_d_o',hgrid%ahx_modl; write(*,*)'ahy_modl:n_d_o',hgrid%ahy_modl; write(*,*)' ax_modl:n_d_o',hgrid%ax_modl; write(*,*)' ay_modl:n_d_o',hgrid%ay_modl; write(*,*)' az_modl:n_d_o',hgrid%az_modl
    write(*,*)'========================='
    !============================================================
    report_ = "report"    ; 
    call from_param("nostack"     , nostack      , .false.)
    call from_param("taper"       , taper        , .true. )
    call from_param("taper_pct"   , taper_pct    , 0.1    )
    call from_param("taper_type"  , taper_type   , 2      )
    call from_param("nodes     "  , nodes        , 1      )
  end subroutine
!========================================================================================== 
  subroutine greens_input_history()
  real :: w
  integer :: iiw
  complex, allocatable:: tmp(:)
    greens_= "greens"; 
    wavelet_="wavelet"
    if (exist_file(wavelet_))source=.true.
    call from_param("jw"  , jw  , 1 ); call from_param("jjw" , jjw , 1 ); 
    call auxin(greens_)
    call hinaxis2(greens_, ggrid%ax_surf, 1); 
    call hinaxis2(greens_, ggrid%ay_surf, 2); 
    call hinaxis2(greens_, ggrid%ax_modl, 3); 
    call hinaxis2(greens_, ggrid%ay_modl, 4); 
    call hinaxis2(greens_, ggrid%az_modl, 5); 
    call hinaxis2(greens_, ggrid%aw , 6);
    ggrid%aw%n= nint(ggrid%aw%n/jw*1.); ggrid%aw%d=     (ggrid%aw%d*jw*1.) 

    if (source) then; call hinaxis2(wavelet_, dgrid%aw, 3); write(6,*)'wavelet....'; else; dgrid%aw = ggrid%aw; end if
    call data_radians_calc    
    allocate(csou(ggrid%aw%n)); allocate(tmp(dgrid%aw%n))
    if (source)then
        if (dgrid%aw%n /= ggrid%aw%n .or. dgrid%aw%d /= ggrid%aw%d .or. dgrid%aw%o /= ggrid%aw%o) then
                call sep_read(tmp, wavelet_, esize=8); 
                do iw=1,ggrid%aw%n; 
                        w   = ggrid%aw%o + (iw-1)*ggrid%aw%d;
                        iiw = 1.5 + (w-dgrid%aw%o)/dgrid%aw%d; 
                        csou(iw)=tmp(iiw); 
                enddo; 
                dgrid%aw=ggrid%aw;
        else
                call sep_read(tmp, wavelet_, esize=8); csou(:)=tmp(:)
        endif
    else
        dgrid%aw=ggrid%aw; csou(:)=1
    endif
    write(*,*)'========================='
    write(*,*)'Summary Greens     ======'
    write(*,*)'========================='
    write(*,*)'x_surf:n_d_o  :',ggrid%ax_surf
    write(*,*)'y_surf:n_d_o  :',ggrid%ay_surf
    write(*,*)'x_modl:n_d_o  :',ggrid%ax_modl
    write(*,*)'y_modl:n_d_o  :',ggrid%ay_modl
    write(*,*)'z_modl:n_d_o  :',ggrid%az_modl
    write(*,*)'dgrid_aw:n_d_o:',dgrid%aw
    write(*,*)'ggrid_aw:n_d_o:',ggrid%aw
    write(*,*)'jw            :',jw
    write(*,*)'========================='
  end subroutine
!==========================================================================================
  subroutine hess_output_history()
    !!!!!Outputs the Hessain details into the history file...
    hessian_= "hessian"; 
    call auxinout(hessian_)
    call houaxis2(hessian_, hgrid%ahx_offd, 1); call houaxis2(hessian_, hgrid%ahy_offd, 2); call houaxis2(hessian_, hgrid%ax_offd , 3); call houaxis2(hessian_, hgrid%ay_offd , 4); call houaxis2(hessian_, hgrid%az_offd , 5); 
    call houaxis2(hessian_, hgrid%ahx_modl, 6); call houaxis2(hessian_, hgrid%ahy_modl, 7); call houaxis2(hessian_, hgrid%ax_modl , 8); call houaxis2(hessian_, hgrid%ay_modl , 9); 
    call auxputch("n10", "i", hgrid%az_modl%n, hessian_); call auxputch("d10", "r", hgrid%az_modl%d, hessian_); call auxputch("o10", "r", hgrid%az_modl%o, hessian_)
    !labels ==========================
    call auxputch("label1","s","hx_offd", hessian_); call auxputch("label2","s","hy_offd", hessian_); call auxputch("label3","s"," x_offd", hessian_); call auxputch("label4","s"," y_offd", hessian_); call auxputch("label5 ","s"," z_offd", hessian_)
    call auxputch("label6","s","hx_modl", hessian_); call auxputch("label7","s","hy_modl", hessian_); call auxputch("label8","s"," x_modl", hessian_); call auxputch("label9","s"," y_modl", hessian_); call auxputch("label10","s"," z_modl", hessian_)
    call auxputch("d11", "r", ggrid%aw%d, hessian_); call auxputch("o11", "r", ggrid%aw%o, hessian_); call auxputch("label11","s","w     ", hessian_)
    if (nostack)then 
        call auxputch("n11", "i", ggrid%aw%n,hessian_);  else; call auxputch("n11", "i", 1         ,hessian_)
    endif
    !===================================================
  end subroutine

!====================================================================================================================
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!initialize the full hessian to zeros.
 subroutine hess_io_init()
    real, allocatable, dimension(:) :: ctmp
    integer                         :: i, nloop
    allocate(ctmp(hgrid%ahx_modl%n*hgrid%ahy_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    ctmp = 0.
          nloop = hgrid%ahx_offd%n*hgrid%ahy_offd%n*hgrid%ax_offd%n*hgrid%ay_offd%n*hgrid%az_offd%n
          write(*,*)'Initializing Hessian.....'
        if (nostack)then
                do i=1, nloop*ggrid%aw%n; call sep_write(ctmp, hessian_, esize=4); end do
        else
                do i=1, nloop           ; call sep_write(ctmp, hessian_, esize=4); end do
        endif
    deallocate(ctmp)
  end subroutine
!====================================================================================================================
  subroutine hess_io_greens(greens, index_ws)
    integer                                   :: index_ws
    integer, dimension(6)                     :: n_cube, j_wind, n_wind, f_wind
    integer                                   :: n_dim_cube, err
    complex, dimension(:,:,:,:,:), allocatable:: greens

    n_dim_cube = 6
    n_cube = (/ggrid%az_modl%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%ax_surf%n, ggrid%ay_surf%n, nint(.5+(ggrid%aw%n*jw))/)
    j_wind = (/1              , 1              , 1              , 1              , 1              , 1                       /)
    n_wind = (/ggrid%az_modl%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%ax_surf%n, ggrid%ay_surf%n, 1                       /)
    f_wind = (/0              , 0              , 0              , 0              , 0              , index_ws-1              /)
    !write(*,*) n_cube; write(*,*) n_wind; write(*,*) f_wind; write(*,*) j_wind
    greens = 0.
    !write(0,*)'reading greens frequency: ', index_ws, 'of ', int(.5+(ggrid%aw%n*jw))
    err = sreed_window(greens_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,greens)
    if (err /= 0) call seperr("data_io_greens: error readin greens")

  end subroutine
!===================================================
  subroutine hess_io_wavelet(wavelet, index_ws)
    complex, dimension(:), pointer :: wavelet
    integer                        :: index_ws
    integer, dimension(1)          :: n_cube, j_wind, n_wind, f_wind
    integer                        :: n_dim_cube, err

    n_dim_cube = 1; n_cube = dgrid%aw%n; j_wind = 1; n_wind = 1; f_wind = index_ws-1
    !write(*,*) n_cube; write(*,*) n_wind; write(*,*) f_wind; write(*,*) j_wind
    wavelet = 0.
    err = sreed_window(wavelet_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,wavelet)
    if (err .lt. 0) call seperr("io_wavelet: trouble reading source wavelet")
  end subroutine

!#####################################################
 subroutine                    hess_io_hess(hessian,index_aw,index_hx_modl,index_hy_modl,index_x_modl,index_y_modl,index_z_modl, flag); 
    integer                                     ::  index_aw,index_hx_modl,index_hy_modl,index_x_modl,index_y_modl,index_z_modl 
    integer, dimension(11)                      ::  n_cube, j_wind, n_wind, f_wind
    integer                                     ::  n_dim_cube, err
    real, dimension(:,:,:,:,:),      allocatable::  hessian
    character(len=4)                            ::  flag

    n_dim_cube = 11
    n_cube = (/hgrid%ahx_offd%n, hgrid%ahy_offd%n, hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, hgrid%ahx_modl%n, hgrid%ahy_modl%n, hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n, ggrid%aw%n/)
    j_wind = (/1               , 1               , 1              , 1              , 1              , 1               , 1               , 1              , 1              , 1              , 1         /)
    n_wind = (/hgrid%ahx_offd%n, hgrid%ahy_offd%n, hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, 1               , 1               , 1              , 1              , 1              , 1         /)
    f_wind = (/0               , 0               , 0              , 0              , 0              , index_hx_modl-1 , index_hy_modl-1 , index_x_modl-1 , index_y_modl-1 , index_z_modl-1 , index_aw-1/)
   ! write(*,*) n_cube; write(*,*) n_wind; write(*,*) f_wind; write(*,*) j_wind,'--'

         if (flag == "rite") then
                !write(0,*)'writing hessian freq: ', index_aw, 'of ', ggrid%aw%n; 
                err = srite_window(hessian_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hessian)
    else if (flag == "reed") then
                !write(0,*)'reading hessian freq: ', index_aw, 'of ', ggrid%aw%n; 
                err = sreed_window(hessian_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hessian)
    else
       call seperr("you have to specify the flag for data_io_data()")
    end if
    if (err /= 0) call seperr("hess_io_hess: error writing hessian")
  end subroutine

!######################################################################
  subroutine data_radians_calc()
        if (.not.radian) then; dgrid%aw%d = dgrid%aw%d * 2.0 * PI; dgrid%aw%o = dgrid%aw%o * 2.0 * PI; end if
  end subroutine

  end module

