module hessian_mod_conv_joint
  use adj_mod
  use type_mod
  use util_
!  use igrad1
!  use igrad2
  use sep
  use minresDataModule

  implicit none
  integer           , save                               , private :: index_rsv_zmin, index_rsv_zmax
  real              , save                               , private ::       rsv_xmin,       rsv_xmax
  integer           , save                               , private :: index_rsv_xmin, index_rsv_xmax
  integer           , save                               , private :: index_zmin    , index_zmax    
  real              , save                               , private ::       rsv_zmin,       rsv_zmax
  real              , save                               , private ::           zmin,           zmax
  real              , save                               , private :: eps_base, eps_4d
  type(hes_grid)    , save                                         :: hgrid 
  real              , dimension(:,:,:)       ,allocatable, private :: hess1
  real              , dimension(:,:,:)       ,allocatable, private :: hess2
  real              , dimension(:,:,:)       ,allocatable, private,save :: hess_1
  real              , dimension(:,:,:)       ,allocatable, private,save :: hess_2
  real              , dimension(:,:,:)       ,allocatable, private,save :: weight1
  real              , dimension(:,:,:)       ,allocatable, private,save :: weight2
  real              , dimension(:,:,:)       ,allocatable, private,save :: mask1
  real              , dimension(:,:,:)       ,allocatable, private,save :: mask2
  real              , dimension(:,:,:)       ,allocatable, private,save :: epsi1
  real              , dimension(:,:,:)       ,allocatable, private,save :: epsi2
  character(len=256),                                      private :: file1_
  character(len=256),                                      private :: file2_
  character(len=256),                                      private :: epsi1_
  character(len=256),                                      private :: epsi2_
  character(len=256),                                      private :: weight1_
  character(len=256),                                      private :: weight2_
  character(len=256),                                      private :: mask1_
  character(len=256),                                      private :: mask2_
  integer           ,                                     external :: srite_window, sreed_window, getch
  logical           , save                                         :: ones
  integer           , save                                ,private :: niter
  real              , save                                ,private :: iter,riter

contains
!#######
  integer function conv_init(hgrid_in, file1_in, file2_in, ones_in, & 
                                zmin_in, rsv_zmin_in, zmax_in, rsv_zmax_in, rsv_xmin_in, rsv_xmax_in, &
                                eps_base_in, eps_4d_in, epsi1_in, epsi2_in, niter_in) result(stat)
    optional:: zmin_in , rsv_zmin_in, zmax_in , rsv_zmax_in, rsv_xmin_in, rsv_xmax_in, eps_base_in, eps_4d_in, epsi1_in, epsi2_in, niter_in
    type(hes_grid)        :: hgrid_in
    character(len=256)    :: file1_in
    character(len=256)    :: file2_in
    character(len=256)    :: epsi1_in
    character(len=256)    :: epsi2_in
    logical               :: ones_in
    real                  :: rsv_xmin_in,rsv_xmax_in
    real                  :: rsv_zmin_in,rsv_zmax_in
    real                  ::     zmin_in,    zmax_in
    real                  :: eps_base_in 
    real                  :: eps_4d_in
    real                  :: max_z
    real                  :: max_x
    integer               :: niter_in
    max_z=hgrid%az_modl%o+(hgrid%az_modl%n-1)*hgrid%az_modl%d
    max_x=hgrid%ax_modl%o+(hgrid%ax_modl%n-1)*hgrid%ax_modl%d
    niter=1 ; iter=1; riter=0.

    stat = 1
    hgrid = hgrid_in
    file1_ = file1_in
    file2_ = file2_in
    ones = ones_in

    if(present(    zmin_in))then;     zmin=    max(zmin_in,hgrid%az_modl%o); else;     zmin=hgrid%az_modl%o; endif
    if(present(rsv_zmin_in))then; rsv_zmin=max(rsv_zmin_in,hgrid%az_modl%o); else; rsv_zmin=hgrid%az_modl%o; endif
    if(present(    zmax_in))then;     zmax=    min(zmax_in,max_z); else;     zmax=max_z; endif
    if(present(rsv_zmax_in))then; rsv_zmax=min(rsv_zmax_in,max_z); else; rsv_zmax=max_z; endif
    if(present(rsv_xmin_in))then; rsv_xmin=max(rsv_xmin_in,hgrid%ax_modl%o); else; rsv_xmin=hgrid%ax_modl%o; endif
    if(present(rsv_xmax_in))then; rsv_xmax=min(rsv_xmax_in,max_x); else; rsv_xmax=max_x; endif


    eps_base =1.; eps_4d=1.    ; niter=1
    if(present(eps_base_in))  eps_base= eps_base_in
    if(present(eps_4d_in  ))  eps_4d  = eps_4d_in
    if(present(epsi1_in  ))  epsi1_  = epsi1_in
    if(present(epsi2_in  ))  epsi2_  = epsi2_in
    if(present(niter_in  ))  niter = niter_in

    allocate(hess1 (hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n))
    allocate(hess2 (hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n))
    allocate(hess_1 (hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    allocate(hess_2 (hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    hess1 = 0.
    hess2= 0.
    stat = 0.
  end function
!  ##########################
  integer function conv2_init(hgrid_in, file1_in, file2_in, ones_in, & 
                                zmin_in, rsv_zmin_in, zmax_in, rsv_zmax_in, rsv_xmin_in, rsv_xmax_in,& 
                                eps_base_in, eps_4d_in, niter_in, mask1_in, mask2_in, weight1_in, weight2_in, epsi1_in, epsi2_in) result(stat)
    optional:: zmin_in, rsv_zmin_in, zmax_in, rsv_zmax_in, rsv_xmin_in, rsv_xmax_in, & 
                weight1_in, weight2_in, mask1_in, mask2_in, eps_base_in, eps_4d_in, epsi1_in, epsi2_in, niter_in
    type(hes_grid)        :: hgrid_in
    character(len=256)    :: file1_in
    character(len=256)    :: file2_in
    character(len=256)    :: weight1_in
    character(len=256)    :: weight2_in
    character(len=256)    :: mask1_in
    character(len=256)    :: mask2_in
    character(len=256)    :: epsi1_in
    character(len=256)    :: epsi2_in
    logical               :: ones_in
    real                  :: rsv_xmin_in,rsv_xmax_in
    real                  :: rsv_zmin_in,rsv_zmax_in
    real                  ::     zmin_in,    zmax_in
    real                  :: eps_base_in 
    real                  :: eps_4d_in
    real                  :: max_z
    real                  :: max_x
    integer               :: niter_in
    max_z=hgrid%az_modl%o+(hgrid%az_modl%n-1)*hgrid%az_modl%d
    max_x=hgrid%ax_modl%o+(hgrid%ax_modl%n-1)*hgrid%ax_modl%d
    niter=1 ; iter=1; riter=0.

    stat = 1
    hgrid = hgrid_in
    file1_ = file1_in
    file2_ = file2_in
    if(present(weight1_in))weight1_  = weight1_in
    if(present(weight2_in))weight2_  = weight2_in
    if(present(mask1_in  ))  mask1_  = mask1_in
    if(present(mask2_in  ))  mask2_  = mask2_in
    if(present(epsi1_in  ))  epsi1_  = epsi1_in
    if(present(epsi2_in  ))  epsi2_  = epsi2_in

    eps_base =1.; eps_4d=1.   ; 
    if(present(eps_base_in  ))  eps_base= eps_base_in
    if(present(eps_4d_in    ))  eps_4d  = eps_4d_in
    if(present(niter_in  ))  niter = niter_in
 
    ones = ones_in
    if(present(    zmin_in))then;     zmin=    max(zmin_in,hgrid%az_modl%o); else;     zmin=hgrid%az_modl%o; endif
    if(present(rsv_zmin_in))then; rsv_zmin=max(rsv_zmin_in,hgrid%az_modl%o); else; rsv_zmin=hgrid%az_modl%o; endif
    if(present(    zmax_in))then;     zmax=    min(zmax_in,max_z); else;     zmax=max_z; endif
    if(present(rsv_zmax_in))then; rsv_zmax=min(rsv_zmax_in,max_z); else; rsv_zmax=max_z; endif
    if(present(rsv_xmin_in))then; rsv_xmin=max(rsv_xmin_in,hgrid%ax_modl%o); else; rsv_xmin=hgrid%ax_modl%o; endif
    if(present(rsv_xmax_in))then; rsv_xmax=min(rsv_xmax_in,max_x); else; rsv_xmax=max_x; endif
        write(0,*)'-------------------------------------------rsv_xmin',rsv_xmin
        write(0,*)'-------------------------------------------rsv_xmax',rsv_xmax
        write(0,*)'-------------------------------------------rsv_zmin',rsv_zmin
        write(0,*)'-------------------------------------------rsv_zmax',rsv_zmax
        write(0,*)'-----------------------------------------------zmin',zmin
        write(0,*)'-----------------------------------------------zmax',zmax

    allocate(hess1 (hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n))
    allocate(hess2 (hgrid%ax_offd%n,hgrid%ay_offd%n,hgrid%az_offd%n))
    allocate(hess_1(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    allocate(hess_2(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    if(present(weight1_in))allocate(weight1(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n))
    if(present(weight2_in))allocate(weight2(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n))
    if(present(mask1_in  ))allocate(  mask1(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n))
    if(present(mask2_in  ))allocate(  mask2(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n))
    !if(present(epsi1_in  ))allocate(  epsi1(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    !if(present(epsi2_in  ))allocate(  epsi2(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    allocate(  epsi1(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    allocate(  epsi2(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n))
    !initialize precond
    call hess_io_diag(hess_1, hgrid, file1_, "reed")
    call hess_io_diag(hess_2, hgrid, file2_, "reed")


    !initialize weight
    if(present(weight1_in))then; 
        if (exist_file(weight1_in))then
                write(0,*)'Weighting1 init.......'
                call sreed(weight1_in   , weight1  , 4*hgrid%az_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n);
        else
                write(0,*)'No weighting1.......'
        endif
    else
        write(0,*)'No weighting1.......'
    endif
    if(present(weight2_in))then;
        if(exist_file(weight2_in))then
                write(0,*)'Weighting2 init.......'
                call sreed(weight2_in   , weight2  , 4*hgrid%az_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n)
        else
                write(0,*)'No weighting2.......'
        endif
    else
        write(0,*)'No weighting2.......'
    endif
    !initialize mask
    if(present(mask1_in))then; 
        if(exist_file(mask1_in   )) then; 
            call sreed(mask1_in     , mask1    , 4*hgrid%az_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n)
        else
                write(0,*)'No Masking1.......'
        endif
    else
        write(0,*)'No Masking1........'
    endif
    if(present(mask2_in))then; 
        if(exist_file(mask2_in   )) then; 
            call sreed(mask2_in     , mask2    , 4*hgrid%az_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n)
        else
                write(0,*)'No Masking2.......'
        endif
    else
        write(0,*)'No Masking2........'
    endif
    !initialize epsi
    if(present(epsi1_in))then; 
        if(exist_file(epsi1_in   )) then; 
                write(0,*)'Epsi-ing1 init.......'
            call sreed(epsi1_in     , epsi1    , 4*hgrid%az_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n)
        else
                write(0,*)'Epsi-hess1-init1.......'
            epsi1=(hess_1+0.00001)
            epsi1=epsi1/maxval(epsi1)
        endif
    else
        write(0,*)'No Epsi-ing1........'
        epsi1=(hess_1+0.00001)
        epsi1=epsi1/maxval(epsi1)
    endif
    if(present(epsi2_in))then; 
        if(exist_file(epsi2_in   )) then; 
                write(0,*)'Epsi-ing2init.......'
            call sreed(epsi2_in     , epsi2    , 4*hgrid%az_modl%n*hgrid%ax_modl%n*hgrid%ay_modl%n)
        else
                write(0,*)'Epsi-hess2-init.......'
            epsi2=(hess_2+0.00001)
            epsi2=epsi2/maxval(epsi2)
        endif
    else
        epsi2=(hess_2+0.00001)
        epsi2=epsi2/maxval(epsi2)
        write(0,*)'No Epsi-ing2........'
    endif
    !epsi1=1.
    !epsi2=1.
    stat = 0.
    !STOP
  end function
!#######
  integer function conv_clos() result(stat)
  if(allocated(hess1))deallocate(hess1)
  if(allocated(hess2))deallocate(hess2)
!  if(allocated(hess_1))deallocate(hess_1)
!  if(allocated(hess_2))deallocate(hess_2)
    stat = 0
  end function
!#######
  integer function conv_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:):: modl, data
    real, dimension(:),allocatable :: modl1, data1
    real, dimension(:),allocatable :: modl2, data2
    allocate(modl1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(modl2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    allocate(data1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(data2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    stat = 1
    if (adj) then;   modl1 = 0.; modl2=0; data1=data(1:size(data1)); data2=data(size(data1)+1:size(data))
    else         ;   data1 = 0.; data2=0; modl1=modl(1:size(modl1)); modl2=modl(size(modl1)+1:size(modl))
    end if
    call hess_convolve(adj, modl1, modl2, data1, data2)
    if (.not. add) then
        if (adj)then;  modl(1:size(modl1))=modl1; modl(size(modl1)+1:size(modl))=modl2;
        else        ;  data(1:size(data1))=data1; data(size(data1)+1:size(data))=data2;
        endif    
    else
        if (adj)then;  modl(1:size(modl1))=modl(1:size(modl1))+modl1; modl(size(modl1)+1:size(modl))=modl(size(modl1)+1:size(modl))+modl2;
        else        ;  data(1:size(data1))=data(1:size(data1))+data1; data(size(data1)+1:size(data))=data(size(data1)+1:size(data))+data2;
        endif    
    end if
    riter=iter/niter; iter=iter+1
    !write(0,*)iter, riter, niter
    stat = 0
  end function
!#######
  integer function conv_precon_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    real, dimension(:),allocatable :: modl1, data1
    real, dimension(:),allocatable :: modl2, data2
    allocate(modl1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(modl2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    allocate(data1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(data2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    stat = 1
    if (adj) then;   modl1 = 0.; modl2=0; data1=data(1:size(data1)); data2=data(size(data1)+1:size(data))
    else         ;   data1 = 0.; data2=0; modl1=modl(1:size(modl1)); modl2=modl(size(modl1)+1:size(modl))
    end if
    call hess_precon_diag2(adj, modl1, modl2, data1, data2)
    if (.not. add) then
        if (adj)then;  modl(1:size(modl1))=modl1; modl(size(modl1)+1:size(modl))=modl2;
        else        ;  data(1:size(data1))=data1; data(size(data1)+1:size(data))=data2;
        endif    
    else
        if (adj)then;  modl(1:size(modl1))=modl(1:size(modl1))+modl1; modl(size(modl1)+1:size(modl))=modl(size(modl1)+1:size(modl))+modl2;
        else        ;  data(1:size(data1))=data(1:size(data1))+data1; data(size(data1)+1:size(data))=data(size(data1)+1:size(data))+data2;
        endif    
    end if
    stat = 0
  end function
!#######
  integer function conv_mask_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    real, dimension(:),allocatable :: modl1, data1
    real, dimension(:),allocatable :: modl2, data2
    allocate(modl1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(modl2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    allocate(data1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(data2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    stat = 1
    if (adj) then;   modl1 = 0.; modl2=0; data1=data(1:size(data1)); data2=data(size(data1)+1:size(data))
    else         ;   data1 = 0.; data2=0; modl1=modl(1:size(modl1)); modl2=modl(size(modl1)+1:size(modl))
    end if
    call hess_mask(adj, modl1, modl2, data1, data2)
    if (.not. add) then
        if (adj)then;  modl(1:size(modl1))=modl1; modl(size(modl1)+1:size(modl))=modl2;
        else        ;  data(1:size(data1))=data1; data(size(data1)+1:size(data))=data2;
        endif    
    else
        if (adj)then;  modl(1:size(modl1))=modl(1:size(modl1))+modl1; modl(size(modl1)+1:size(modl))=modl(size(modl1)+1:size(modl))+modl2;
        else        ;  data(1:size(data1))=data(1:size(data1))+data1; data(size(data1)+1:size(data))=data(size(data1)+1:size(data))+data2;
        endif    
    end if
    stat = 0
  end function
!#######
  integer function conv_weight_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    real, dimension(:),allocatable :: modl1, data1
    real, dimension(:),allocatable :: modl2, data2
    allocate(modl1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(modl2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    allocate(data1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(data2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    stat = 1
    if (adj) then;   modl1 = 0.; modl2=0; data1=data(1:size(data1)); data2=data(size(data1)+1:size(data))
    else         ;   data1 = 0.; data2=0; modl1=modl(1:size(modl1)); modl2=modl(size(modl1)+1:size(modl))
    end if
    call hess_weight(adj, modl1, modl2, data1, data2)
    if (.not. add) then
        if (adj)then;  modl(1:size(modl1))=modl1; modl(size(modl1)+1:size(modl))=modl2;
        else        ;  data(1:size(data1))=data1; data(size(data1)+1:size(data))=data2;
        endif    
    else
        if (adj)then;  modl(1:size(modl1))=modl(1:size(modl1))+modl1; modl(size(modl1)+1:size(modl))=modl(size(modl1)+1:size(modl))+modl2;
        else        ;  data(1:size(data1))=data(1:size(data1))+data1; data(size(data1)+1:size(data))=data(size(data1)+1:size(data))+data2;
        endif    
    end if
    stat = 0
  end function
!#######
  integer function conv_grad_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    real, dimension(:),allocatable :: modl1, data1
    real, dimension(:),allocatable :: modl2, data2
    allocate(modl1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(modl2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    allocate(data1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(data2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    stat = 1
    if (adj) then;   modl1 = 0.; modl2=0; data1=data(1:size(data1)); data2=data(size(data1)+1:size(data))
    else         ;   data1 = 0.; data2=0; modl1=modl(1:size(modl1)); modl2=modl(size(modl1)+1:size(modl))
    end if
    call hess_grad1(adj, modl1, modl2, data1, data2)
    if (.not. add) then
        if (adj)then;  modl(1:size(modl1))=modl1; modl(size(modl1)+1:size(modl))=modl2;
        else        ;  data(1:size(data1))=data1; data(size(data1)+1:size(data))=data2;
        endif    
    else
        if (adj)then;  modl(1:size(modl1))=modl(1:size(modl1))+modl1; modl(size(modl1)+1:size(modl))=modl(size(modl1)+1:size(modl))+modl2;
        else        ;  data(1:size(data1))=data(1:size(data1))+data1; data(size(data1)+1:size(data))=data(size(data1)+1:size(data))+data2;
        endif    
    end if
    stat = 0
  end function
!#######
  integer function conv_grad_oper2(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    real, dimension(:),allocatable :: modl1, data1
    real, dimension(:),allocatable :: modl2, data2
    allocate(modl1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(modl2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    allocate(data1(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n)); allocate(data2(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    stat = 1
    if (adj) then;   modl1 = 0.; modl2=0; data1=data(1:size(data1)); data2=data(size(data1)+1:size(data))
    else         ;   data1 = 0.; data2=0; modl1=modl(1:size(modl1)); modl2=modl(size(modl1)+1:size(modl))
    end if
    call hess_grad_base(adj, modl1, data1)
    !call   hess_grad_4d(adj, modl2, data2)
    call   hess_grad_4d_xz(adj, modl2, data2)
    if (.not. add) then
        if (adj)then;  modl(1:size(modl1))=modl1; modl(size(modl1)+1:size(modl))=modl2;
        else        ;  data(1:size(data1))=data1; data(size(data1)+1:size(data))=data2;
        endif    
    else
        if (adj)then;  modl(1:size(modl1))=modl(1:size(modl1))+modl1; modl(size(modl1)+1:size(modl))=modl(size(modl1)+1:size(modl))+modl2;
        else        ;  data(1:size(data1))=data(1:size(data1))+data1; data(size(data1)+1:size(data))=data(size(data1)+1:size(data))+data2;
        endif    
    end if
    stat = 0
  end function
!#####################################
  subroutine hess_convolve(adj, modl1, modl2, data1, data2)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl
    integer                :: ix_offd, iy_offd, iz_offd
    integer                :: index_x_offd, index_y_offd, index_z_offd
    integer                :: index_x_modl, index_y_modl, index_z_modl
    real                   :: x_offd, y_offd, z_offd
    real                   :: x_modl, y_modl, z_modl
    real                   :: maxz,maxx,maxy
    integer::i,j,k,ii
        
    maxz=hgrid%az_modl%o + (hgrid%az_modl%n-1)*hgrid%az_modl%d
    maxx=hgrid%ax_modl%o + (hgrid%ax_modl%n-1)*hgrid%ax_modl%d
    maxy=hgrid%ay_modl%o + (hgrid%ay_modl%n-1)*hgrid%ay_modl%d
         do iz_modl= 1,hgrid%az_modl%n ;  z_modl = hgrid%az_modl%o + (iz_modl-1)*hgrid%az_modl%d; index_z_modl = 1.5 + (z_modl - hgrid%az_modl%o)/hgrid%az_modl%d;
         do iy_modl= 1,hgrid%ay_modl%n ;  y_modl = hgrid%ay_modl%o + (iy_modl-1)*hgrid%ay_modl%d; index_y_modl = 1.5 + (y_modl - hgrid%ay_modl%o)/hgrid%ay_modl%d;
         do ix_modl= 1,hgrid%ax_modl%n ;  x_modl = hgrid%ax_modl%o + (ix_modl-1)*hgrid%ax_modl%d; index_x_modl = 1.5 + (x_modl - hgrid%ax_modl%o)/hgrid%ax_modl%d;
                call hess_io(hess1, hgrid, ix_modl, iy_modl, iz_modl, file1_, "reed")
                call hess_io(hess2, hgrid, ix_modl, iy_modl, iz_modl, file2_, "reed")
                do  iz_offd = 1,hgrid%az_offd%n; z_offd = z_modl + hgrid%az_offd%o + (iz_offd-1)*hgrid%az_offd%d; 
                        if (z_offd < hgrid%az_modl%o .or. z_offd > maxz ) cycle
                        index_z_offd = 1.5 + (z_offd - hgrid%az_modl%o)/hgrid%az_modl%d;
                do  iy_offd= 1,hgrid%ay_offd%n; y_offd = y_modl + hgrid%ay_offd%o + (iy_offd-1)*hgrid%ay_offd%d;
                        if (y_offd < hgrid%ay_modl%o .or. y_offd > maxy ) cycle
                        index_y_offd = 1.5 + (y_offd - hgrid%ay_modl%o)/hgrid%ay_modl%d;
                do  ix_offd= 1,hgrid%ax_offd%n; x_offd = x_modl + hgrid%ax_offd%o + (ix_offd-1)*hgrid%ax_offd%d;
                        if (x_offd < hgrid%ax_modl%o .or. x_offd > maxx ) cycle
                        if (iz_offd == hgrid%az_offd%n .and. ix_offd > ceiling(hgrid%ax_offd%n/2.)) cycle
                        index_x_offd = 1.5 + (x_offd - hgrid%ax_modl%o)/hgrid%ax_modl%d;
                        if (adj) then
                                modl1(index_z_offd,index_x_offd,index_y_offd) = modl1(index_z_offd,index_x_offd,index_y_offd) &
              + (hess1(ix_offd,iy_offd,iz_offd)+hess2(ix_offd,iy_offd,iz_offd))*data1(index_z_modl,index_x_modl,index_y_modl) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*data2(index_z_modl,index_x_modl,index_y_modl) 
                                modl1(index_z_modl,index_x_modl,index_y_modl) = modl1(index_z_modl,index_x_modl,index_y_modl) &
              + (hess1(ix_offd,iy_offd,iz_offd)+hess2(ix_offd,iy_offd,iz_offd))*data1(index_z_offd,index_x_offd,index_y_offd) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*data2(index_z_offd,index_x_offd,index_y_offd) 
                                modl2(index_z_offd,index_x_offd,index_y_offd) = modl2(index_z_offd,index_x_offd,index_y_offd) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*data1(index_z_modl,index_x_modl,index_y_modl) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*data2(index_z_modl,index_x_modl,index_y_modl) 
                                modl2(index_z_modl,index_x_modl,index_y_modl) = modl2(index_z_modl,index_x_modl,index_y_modl) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*data1(index_z_offd,index_x_offd,index_y_offd) & 
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*data2(index_z_offd,index_x_offd,index_y_offd) 
                        else 
                                data1(index_z_modl,index_x_modl,index_y_modl) = data1(index_z_modl,index_x_modl,index_y_modl) & 
              + (hess1(ix_offd,iy_offd,iz_offd)+hess2(ix_offd,iy_offd,iz_offd))*modl1(index_z_offd,index_x_offd,index_y_offd) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*modl2(index_z_offd,index_x_offd,index_y_offd) 
                                data1(index_z_offd,index_x_offd,index_y_offd) = data1(index_z_offd,index_x_offd,index_y_offd) & 
              + (hess1(ix_offd,iy_offd,iz_offd)+hess2(ix_offd,iy_offd,iz_offd))*modl1(index_z_modl,index_x_modl,index_y_modl) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*modl2(index_z_modl,index_x_modl,index_y_modl)
                                data2(index_z_modl,index_x_modl,index_y_modl) = data2(index_z_modl,index_x_modl,index_y_modl) & 
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*modl1(index_z_offd,index_x_offd,index_y_offd) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*modl2(index_z_offd,index_x_offd,index_y_offd)
                                data2(index_z_offd,index_x_offd,index_y_offd) = data2(index_z_offd,index_x_offd,index_y_offd) & 
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*modl1(index_z_modl,index_x_modl,index_y_modl) &
              + (hess2(ix_offd,iy_offd,iz_offd)                               )*modl2(index_z_modl,index_x_modl,index_y_modl)

                        end if
                end do
                end do
                end do
        end do
        end do
        end do
    ii=1
       ! data2(hgrid%az_modl%n*2/3:hgrid%az_modl%n,:,:)=0

    return
  end subroutine
!#####################################
  subroutine hess_precon_diag(adj, modl1, modl2, data1, data2)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl
    integer                :: ix_offd, iy_offd, iz_offd
    integer                :: index_x_offd, index_y_offd, index_z_offd
    integer                :: index_x_modl, index_y_modl, index_z_modl
    real                   :: x_offd, y_offd, z_offd
    real                   :: x_modl, y_modl, z_modl
    real                   :: maxz,maxx,maxy
    integer::i,j,k,ii
        

         do iz_modl= 1,hgrid%az_modl%n ;  
         do iy_modl= 1,hgrid%ay_modl%n ;  
         do ix_modl= 1,hgrid%ax_modl%n ;  
                        if (adj) then
                                modl1(iz_modl,ix_modl,iy_modl) = modl1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_1(ix_modl,iy_modl,iz_modl)+hess_2(ix_modl,iy_modl,iz_modl))*data1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*data2(iz_modl,ix_modl,iy_modl) 
                                                modl2(iz_modl,ix_modl,iy_modl) = modl2(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*data1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*data2(iz_modl,ix_modl,iy_modl) 
                        else 
                                data1(iz_modl,ix_modl,iy_modl) = data1(iz_modl,ix_modl,iy_modl)  & 
                                      + (hess_1(ix_modl,iy_modl,iz_modl)+hess_2(ix_modl,iy_modl,iz_modl))*modl1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*modl2(iz_modl,ix_modl,iy_modl) 
                                data2(index_z_modl,index_x_modl,index_y_modl) = data2(iz_modl,ix_modl,iy_modl) & 
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*modl1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*modl2(iz_modl,ix_modl,iy_modl)
                        end if
        end do
        end do
        end do
    return
  end subroutine
!#####################################
  subroutine hess_precon_diag2(adj, modl1, modl2, data1, data2)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl
write(0,*)ones
    if (ones)then
        if (adj)then
            modl1=data1
            modl2=data2
        else
            data1=modl1
            data2=modl2
        endif
    else

         call hess_io_diag(hess_1, hgrid, file1_, "reed")
         call hess_io_diag(hess_2, hgrid, file2_, "reed")

         do iz_modl= 1,hgrid%az_modl%n ;  
         do iy_modl= 1,hgrid%ay_modl%n ;  
         do ix_modl= 1,hgrid%ax_modl%n ;  
         if (adj) then
               modl1(iz_modl,ix_modl,iy_modl) = modl1(iz_modl,ix_modl,iy_modl) + 1/(hess_1(ix_modl,iy_modl,iz_modl)+hess_2(ix_modl,iy_modl,iz_modl))*data1(iz_modl,ix_modl,iy_modl) 
               modl2(iz_modl,ix_modl,iy_modl) = modl2(iz_modl,ix_modl,iy_modl) + 1/(hess_2(ix_modl,iy_modl,iz_modl)                               )*data2(iz_modl,ix_modl,iy_modl) 
         else 
               data1(iz_modl,ix_modl,iy_modl) = data1(iz_modl,ix_modl,iy_modl) + 1/(hess_1(ix_modl,iy_modl,iz_modl)+hess_2(ix_modl,iy_modl,iz_modl))*modl1(iz_modl,ix_modl,iy_modl) 
               data2(iz_modl,ix_modl,iy_modl) = data2(iz_modl,ix_modl,iy_modl) + 1/(hess_2(ix_modl,iy_modl,iz_modl)                               )*modl2(iz_modl,ix_modl,iy_modl)
         end if
         end do
         end do
         end do
    endif
    return
  end subroutine
!  ######################
  subroutine hess_precon_block_diag(adj, inv, modl1, modl2, data1, data2)
    logical                :: adj,inv
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl
    integer                :: ix_offd, iy_offd, iz_offd
    integer                :: index_x_offd, index_y_offd, index_z_offd
    integer                :: index_x_modl, index_y_modl, index_z_modl
    real                   :: x_offd, y_offd, z_offd
    real                   :: x_modl, y_modl, z_modl
    real                   :: maxz,maxx,maxy
    integer::i,j,k,ii
        
    call hess_io_diag(hess_1, hgrid, file1_, "reed")
    call hess_io_diag(hess_2, hgrid, file2_, "reed")

         if(inv)then
                
         else

         do iz_modl= 1,hgrid%az_modl%n ;  
         do iy_modl= 1,hgrid%ay_modl%n ;  
         do ix_modl= 1,hgrid%ax_modl%n ;  
                        if (adj) then
                                modl1(iz_modl,ix_modl,iy_modl) = modl1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_1(ix_modl,iy_modl,iz_modl)+hess_2(ix_modl,iy_modl,iz_modl))*data1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*data2(iz_modl,ix_modl,iy_modl) 
                                                modl2(iz_modl,ix_modl,iy_modl) = modl2(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*data1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*data2(iz_modl,ix_modl,iy_modl) 
                        else 
                                data1(iz_modl,ix_modl,iy_modl) = data1(iz_modl,ix_modl,iy_modl)  & 
                                      + (hess_1(ix_modl,iy_modl,iz_modl)+hess_2(ix_modl,iy_modl,iz_modl))*modl1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*modl2(iz_modl,ix_modl,iy_modl) 
                                data2(index_z_modl,index_x_modl,index_y_modl) = data2(iz_modl,ix_modl,iy_modl) & 
                                      + (hess_1(ix_modl,iy_modl,iz_modl)                               )*modl1(iz_modl,ix_modl,iy_modl) &
                                      + (hess_2(ix_modl,iy_modl,iz_modl)                               )*modl2(iz_modl,ix_modl,iy_modl)
                        end if
        end do
        end do
        end do
    endif
  return
  end subroutine

!#####################################
  subroutine hess_weight(adj, modl1, modl2, data1, data2)
    optional::  modl1, modl2, data1, data2
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl

    if(present(modl1).and.present(data1).and.exist_file(weight1_))then
        do ix_modl= 1,hgrid%ax_modl%n ;  
        do iy_modl= 1,hgrid%ay_modl%n ; 
        do iz_modl= 1,hgrid%az_modl%n ;  
                        if (adj) then
                                modl1(iz_modl,ix_modl,iy_modl) = modl1(iz_modl,ix_modl,iy_modl) + (weight1(iz_modl,ix_modl,iy_modl))*data1(iz_modl,ix_modl,iy_modl) 
                        else 
                                data1(iz_modl,ix_modl,iy_modl) = data1(iz_modl,ix_modl,iy_modl) + (weight1(iz_modl,ix_modl,iy_modl))*modl1(iz_modl,ix_modl,iy_modl) 
                        end if
        end do
        end do
!                        if (ix_modl==hgrid%ax_modl%n/2)write(0,*) weight1(:,ix_modl,:) 
     
        end do
    elseif(present(modl1).and.present(data1))then
                        if (adj) then
                                modl1 = data1 
                        else 
                                data1 = modl1 
                        end if
    else
        call seperr('hess_wght1: Trouble getting to your weighting file...hv you read it in??!!!!!')
    endif
    if(present(modl2).and.present(data2).and.exist_file(weight2_))then
        do ix_modl= 1,hgrid%ax_modl%n ;  
        do iy_modl= 1,hgrid%ay_modl%n ; 
        do iz_modl= 1,hgrid%az_modl%n ;  
                        if (adj) then
                                modl2(iz_modl,ix_modl,iy_modl) = modl2(iz_modl,ix_modl,iy_modl) + (weight2(iz_modl,ix_modl,iy_modl))*data2(iz_modl,ix_modl,iy_modl) 
                        else 
                                data2(iz_modl,ix_modl,iy_modl) = data2(iz_modl,ix_modl,iy_modl) + (weight2(iz_modl,ix_modl,iy_modl))*modl2(iz_modl,ix_modl,iy_modl)
                        end if
        end do
        end do
        end do
    elseif(present(modl2).and.present(data2))then
                        if (adj) then
                                modl2 = data2 
                        else 
                                data2 = modl2 
                        end if
    else
        call seperr('hess_wght2: Trouble getting to your weighting file...hv you read it in??!!!!!')
    endif
    return
  end subroutine
!#####################################
  subroutine hess_mask(adj, modl1, modl2, data1, data2)
    optional:: modl1, modl2, data1, data2
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl

    if(present(modl1).and.present(data1))then
        do ix_modl= 1,hgrid%ax_modl%n ;  
        do iy_modl= 1,hgrid%ay_modl%n ; 
        do iz_modl= 1,hgrid%az_modl%n ; 
                        if (adj) then
                                modl1(iz_modl,ix_modl,iy_modl) = modl1(iz_modl,ix_modl,iy_modl) + (mask1(iz_modl,ix_modl,iy_modl))*data1(iz_modl,ix_modl,iy_modl) 
                        else 
                                data1(iz_modl,ix_modl,iy_modl) = data1(iz_modl,ix_modl,iy_modl) + (mask1(iz_modl,ix_modl,iy_modl))*modl1(iz_modl,ix_modl,iy_modl) 
                        end if
        end do
        end do
        end do
    elseif(present(modl1).and.present(data1))then
        call seperr('hess_mask: Need to specify both modl1 and data1 or none!!!')
    endif
    if(present(modl2).and.present(data2))then
        do iz_modl= 1,hgrid%az_modl%n ; !write(*,*)weight2(iz_modl,1,1) 
        do iy_modl= 1,hgrid%ay_modl%n ; 
        do ix_modl= 1,hgrid%ax_modl%n ;  
                        if (adj) then
                                modl2(iz_modl,ix_modl,iy_modl) = modl2(iz_modl,ix_modl,iy_modl) + (mask2(iz_modl,ix_modl,iy_modl))*data2(iz_modl,ix_modl,iy_modl) 
                        else 
                                data2(iz_modl,ix_modl,iy_modl) = data2(iz_modl,ix_modl,iy_modl) + (mask2(iz_modl,ix_modl,iy_modl))*modl2(iz_modl,ix_modl,iy_modl)
                        end if
        end do
        end do
        end do
    elseif(present(modl2).and.present(data2))then
        call seperr('hess_mask: Need to specify both modl2 and data2 or none!!!')
    endif

    return
  end subroutine
!#####################################
  subroutine hess_grad1(adj, modl1, modl2, data1, data2)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data1
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl2
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data2

    integer                :: ix_modl, iy_modl, iz_modl

do ix_modl= 1,hgrid%ax_modl%n-1;  
do iy_modl= 1,hgrid%ay_modl%n ;  
do iz_modl= 1,hgrid%az_modl%n ;  
        if (adj) then
                modl1(iz_modl,ix_modl+1,iy_modl)=modl1(iz_modl,ix_modl+1,iy_modl)+epsi1(ix_modl,iy_modl,iz_modl)*data1(iz_modl,ix_modl ,iy_modl) 
                modl1(iz_modl,ix_modl  ,iy_modl)=modl1(iz_modl,ix_modl  ,iy_modl)-epsi1(ix_modl,iy_modl,iz_modl)*data1(iz_modl,ix_modl ,iy_modl) 
                modl2(iz_modl,ix_modl+1,iy_modl)=modl2(iz_modl,ix_modl+1,iy_modl)+epsi1(ix_modl,iy_modl,iz_modl)*data2(iz_modl,ix_modl ,iy_modl) 
                modl2(iz_modl,ix_modl  ,iy_modl)=modl2(iz_modl,ix_modl  ,iy_modl)-epsi1(ix_modl,iy_modl,iz_modl)*data2(iz_modl,ix_modl ,iy_modl) 
        else 
                data1(iz_modl,ix_modl  ,iy_modl)=data1(iz_modl,ix_modl  ,iy_modl)+epsi1(ix_modl,iy_modl,iz_modl)*(modl1(iz_modl,ix_modl+1,iy_modl)-& 
                                                                                                                modl1(iz_modl,ix_modl  ,iy_modl)) 
                data2(iz_modl,ix_modl  ,iy_modl)=data2(iz_modl,ix_modl  ,iy_modl)+epsi1(ix_modl,iy_modl,iz_modl)*(modl2(iz_modl,ix_modl+1,iy_modl)-& 
                                                                                                                modl2(iz_modl,ix_modl  ,iy_modl)) 
        end if
end do
end do
end do
    return
  end subroutine
                                                                                  
!#####################################
  subroutine hess_grad_4d(adj, modl, data)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data
    integer                :: ix_modl, iy_modl, iz_modl

    index_rsv_zmin=1.5+(rsv_zmin-hgrid%az_modl%o)/hgrid%az_modl%d
do ix_modl= 1,hgrid%ax_modl%n-1;  
do iy_modl= 1,hgrid%ay_modl%n ;  
do iz_modl= index_rsv_zmin,hgrid%az_modl%n ; 
        if (adj) then
                modl(iz_modl,ix_modl+1,iy_modl)=modl(iz_modl,ix_modl+1,iy_modl)+1.2*epsi2(ix_modl,iy_modl,iz_modl)*data(iz_modl,ix_modl ,iy_modl) 
                modl(iz_modl,ix_modl  ,iy_modl)=modl(iz_modl,ix_modl  ,iy_modl)-1.2*epsi2(ix_modl,iy_modl,iz_modl)*data(iz_modl,ix_modl ,iy_modl) 
        else 
                data(iz_modl,ix_modl  ,iy_modl)=data(iz_modl,ix_modl  ,iy_modl)+1.2*epsi2(ix_modl,iy_modl,iz_modl)*(modl(iz_modl,ix_modl+1,iy_modl)-& 
                                                                                                                modl(iz_modl,ix_modl  ,iy_modl)) 
        end if
end do
end do
end do
    return
  end subroutine
!#####################################
  subroutine hess_grad_base(adj, modl, data)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data
    integer                :: ix_modl, iy_modl, iz_modl

    index_zmin=1.5+(zmin-hgrid%az_modl%o)/hgrid%az_modl%d
    index_zmax=1.5+(zmax-hgrid%az_modl%o)/hgrid%az_modl%d
!    write(0,*) 'bas',index_zmin,index_zmax, zmin,zmax
do ix_modl= 1,hgrid%ax_modl%n-1;  
do iy_modl= 1,hgrid%ay_modl%n ;  
do iz_modl= index_zmin,index_zmax;  
        if (adj) then
                modl(iz_modl,ix_modl+1,iy_modl)=modl(iz_modl,ix_modl+1,iy_modl)+1.2*epsi1(ix_modl,iy_modl,iz_modl)* data(iz_modl,ix_modl ,iy_modl) 
                modl(iz_modl,ix_modl  ,iy_modl)=modl(iz_modl,ix_modl  ,iy_modl)-1.2*epsi1(ix_modl,iy_modl,iz_modl)* data(iz_modl,ix_modl ,iy_modl) 
        else
                data(iz_modl,ix_modl  ,iy_modl)=data(iz_modl,ix_modl  ,iy_modl)+1.2*epsi1(ix_modl,iy_modl,iz_modl)*(modl(iz_modl,ix_modl+1,iy_modl)-& 
                                                                                                                modl(iz_modl,ix_modl  ,iy_modl)) 
        end if
end do
end do
end do
if (index_zmin /=1)then
        if(adj)then
                modl(1:index_zmin-1,:,:)=data(1:index_zmin-1,:,:)
        else
                data(1:index_zmin-1,:,:)=modl(1:index_zmin-1,:,:) 
        endif
endif
do ix_modl= 1,hgrid%ax_modl%n;  
do iy_modl= 1,hgrid%ay_modl%n ;  
do iz_modl= 1,hgrid%az_modl%n ;  
        if (adj) then
                modl(iz_modl,ix_modl,iy_modl)=modl(iz_modl,ix_modl,iy_modl)*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl))
        else 
                data(iz_modl,ix_modl,iy_modl)=data(iz_modl,ix_modl,iy_modl)+(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl))
        end if
end do
end do
end do
return
  end subroutine
!##################
  subroutine hess_grad_4d_xz(adj, modl, data)
    logical                :: adj
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: modl
    real, dimension(hgrid%az_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n) :: data
    integer                :: ix_modl, iy_modl, iz_modl
    real                  :: xfact,zfact
    real,dimension(hgrid%az_modl%n) :: xxfact,zzfact

    index_rsv_zmin=1.5+abs(rsv_zmin-hgrid%az_modl%o)/hgrid%az_modl%d
    index_rsv_zmax=1.5+abs(rsv_zmax-hgrid%az_modl%o)/hgrid%az_modl%d
    index_rsv_xmin=1.5+abs(abs(rsv_xmin)-abs(hgrid%ax_modl%o))/hgrid%ax_modl%d
    index_rsv_xmax=1.5+abs(rsv_xmax-hgrid%ax_modl%o)/hgrid%ax_modl%d
    write(0,*)'rsv', index_rsv_zmin,index_rsv_zmax, rsv_zmin,rsv_zmax
    write(0,*)'rsv', index_rsv_xmin,index_rsv_xmax, rsv_xmin,rsv_xmax,maxval(epsi1),minval(epsi2)

!do ix_modl= index_rsv_xmin,index_rsv_xmax-1;  
do ix_modl= 1,hgrid%ax_modl%n-1;  
do iy_modl= 1,hgrid%ay_modl%n ;  
do iz_modl= index_rsv_zmin-4,min(index_rsv_zmax+6, hgrid%az_modl%n) ; 
        if (adj) then
                modl(iz_modl,ix_modl+1,iy_modl)=modl(iz_modl,ix_modl+1,iy_modl)+.5*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) & 
                                                                                 *data(iz_modl  ,ix_modl,iy_modl) 
                modl(iz_modl,ix_modl  ,iy_modl)=modl(iz_modl,ix_modl  ,iy_modl)-.5*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) &
                                                                                 *data(iz_modl  ,ix_modl,iy_modl) 
        else 
                data(iz_modl,ix_modl  ,iy_modl)=data(iz_modl,ix_modl  ,iy_modl)+.5*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) &
                                                                                 *(modl(iz_modl,ix_modl+1,iy_modl)-modl(iz_modl ,ix_modl,iy_modl)) 
        end if
end do
end do
end do

do ix_modl= 1,hgrid%ax_modl%n;  
do iy_modl= 1,hgrid%ay_modl%n;  
do iz_modl= 1,index_rsv_zmin;
!write(0,*)ix_modl,iz_modl,sum(epsi1(ix_modl,iy_modl,:)),epsi1(ix_modl,iy_modl,iz_modl)
        if (adj) then
                modl(iz_modl+1,ix_modl,iy_modl)=modl(iz_modl+1,ix_modl,iy_modl)+& 
                                                0.5*data(iz_modl  ,ix_modl,iy_modl) 
                modl(iz_modl  ,ix_modl,iy_modl)=modl(iz_modl  ,ix_modl,iy_modl)- & 
                                                0.5*data(iz_modl  ,ix_modl,iy_modl) 
        else 
                data(iz_modl  ,ix_modl,iy_modl)=data(iz_modl  ,ix_modl,iy_modl)+ & 
                                                0.5*(modl(iz_modl+1,ix_modl,iy_modl)-& 
                                                                                                                modl(iz_modl ,ix_modl,iy_modl)) 
        end if
end do
do iz_modl=index_rsv_zmax-2, hgrid%az_modl%n-1; 
        if (adj) then
                modl(iz_modl+1,ix_modl,iy_modl)=modl(iz_modl+1,ix_modl,iy_modl)+ & 
                                                0.5*data(iz_modl  ,ix_modl,iy_modl) 
                modl(iz_modl  ,ix_modl,iy_modl)=modl(iz_modl  ,ix_modl,iy_modl)- & 
                                                0.5*data(iz_modl  ,ix_modl,iy_modl) 
        else 
                data(iz_modl  ,ix_modl,iy_modl)=data(iz_modl  ,ix_modl,iy_modl)+ & 
                                                0.5*(modl(iz_modl+1,ix_modl,iy_modl)-& 
                                                                                                                modl(iz_modl ,ix_modl,iy_modl)) 
        end if
end do
end do
end do
!weights
zzfact=1
do iz_modl= 1,index_rsv_zmin;
        zzfact(iz_modl)=1.2+1.2*(index_rsv_zmin-iz_modl)/(index_rsv_zmin-1.)
enddo
do iz_modl=index_rsv_zmax-2, hgrid%az_modl%n;
        zzfact(iz_modl)=1.2+1.2*(-(index_rsv_zmax-2.)+iz_modl)/(hgrid%az_modl%n-index_rsv_zmax-2)
end do
!
do ix_modl= 1,hgrid%ax_modl%n;  
do iy_modl= 1,hgrid%ay_modl%n;  
        if (adj) then
                modl(:,ix_modl,iy_modl)=modl(:,ix_modl,iy_modl)*zzfact
        else 
                data(:,ix_modl,iy_modl)=data(:,ix_modl,iy_modl)*zzfact
        end if
end do
end do
!%edges//////////////
!do ix_modl= 1,hgrid%ax_modl%n-1;  
!do iy_modl= 1,hgrid%ay_modl%n;  
!do iz_modl= index_rsv_zmin-5,index_rsv_zmin
!write(0,*)ix_modl,iz_modl,sum(epsi1(ix_modl,iy_modl,:)),epsi1(ix_modl,iy_modl,iz_modl)
!zfact=1.*abs(index_rsv_zmin-iz_modl)/5.
!        if (adj) then
!                modl(iz_modl,ix_modl+1,iy_modl)=modl(iz_modl,ix_modl+1,iy_modl)+1.*zfact*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) & 
!                                                                                 *data(iz_modl  ,ix_modl,iy_modl) 
!                modl(iz_modl,ix_modl  ,iy_modl)=modl(iz_modl,ix_modl  ,iy_modl)-1.*zfact*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) &
!                                                                                 *data(iz_modl  ,ix_modl,iy_modl) 
!        else 
!                data(iz_modl,ix_modl  ,iy_modl)=data(iz_modl,ix_modl  ,iy_modl)+1.*zfact*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) &
!                                                                                 *(modl(iz_modl,ix_modl+1,iy_modl)-modl(iz_modl ,ix_modl,iy_modl)) 
!        end if
!end do
!do iz_modl=index_rsv_zmax-2,index_rsv_zmax+5;
!zfact=1.*abs(index_rsv_zmax+5.-iz_modl)/7.
!        if (adj) then
!                modl(iz_modl,ix_modl+1,iy_modl)=modl(iz_modl,ix_modl+1,iy_modl)+1.*zfact*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) & 
!                                                                                 *data(iz_modl  ,ix_modl,iy_modl) 
!                modl(iz_modl,ix_modl  ,iy_modl)=modl(iz_modl,ix_modl  ,iy_modl)-1.*zfact*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) &
!                                                                                 *data(iz_modl  ,ix_modl,iy_modl) 
!        else 
!                data(iz_modl,ix_modl  ,iy_modl)=data(iz_modl,ix_modl  ,iy_modl)+1.*zfact*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl)) &
!                                                                                 *(modl(iz_modl,ix_modl+1,iy_modl)-modl(iz_modl ,ix_modl,iy_modl)) 
!        end if
!end do
!end do
!end do
!
do ix_modl= 1,index_rsv_xmin;
xfact=float(index_rsv_xmin-ix_modl)/index_rsv_xmin
!write(0,*)ix_modl,index_rsv_xmin,index_rsv_xmin-ix_modl,xfact
do iy_modl= 1,hgrid%ay_modl%n;  
do iz_modl= index_rsv_zmin,index_rsv_zmax-1 ; 
        if (adj) then
                modl(iz_modl+1,ix_modl,iy_modl)=modl(iz_modl+1,ix_modl,iy_modl)+xfact* & 
                                                .4*data(iz_modl  ,ix_modl,iy_modl) 
                modl(iz_modl  ,ix_modl,iy_modl)=modl(iz_modl  ,ix_modl,iy_modl)- xfact* & 
                                                .4*data(iz_modl  ,ix_modl,iy_modl) 
        else 
                data(iz_modl  ,ix_modl,iy_modl)=data(iz_modl  ,ix_modl,iy_modl)+ xfact* & 
                                                .4*(modl(iz_modl+1,ix_modl,iy_modl)-& 
                                                                                                                modl(iz_modl ,ix_modl,iy_modl)) 
        end if
end do
end do
end do
!write(*,*)'----'
do ix_modl= index_rsv_xmax,hgrid%ax_modl%n ;
xfact=1.-float(hgrid%ax_modl%n-ix_modl)/(hgrid%ax_modl%n-index_rsv_xmax)
!write(0,*)ix_modl,xfact
do iy_modl= 1,hgrid%ay_modl%n;  
do iz_modl= index_rsv_zmin,index_rsv_zmax-1 ; 
        if (adj) then
                modl(iz_modl+1,ix_modl,iy_modl)=modl(iz_modl+1,ix_modl,iy_modl)+ xfact* & 
                                                .6*data(iz_modl  ,ix_modl,iy_modl) 
                modl(iz_modl  ,ix_modl,iy_modl)=modl(iz_modl  ,ix_modl,iy_modl)- xfact* & 
                                                .6*data(iz_modl  ,ix_modl,iy_modl) 
        else 
                data(iz_modl  ,ix_modl,iy_modl)=data(iz_modl  ,ix_modl,iy_modl)+ xfact* & 
                                                .6*(modl(iz_modl+1,ix_modl,iy_modl)-& 
                                                                                                                modl(iz_modl ,ix_modl,iy_modl)) 
        end if
end do
end do
end do

do ix_modl= 1,hgrid%ax_modl%n;  
do iy_modl= 1,hgrid%ay_modl%n ;  
do iz_modl= 1,hgrid%az_modl%n ;  
        if (adj) then
                modl(iz_modl,ix_modl,iy_modl)=modl(iz_modl,ix_modl,iy_modl)*(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl))
        else 
                data(iz_modl,ix_modl,iy_modl)=data(iz_modl,ix_modl,iy_modl)+(epsi1(ix_modl,iy_modl,iz_modl)+epsi2(ix_modl,iy_modl,iz_modl))
        end if
end do
end do
end do

    return
  end subroutine
!##################
  subroutine hess_io(hess, hgrid, ix_modl, iy_modl, iz_modl, file_, flag)
    real, dimension(:,:,:) :: hess
    type(hes_grid)         :: hgrid
    integer                :: ix_modl, iy_modl, iz_modl
    character(len=*)       :: file_
    character(len=4)       :: flag
    integer, dimension(10) :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 10
    n_cube = (/hgrid%ahx_offd%n, hgrid%ahy_offd%n, hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, &
               hgrid%ahx_modl%n, hgrid%ahy_modl%n, hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n/)
    j_wind = (/1               , 1               , 1              , 1              , 1              , & 
               1               , 1               , 1              , 1              , 1              /)
    n_wind = (/1               , 1               , hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, &
               1               , 1               , 1              , 1              , 1              /)
    f_wind = (/0               , 0               , 0              , 0              , 0              , &
               0               , 0               , ix_modl-1      , iy_modl-1      , iz_modl-1      /)
    if (flag == "rite") then
    !   err = srite_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else if (flag == "reed") then
       hess = 0.
       err = sreed_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    end if
    return
  end subroutine
!###################  
  subroutine hess_io2(hess, hgrid, file_, flag)
    real, dimension(:,:,:,:,:,:) :: hess
    type(hes_grid)         :: hgrid
    integer                :: ix_modl, iy_modl, iz_modl
    character(len=*)       :: file_
    character(len=4)       :: flag
    integer, dimension(10) :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 10
    n_cube = (/hgrid%ahx_offd%n, hgrid%ahy_offd%n, hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, &
               hgrid%ahx_modl%n, hgrid%ahy_modl%n, hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n/)
    j_wind = (/1               , 1               , 1              , 1              , 1              , & 
               1               , 1               , 1              , 1              , 1              /)
    n_wind = (/1               , 1               , hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, &
               1               , 1               , hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n/)
    f_wind = (/0               , 0               , 0              , 0              , 0              , &
               0               , 0               , 0              , 0              , 0              /)

    if (flag == "rite") then
    !   err = srite_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else if (flag == "reed") then
       hess = 0.
       err = sreed_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    end if
    return
  end subroutine
!###################  
  subroutine hess_io_diag(hess, hgrid, file_, flag)
    real, dimension(:,:,:) :: hess
    type(hes_grid)         :: hgrid
    integer                :: ix_modl, iy_modl, iz_modl
    character(len=*)       :: file_
    character(len=4)       :: flag
    integer, dimension(10) :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 10
    n_cube = (/hgrid%ahx_offd%n, hgrid%ahy_offd%n, hgrid%ax_offd%n  , hgrid%ay_offd%n  , hgrid%az_offd%n  , &
               hgrid%ahx_modl%n, hgrid%ahy_modl%n, hgrid%ax_modl%n  , hgrid%ay_modl%n  , hgrid%az_modl%n  /)
    j_wind = (/1               , 1               , 1                , 1                , 1                , & 
               1               , 1               , 1                , 1                , 1                /)
    n_wind = (/1               , 1               , 1                , 1                , 1                , &
               1               , 1               , hgrid%ax_modl%n  , hgrid%ay_modl%n  , hgrid%az_modl%n  /)
    f_wind = (/0               , 0               , hgrid%ax_offd%n/2, hgrid%ay_offd%n/2, hgrid%az_offd%n-1, &
               0               , 0               , 0                , 0                , 0                /)
!    write(*,*)'--'
!    write(0,*)n_cube
!    write(0,*)n_wind
!    write(0,*)j_wind
!    write(0,*)f_wind

    if (flag == "rite") then
    !   err = srite_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else if (flag == "reed") then
       hess = 0.
!      write(0,*) 'In predonditioner ==> diagonal hessian....'
       err = sreed_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    end if
!    call auxinout('junk_')
!    call houaxis2('junk_', hgrid%az_modl, 1); call houaxis2('junk_', hgrid%ax_modl, 2); call houaxis2('junk_', hgrid%ay_modl, 3);
!    call sep_write(hess,"junk_",esize=4)
    return
  end subroutine
!###################  
  subroutine weight_io(weight, hgrid, file_, flag)
    real, dimension(:,:,:) :: weight
    type(hes_grid)         :: hgrid
    integer                :: ix_modl, iy_modl, iz_modl
    character(len=*)       :: file_
    character(len=4)       :: flag
    integer, dimension(3)  :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 3
    n_cube = (/hgrid%az_modl%n, hgrid%ax_modl%n, hgrid%ay_modl%n/)
    j_wind = (/1              , 1              , 1              /)
    n_wind = (/hgrid%az_modl%n, hgrid%ax_modl%n, hgrid%ay_modl%n/)
    f_wind = (/0              , 0              , 0              /)

    write(*,*)'--'
    write(0,*)n_cube
    write(0,*)n_wind
    write(0,*)j_wind
    write(0,*)f_wind
    if (flag == "rite") then
    !   err = srite_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,wieght)
    else if (flag == "reed") then
       weight = 0.
       write(0,*) 'Weights ==> ....'
       write(0,*)size(weight,1),size(weight,2), size(weight,3),file_
       err = sreed_window(file_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,weight)
    end if
    return
  end subroutine



end module

