module hessian_mod_convolve
  use hessian_convolve_init
  use hessian_convolve_module
  use hessian_convolve_module2
  use hessian_convolve_precond
  use hessian_convolve_axis
  use hessian_convolve_array
  use hessian_convolve_adjnull
  use hessian_convolve_add
  use hessian_convolve_regs
  use chain_mod
  implicit none

contains
!#######
  integer function conv_init() result(stat) 
        stat = 1
        call hessian_conv_params_init() 
        call hessian_conv_aux_init()
        call hessian_conv_axis_init()
        call hessian_conv_array_alloc()

        if (adj .or. inv) then
            if (inv) write(*,*)'inversion............................'
            if (adj) write(*,*)'adjoint only.........................'
            call hessian_conv_inout_init()
        else
            write(*,*)'modeling only.................................'
            call hessian_conv_inout_model_init()
        endif
        stat = 0.
  end function
!#######
  integer function conv_clos() result(stat)
        stat = 1
        call  hessian_conv_array_clos()
        stat = 0
  end function
!#######
  integer function conv_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,    modl,data)
    if(method==1)then
        call hessian_convolve    (adj,add          )
    else
        call hessian_convolve2   (adj,add          )
    endif
    call hessian_conv_add    (adj,add,modl,data)
    if (timing2)then
        if (iter==0)then
                call date_and_time(values=time_array_5)
                start_t1 = time_array_5 (5) * 3600 + time_array_5 (6) * 60  + time_array_5 (7) + 0.001 * time_array_5 (8)
                start_t0=start_t1
                finish_t0=start_t0
        else
                call date_and_time(values=time_array_6)
                finish_t1 = time_array_6 (5) * 3600 + time_array_6 (6) * 60  + time_array_6 (7) + 0.001 * time_array_6 (8)
                int_t = finish_t1-finish_t0
                tot_t = finish_t1-start_t0
                finish_t0=finish_t1
                write(0,*)'Time: Interv | Spent | Remain :',int_t,tot_t, int_t*(niter-iter-1)
        endif
   endif
    riter=iter/niter; 
    iter=iter+1
    stat = 0
  end function
!#######
  integer function conv_precon_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:):: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,    modl,data)
    call hessian_precond     (adj,add          )
    call hessian_conv_add    (adj,add,modl,data)
    stat = 0
  end function
!#######
  integer function conv_grad_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call hessian_conv_adjnull  (adj,     modl,data)
    call hessian_conv_gradient (adj               )
    call hessian_conv_add      (adj, add,modl,data)
    stat = 0
  end function
!#######
  integer function conv_temp_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call hessian_conv_adjnull  (adj,    modl,data)
    call hessian_conv_grad_temp(adj              )
    call hessian_conv_add      (adj,add,modl,data)
    stat = 0
  end function
!#######
  integer function conv_grad_temp_oper(adj, add, modl, data) result(stat)
    logical                     :: adj, add
    real, dimension(:)          :: modl, data
    real, dimension(nmod*nsurv) :: tmp1
    stat = 1
    call chain2 (oper1=conv_grad_oper, oper2=conv_temp_oper, adj=adj, add=.true., mod=modl, dat=data, tmp1=tmp1)
    stat = 0
  end function
!#######
  integer function conv_temp_grad_oper(adj, add, modl, data) result(stat)
    logical                     :: adj, add
    real, dimension(:)          :: modl, data
    real, dimension(nmod*nsurv) :: tmp1
    stat = 1
    call chain2 (oper1=conv_temp_oper, oper2=conv_grad_oper, adj=adj, add=.true., mod=modl, dat=data, tmp1=tmp1)
    stat = 0
  end function
!#######
!#######
  integer function conv_weight_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,modl,data)
    call hessian_conv_weight (adj,modl,data)
    call hessian_conv_add    (adj,add,modl,data)
    stat = 0
  end function
!#######
  integer function conv_mask_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call hessian_conv_adjnull(adj,modl,data)
    call hessian_conv_mask   (adj,modl,data)
    call hessian_conv_add    (adj,add,modl,data)
    stat = 0
  end function
!#####################################
  subroutine hessian_conv_weight(adj, modl, data)
    logical                     :: adj
    real, dimension(:)          :: modl, data
    if ( adj) then
                modl = modl + data*wght
    else
                data = data + modl*wght
    end if
  end subroutine
!#####################################
  subroutine hessian_conv_mask(adj, modl, data)
    logical                     :: adj
    real, dimension(:)          :: modl, data
    if ( adj) then
        where( mask==1)
                modl = modl + data
        end where
    else
        where( mask==1)
                data = data + modl
        end where
    end if
  end subroutine

!#####################################

end module

