module history_mod
  use global_params_mod
  use reflect_mod
  use map_mod, only: data_radians_calc, gre_radians_calc 

  implicit none


contains
!==========================================================================================
 subroutine ref_input_history()
    integer             :: num_match
    character(len=1024) :: string
    ref_= "ref"; 
    !read the velocity from file   
    call auxin(ref_)
    call ainaxis(ref_, rgrid%az,1); call ainaxis(ref_, rgrid%ax,2); call ainaxis(ref_,rgrid%ay,3)
    write(0,*)'Reflectivity Model summary'
    write(0,*)'=========z x y============'
    write(0,*)rgrid%az
    write(0,*)rgrid%ax
    write(0,*)rgrid%ay
    write(0,*)'======================'
  end subroutine
!====================================
  subroutine vel_ref_input_history()
    integer             :: num_match
    character(len=1024) :: string
    vel_= "vel"; ref_ = "ref"
    call from_param("for" , for , .false. ) 

    !read the velocity from file   
    call auxin(vel_)
    call from_aux(vel_, "n1", vgrid%ax%n); call from_aux(vel_, "d1", vgrid%ax%d); call from_aux(vel_, "o1", vgrid%ax%o) 
    call from_aux(vel_, "n2", vgrid%ay%n); call from_aux(vel_, "d2", vgrid%ay%d); call from_aux(vel_, "o2", vgrid%ay%o)
    call from_aux(vel_, "n3", vgrid%az%n); call from_aux(vel_, "d3", vgrid%az%d); call from_aux(vel_, "o3", vgrid%az%o)

    !read the reflectivity from file if provided...if not generate reflectivity
    num_match =  exist_file(ref_)
    if ( num_match==1 .or. for ) then
        call auxin(ref_)
        call from_aux(ref_, "n1", rgrid%ax%n); call from_aux(ref_, "d1", rgrid%ax%d); call from_aux(ref_, "o1", rgrid%ax%o) 
        call from_aux(ref_, "n2", rgrid%ay%n); call from_aux(ref_, "d2", rgrid%ay%d); call from_aux(ref_, "o2", rgrid%ay%o)
        call from_aux(ref_, "n3", rgrid%az%n); call from_aux(ref_, "d3", rgrid%az%d); call from_aux(ref_, "o3", rgrid%az%o)

        if (rgrid%az%n /= vgrid%az%n .or. rgrid%ax%n /= vgrid%ax%n .or. rgrid%ay%n /= vgrid%ay%n) &
        call seperr("reflectivity and velocity grids must match...see vel_ref_input_history()")
    else
        if (.not. exist_file(ref_)) call reflec(vel, ref, ref_, vgrid, den) !generate reflectivity model
        rgrid%az=vgrid%az; rgrid%ax=vgrid%ax; rgrid%ay=vgrid%ay
    end if
    write(*,*)rgrid%az%o, rgrid%az%n, rgrid%az%d
    write(*,*)vgrid%az%o, vgrid%az%n, vgrid%az%d

  end subroutine
!==========================================================================================
  subroutine data_output_history()
        if (lfreq)then
                write(*,*)'data===========>       freq!'
                call auxinout(data_)
                call houaxis(data_,dgrid%agx,1);call houaxis(data_,dgrid%agy,2); call houaxis(data_,dgrid%aw,5); call houaxis(data_,dgrid%asx,3); call houaxis(data_,dgrid%asy,4);
                call auxputch("label1", "s", "geoph_x", data_)
                call auxputch("label2", "s", "geoph_y", data_)
                call auxputch("label5", "s", "freq.  ", data_)
                call auxputch("label3", "s", "shot_x", data_)
                call auxputch("label4", "s", "shot_y", data_)
                call to_history('esize', 8, data_)

        else
                call auxinout(data_)
                call houaxis(data_,dgrid%agx,1);call houaxis(data_,dgrid%agy,2); call houaxis(data_,dgrid%at,5); call houaxis(data_,dgrid%asx,3); call houaxis(data_,dgrid%asy,4);
                call auxputch("label1", "s", "geoph_x", data_)
                call auxputch("label2", "s", "geoph_y", data_)
                call auxputch("label5", "s", "time.  ", data_)
                call auxputch("label3", "s", "shot_x ", data_)
                call auxputch("label4", "s", "shot_y ", data_)
                call to_history('esize', 4, data_)
                write(*,*)'data=========>       time'
        endif
    	if (.not. radian) call data_radians_calc()
  end subroutine
!#####################################
  subroutine mig_output_history()
                call houaxis(mig_,igrid%ahx,1);call houaxis(mig_,igrid%ahy,2); call houaxis(mig_,igrid%ax,3); call houaxis(mig_,igrid%ay,4); call houaxis(mig_,igrid%az,5);
                call auxputch("label1", "s", "hx", mig_); call auxputch("label2", "s", "hy", mig_); call auxputch("label3", "s", "x" , mig_); call auxputch("label4", "s", "y" , mig_); call auxputch("label5", "s", "z" , mig_)
                if (nostack_w .and. nostack_s)then
                    call houaxis(mig_, dgrid%aw, 6); call auxputch("label6", "s", "w", mig_)
                    call houaxis(mig_, dgrid%aw, 7); call auxputch("label7", "s", "s", mig_)
                elseif (nostack_w) then; 
                    call houaxis(mig_, dgrid%aw, 6); call auxputch("label6", "s", "w", mig_)
                elseif(nostack_s) then; 
                    call houaxis(mig_, dgrid%aw, 6); call auxputch("label6", "s", "s", mig_)
                else
                endif
                call to_history('esize', 4, mig_)
  end subroutine
!#####################################
 subroutine green_output_history()
!        if(loop=='s')then
!                call houaxis(greens_,ggrid%ax_surf,5); call houaxis(greens_,ggrid%ay_surf,6); call houaxis(greens_,ggrid%ax_modl,1); call houaxis(greens_,ggrid%ay_modl,2); call houaxis(greens_,ggrid%az_modl,3); call houaxis(greens_,ggrid%aw,4);  
!                call auxputch("esize" , "i",         8   , greens_)
!                call auxputch("label5", "s", "surface_x" , greens_)
!                call auxputch("label6", "s", "surface_y" , greens_)
!                call auxputch("label1", "s", "model_x"   , greens_)
!                call auxputch("label2", "s", "model_y"   , greens_)
!                call auxputch("label3", "s", "model_z"   , greens_)
!                call auxputch("label4", "s", "frequency" , greens_)
!        else
                call houaxis(greens_,ggrid%ax_surf,1); call houaxis(greens_,ggrid%ay_surf,2); call houaxis(greens_,ggrid%ax_modl,3); call houaxis(greens_,ggrid%ay_modl,4); call houaxis(greens_,ggrid%az_modl,5); call houaxis(greens_,ggrid%aw,6);  
                call auxputch("esize" , "i",         8   , greens_)
                call auxputch("label1", "s", "surface_x" , greens_)
                call auxputch("label2", "s", "surface_y" , greens_)
                call auxputch("label3", "s", "model_x"   , greens_)
                call auxputch("label4", "s", "model_y"   , greens_)
                call auxputch("label5", "s", "model_z"   , greens_)
                call auxputch("label6", "s", "frequency" , greens_)
!        endif
    if (.not. radian) call gre_radians_calc()
  end subroutine
!##################################################
  subroutine wfld_output_history(wfld_)
        character(len=256)        :: wfld_
        call auxinout(wfld_d_)
        call auxputch("n1","i",vgrid%ax%n,wfld_); call auxputch("n2","i",vgrid%ay%n,wfld_); call auxputch("n3","i",vgrid%az%n,wfld_); call auxputch("n4","i",dgrid%at%n,wfld_);
        call auxputch("d1","r",vgrid%ax%d,wfld_); call auxputch("d2","r",vgrid%ay%d,wfld_); call auxputch("d3","r",vgrid%az%d,wfld_); call auxputch("d4","r",dgrid%at%d,wfld_);
        call auxputch("o1","r",vgrid%ax%o,wfld_); call auxputch("o2","r",vgrid%ay%o,wfld_); call auxputch("o3","r",vgrid%az%o,wfld_); call auxputch("o4","r",dgrid%at%o,wfld_); 
        call auxputch("label1","s","nx",wfld_)
        call auxputch("label2","s","ny",wfld_)
        call auxputch("label3","s","nz" ,wfld_)
        call auxputch("label4","s","t" ,wfld_)
        call auxputch("esize" ,"i", 4  ,wfld_)
  end subroutine


  end module
