!OW mig io
module io_mod
  use global_params_mod
  implicit none

contains
!#####################################################
  subroutine data_io_init()
    complex, allocatable, dimension(:) :: ctmp
    integer                            :: i, nloop
    write(*,*)"Initializing the born data"
    if(lfreq)then
        allocate(ctmp(dgrid%agx%n*dgrid%agy%n*dgrid%asx%n*dgrid%asy%n))
        ctmp = (0,0)
        nloop = dgrid%aw%n
        do i=1, nloop
                call sep_write(ctmp, data_, esize=8)
        end do
    else
        allocate(ctmp(dgrid%at%n))
        ctmp = 0
        nloop = dgrid%agx%n*dgrid%agy%n*dgrid%asx%n*dgrid%asy%n
        do i=1, nloop
                call sep_write(ctmp, data_, esize=4)
        end do
    endif
    deallocate(ctmp)

  end subroutine
!#####################################################
  subroutine mig_io_init()
    real, allocatable, dimension(:) :: ctmp
    integer                            :: i, nloop
        allocate(ctmp(igrid%ax%n*igrid%ay%n*igrid%ahx%n*igrid%ahy%n))
        ctmp = (0,0)
        if (nostack_w .and. nostack_s)then
               nloop = igrid%az%n*dgrid%aw%n*dgrid%asx%n
        elseif (nostack_w)then
               nloop = igrid%az%n*dgrid%aw%n
        elseif (nostack_s)then
               nloop = igrid%az%n*dgrid%asx%n
        else
               nloop = igrid%az%n
        endif
        do i=1, nloop
                call sep_write(ctmp, mig_, esize=4)
        end do
        deallocate(ctmp)
  end subroutine
!#####################################################
  subroutine greens_io_init()
    complex, allocatable, dimension(:) :: ctmp
    integer                            :: i, nloop
        allocate(ctmp(ggrid%ax_surf%n*ggrid%ay_surf%n*ggrid%ay_modl%n*ggrid%aw%n*max(ggrid%az_modl%n,ggrid%ax_modl%n)))
        ctmp = (0,0)
        nloop = min(ggrid%az_modl%n,ggrid%ax_modl%n)
        write(*,*)"Initializing the greens functions......"
        do i=1, nloop
                if(modulo(i,50)==0)write(0,*)'done:', i,'of ', nloop
                call sep_write(ctmp, greens_, esize=8)	
        end do
        deallocate(ctmp)
  end subroutine
!#####################################################
  subroutine wfld_io_init(wfld_)
    complex, allocatable, dimension(:) :: ctmp
    integer                            :: i, nloop
    character(len=256)                 :: wfld_

    allocate(ctmp(vgrid%ax%n*vgrid%ay%n*dgrid%at%n))
    ctmp = (0,0)
    nloop = vgrid%az%n
    write(*,*)"Initializing the wfld_d"
    do i=1, nloop
!            call sep_write(ctmp, wfld_, esize=4)
    end do
    deallocate(ctmp)
  end subroutine

!#####################################################
subroutine data_io_data(flag, data, index_aw )
    complex, dimension(:,:,:,:  ), allocatable :: data
    character(len=4)                    :: flag
    integer                             :: index_aw
    integer, dimension(5)               :: n_cube, j_wind, n_wind, f_wind
    integer                             :: n_dim_cube, err
    character :: y

    n_dim_cube = 5
    n_cube = (/dgrid%agx%n , dgrid%agy%n   , dgrid%aw%n    , dgrid%asx%n    ,  dgrid%asy%n/)
    j_wind = (/1           , 1             , 1             , 1              ,           1/)
    n_wind = (/dgrid%agx%n , dgrid%agy%n   , 1             , dgrid%asx%n    ,  dgrid%asy%n/)
    f_wind = (/0           , 0             , index_aw-1    , 0              ,  0          /)
    !write(0,*)n_cube; write(0,*)n_wind; write(0,*)j_wind; write(0,*)f_wind; 
    !write(0,*)index_aw; write(0,*)size(data,1),size(data,2),size(data,3),size(data,4)

    if (flag == "rite") then
     !  write(0,*)'writing data frequency: ', index_aw
       err = srite_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,data)
    else if (flag == "reed") then
       data = 0.
!      write(0,*)'reading data frequency: ', index_aw
       err = sreed_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,data)
    else
       call seperr("you have to specify the flag for data_io_data()")
    end if
    if (err /= 0) call seperr("data_io_data: error reading data")
  end subroutine
!#####################################################
subroutine data_1w_io(flag, data, index_sx, index_sy )
    complex, dimension(:,:,:,:  ), allocatable :: data
    character(len=4)                    :: flag
    integer                             :: index_sx, index_sy
    integer, dimension(5)               :: n_cube, j_wind, n_wind, f_wind
    integer                             :: n_dim_cube, err
    character :: y

    n_dim_cube = 5
    n_cube = (/dgrid%agx%n , dgrid%agy%n   , dgrid%aw%n    , dgrid%asx%n    ,  dgrid%asy%n/)
    j_wind = (/1           , 1             , 1             , 1              ,           1/)
    n_wind = (/dgrid%agx%n , dgrid%agy%n   , dgrid%aw%n    , 1              ,           1/)
    f_wind = (/0           , 0             , 0             , index_sx-1     ,  index_sy-1/)

    !write(0,*)n_cube; write(0,*)n_wind; write(0,*)j_wind; write(0,*)f_wind; 
    !write(0,*)index_sx, index_sy; write(0,*)size(data,1),size(data,2),size(data,3),size(data,4)
    if (flag == "rite") then
       err = srite_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,data)
    else if (flag == "reed") then
       data = 0.
       err = sreed_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,data)
    else
       call seperr("you have to specify the flag for data_io_data()")
    end if
    if (err /= 0) call seperr("data_io_data: error reading data")
  end subroutine
!#####################################################
subroutine data_io_fdata(flag, fdata)
    complex, dimension(:,:,:,:,:), allocatable 	:: fdata
    character(len=4)                    	:: flag
    integer, dimension(5)               	:: n_cube, j_wind, n_wind, f_wind
    integer                             	:: n_dim_cube, err
    character :: y

    n_dim_cube = 5
    n_cube = (/dgrid%agx%n , dgrid%agy%n   , dgrid%asx%n   , dgrid%asy%n   ,dgrid%aw%n  /)
    j_wind = (/1           , 1             , 1             , 1             ,           1/)
    n_wind = (/dgrid%agx%n , dgrid%agy%n   , dgrid%asx%n   , dgrid%asy%n   ,dgrid%aw%n  /)
    f_wind = (/0           , 0             , 0             , 0             ,	       0/)

    if (flag == "rite") then
       err = srite_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,fdata)
    else if (flag == "reed") then
       data = 0.
       err = sreed_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,fdata)
    else
       call seperr("you have to specify the flag for data_io_data()")
    end if
  end subroutine
!#####################################################
subroutine data_io_tdata(flag, fdata)
    use myfft_mod
    complex, dimension(:,:,:,:,:), allocatable 	:: fdata
    complex, dimension(size(fdata,5))		:: tmp
    character(len=4)                    	:: flag
    integer, dimension(5)               	:: n_cube, j_wind, n_wind, f_wind
    integer                             	:: n_dim_cube, err
    character :: y

    n_dim_cube = 5
    n_cube = (/dgrid%agx%n , dgrid%agy%n   , dgrid%asx%n   , dgrid%asy%n   ,dgrid%at%n  /);
    j_wind = (/1           , 1             , 1             , 1             ,           1/);
    n_wind = (/dgrid%agx%n , dgrid%agy%n   , dgrid%asx%n   , dgrid%asy%n   ,dgrid%at%n  /);
    f_wind = (/0           , 0             , 0             , 0             ,	       0/);

        do isy=1, dgrid%asy%n;         
        do isx=1, dgrid%asx%n;         
                do igy=1, dgrid%agy%n; 
                do igx=1, dgrid%agx%n; 
                        call FFT1DT(fdata(dgrid%agx%n, dgrid%agy%n, dgrid%asx%n, dgrid%asy%n, :), 1);
                end do; 
                end do; 
        end do; 
        end do;   

    if (flag == "rite") then
       err = srite_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,real(fdata))
    else if (flag == "reed") then
       data = 0.
       err = sreed_window(data_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,real(fdata))
    else
       call seperr("you have to specify the flag for data_io_data()")
    end if
  end subroutine

!#####################################################
  subroutine data_io_greens(greens, index_ws)
    integer                                   :: index_ws
    integer, dimension(6)                     :: n_cube, j_wind, n_wind, f_wind
    integer                                   :: n_dim_cube, err
    complex, dimension(:,:,:,:,:), allocatable:: greens

    n_dim_cube = 6
    n_cube = (/ggrid%az_modl%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%aw%n/)
    j_wind = (/1              , 1              , 1              , 1              , 1              , 1         /)
    n_wind = (/ggrid%az_modl%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%ax_surf%n, ggrid%ay_surf%n, 1         /)
    f_wind = (/0              , 0              , 0              , 0              , 0              , index_ws-1/)
    write(0,*)'writing greens freq.: ', index_ws, 'of ', ggrid%aw%n, 'at node', omp_get_thread_num();  
    err = srite_window(greens_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,greens)
    if (err /= 0) call seperr("data_io_greens: error writing wavefield")

  end subroutine
!#####################################################
  subroutine green_io_greens(greens, index_ws)
    integer                                   :: index_ws
    integer, dimension(6)                     :: n_cube, j_wind, n_wind, f_wind
    integer                                   :: n_dim_cube, err
    complex, dimension(:,:,:,:,:), allocatable:: greens

    n_dim_cube = 6
    n_cube = (/ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%aw%n/)
    j_wind = (/1              , 1              , 1              , 1              , 1              , 1         /)
    n_wind = (/ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, 1         /)
    f_wind = (/0              , 0              , 0              , 0              , 0              , index_ws-1/)
    write(0,'(A25,I5,A4,I5,A8,I4,2F15.5)')'writing greens freq.: ', index_ws, 'of ', ggrid%aw%n, 'at node', omp_get_thread_num(),sum(greens);  
!    write(0,*)n_cube; write(0,*)n_wind; write(0,*)j_wind; write(0,*)f_wind,size(greens),sum(greens);
!    write(0,*)size(greens,1),size(greens,2),size(greens,3),size(greens,4),size(greens,5)
    err = srite_window(greens_, n_dim_cube, n_cube, n_wind, f_wind, j_wind, 8, greens)
!   call srite(greens_, greens, 8*size(greens))
    if (err /= 0) call seperr("data_io_greens: error writing wavefield")
  end subroutine
!#####################################################
  subroutine green_io_greens2(greens, igx, igy)
    integer                                   :: igx,igy
    integer, dimension(6)                     :: n_cube, j_wind, n_wind, f_wind
    integer                                   :: n_dim_cube, err
    complex, dimension(:,:,:,:,:), allocatable:: greens

    n_dim_cube = 6
    n_cube = (/ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%aw%n ,ggrid%ax_surf%n, ggrid%ay_surf%n/)
    j_wind = (/1              , 1              , 1              , 1          , 1              , 1             /)
    n_wind = (/ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%aw%n , 1              , 1             /)
    f_wind = (/0              , 0              , 0              , 0          , igx-1          , igy-1         /)
    write(0,*)'writing greens igx/igy.: ', igx,igy, '@ node', omp_get_thread_num();
    !write(0,*)size(greens,1),size(greens,2),size(greens,3),size(greens,4),size(greens,5) 
    !write(0,*)n_cube; write(0,*)n_wind; write(0,*)j_wind; write(0,*)f_wind;
    err = srite_window(greens_, n_dim_cube, n_cube, n_wind, f_wind, j_wind, 8, greens)
    if (err /= 0) call seperr("data_io_greens: error writing wavefield")
  end subroutine
!#####################################################
  subroutine green_io_greens3(greens, index_gy, index_ws)
    integer                                   :: index_gy,index_ws
    integer, dimension(6)                     :: n_cube, j_wind, n_wind, f_wind
    integer                                   :: n_dim_cube, err
    complex, dimension(:,:,:,:,:), allocatable:: greens

    n_dim_cube = 6
    n_cube = (/ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%aw%n/)
    j_wind = (/1              , 1              , 1              , 1              , 1              , 1         /)
    n_wind = (/ggrid%ax_surf%n, 1              , ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, 1         /)
    f_wind = (/0              , index_gy-1     , 0              , 0              , 0              , index_ws-1/)
    write(0,'(A25,I5,A4,I4,A10,I4,A4,I4,A8,I4,2F15.5)')'writing greens iw.: ', index_ws, 'of ', ggrid%aw%n, 'xline.: ', index_gy, 'of ', ggrid%ay_surf%n, ' | At node', omp_get_thread_num(),sum(greens);  
!    write(0,*)n_cube; write(0,*)n_wind; write(0,*)j_wind; write(0,*)f_wind,size(greens),sum(greens);
!    write(0,*)size(greens,1),size(greens,2),size(greens,3),size(greens,4),size(greens,5)
    err = srite_window(greens_, n_dim_cube, n_cube, n_wind, f_wind, j_wind, 8, greens)
!   call srite(greens_, greens, 8*size(greens))
    if (err /= 0) call seperr("data_io_greens: error writing wavefield")
  end subroutine

!#####################################################
  subroutine mig_io_mig(mig, index_ws)
    integer                                   :: index_ws
    integer, dimension(4)                     :: n_cube, j_wind, n_wind, f_wind
    integer                                   :: n_dim_cube, err
    real   , dimension(:,:,:), allocatable    :: mig

    n_dim_cube = 4
    n_cube = (/igrid%az%n, igrid%ax%n, igrid%ay%n, unw         /)
    j_wind = (/1         , 1         , 1         , 1           /)
    n_wind = (/igrid%az%n, igrid%ax%n, igrid%ay%n, 1           /)
    f_wind = (/0         , 0         , 0         , index_ws-1  /)

    write(0,*)'writing migration freq.: ', index_ws, 'of ', unw, 'at node', omp_get_thread_num();  
    err = srite_window(mig_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,mig)
    if (err /= 0) call seperr("mig_io_mig: error writing mig")

  end subroutine
!#####################################################
  subroutine mig_io(mig)
    integer                                     :: index_ws
    integer, dimension(5)                       :: n_cube, j_wind, n_wind, f_wind
    integer                                     :: n_dim_cube, err
    real   , dimension(:,:,:,:,:), allocatable  :: mig

    n_dim_cube = 5
    n_cube = (/igrid%ahx%n, igrid%ahy%n , igrid%ax%n              , igrid%ay%n            , igrid%az%n             /)
    n_wind = (/igrid%ahx%n, igrid%ahy%n , igrid%ax%n              , igrid%ay%n            , igrid%az%n             /)
    j_wind = (/1          , 1           , 1                       , 1                     , 1                      /)
    f_wind = (/0          , 0           , 0                       , 0                     , 0                      /)
!    write(0,*)'writing migration freq.: ', index_ws, 'of ', unw, 'at node', omp_get_thread_num();  
    err = srite_window(mig_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,mig)
    if (err /= 0) call seperr("mig_io_mig: error writing mig")

  end subroutine

!#####################################################
  subroutine data_io_wfld(wfld_, wfld, index_z)
    integer                                :: index_z
    integer, dimension(4)                  :: n_cube, j_wind, n_wind, f_wind
    integer                                :: n_dim_cube, err
    real , dimension(:,:,:,:), allocatable :: wfld
    character(len=256)                     :: wfld_

    n_dim_cube = 4
    n_cube = (/vgrid%az%n, vgrid%ax%n, vgrid%ay%n, dgrid%at%n/wfld_out_jt /)
    j_wind = (/1         , 1         , 1         , 1                      /)
    n_wind = (/vgrid%az%n, vgrid%ax%n, vgrid%ay%n, dgrid%at%n/wfld_out_jt /)
    f_wind = (/0         , 0        ,  0         , 0                      /)

    wfld = 0.
    err = srite_window(wfld_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,wfld)
    if (err /= 0) call seperr("data_io_greens: error reading Green's function")
  end subroutine

!#####################################################
  subroutine data_io_vel(vel, index_z)
    integer                              :: index_z
    integer, dimension(3)                :: n_cube, j_wind, n_wind, f_wind
    real   , dimension(:,:), allocatable :: vel
    integer                              :: n_dim_cube, err

    n_dim_cube = 3
    n_cube = (/vgrid%az%n, vgrid%ax%n        , vgrid%ay%n        /)
    j_wind = (/1         , 1                 , 1                 /)
    n_wind = (/1         , pgrid%ax%n        , pgrid%ay%n        /)
    f_wind = (/index_z-1 , pgrid%index_xmin-1, pgrid%index_ymin-1/)

    vel = 0.
    err = sreed_window(vel_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,vel)
    if (err /= 0) call seperr("data_io_vel: error reading vel slice")
  end subroutine
!#####################################################
!#####################################################
  subroutine data_io_ref(ref, index_z)
    integer                              :: index_z
    integer, dimension(3)                :: n_cube, j_wind, n_wind, f_wind
    real   , dimension(:,:), allocatable :: ref
    integer                              :: n_dim_cube, err

    n_dim_cube = 3
    n_cube = (/rgrid%az%n, rgrid%ax%n        , rgrid%ay%n        /)
    j_wind = (/1         , 1                 , 1                 /)
    n_wind = (/1         , pgrid%ax%n        , pgrid%ay%n        /)
    f_wind = (/index_z-1 , pgrid%index_xmin-1, pgrid%index_ymin-1/)

    ref = 0.
    err = sreed_window(ref_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,ref)
    if (err /= 0) call seperr("data_io_ref: error reading ref slice")
  end subroutine

!#####################################################
 subroutine wfld_out(wfld,wfld_)
        use myfft_mod
        complex    , dimension(:,:,:,:,:), allocatable     :: wfld
        complex    , dimension(:      ), allocatable :: tmp_wfld
        real       , dimension(:,:,:,:), allocatable :: tmp_wfld2
        real       , dimension(:,:,:,:), allocatable :: tmp_wfld3
        integer                                         :: nshare
        character (len=256) :: wfld_

        if (allocated(tmp_wfld )) deallocate (tmp_wfld ); allocate(tmp_wfld(dgrid%at%n))
        if (allocated(tmp_wfld2)) deallocate (tmp_wfld2); allocate(tmp_wfld2(vgrid%ax%n,vgrid%ay%n,vgrid%az%n,dgrid%at%n)) 
write(0,*)sum(wfld),'yyyyyyy'


        nnode =omp_get_num_procs(); thread=omp_get_thread_num(); 
        nodes=floor(nnode*.98)
        call omp_set_num_threads(nodes); write(*,*) 'using', nodes, 'of' , nnode
        nshare=nint(1.*vgrid%az%n/nodes)
        !$OMP parallel do private(it,itt,ix,iz,inode,tmp_wfld) schedule(dynamic,1) 
        do iz=1,vgrid%az%n
                if (allocated(tmp_wfld )) deallocate (tmp_wfld ); allocate(tmp_wfld(dgrid%at%n))
                inode=omp_get_thread_num();  
                !if(verb>1) write(*,*)'iz.:', iz ,' of ',vgrid%az%n,' @ node ', inode         
                write(*,*)'iz.:', iz ,' of ',vgrid%az%n,' @ node ', inode         
                do iy=1,vgrid%ay%n 
                do ix=1,vgrid%ax%n 
                        tmp_wfld=wfld(ix,iy,iz,:,1); call FFT1DT(tmp_wfld, -1); tmp_wfld2(ix,iy,iz,:)=tmp_wfld
                enddo
                enddo
        enddo
        !$OMP end parallel do
        itt=1;


        call sep_write(tmp_wfld2, wfld_, esize=4)
        write(*,*)'written wavefields for one shot successfully'
        deallocate(tmp_wfld,tmp_wfld2)
 end subroutine
!#####################################################
 subroutine mig_velocity()
           if(lloyd)then
                call sreed(vel_, vel, 4*vgrid%ax%n*vgrid%ay%n*vgrid%az%n)
                if(.not. exist_file(vel_ref_)) call seperr('Provide file for refernce velocities...');
                call from_aux(vel_ref_,"n1",vnum);
                call sreed(vel_ref_, vel_ref, 4*vnum*vgrid%az%n*2)
                nref=0;
                do iz=1,vgrid%az%n
                        do iv=1,vnum; if(vel_ref(iv,iz,1)>1e-05)then; nref(iz)=nref(iz)+1; else; exit; endif; enddo
                enddo
           else
                call sreed(vel_, vel, 4*vgrid%ax%n*vgrid%ay%n*vgrid%az%n)
                write(0,*)lloyd
                stop
           endif
  end subroutine
!#####################################################
  subroutine green_velocity()
           if(lloyd)then
                call sreed(vel_, vel, 4*vgrid%ax%n*vgrid%ay%n*vgrid%az%n)
                if(.not. exist_file(vel_ref_)) call seperr('Provide file for refernce velocities...');
                call from_aux(vel_ref_,"n1",vnum);
                call sreed(vel_ref_, vel_ref, 4*vnum*vgrid%az%n*2)
                nref=0;
                do iz=1,vgrid%az%n
                        do iv=1,vnum; if(vel_ref(iv,iz,1)>1e-05)then; nref(iz)=nref(iz)+1; else; exit; endif; enddo
                enddo
           else
                call sreed(vel_, vel, 4*vgrid%ax%n*vgrid%ay%n*vgrid%az%n)
           endif
	   if(ldata)call sreed(ref_, refl, 4*vgrid%ax%n*vgrid%ay%n*vgrid%az%n)
  end subroutine
!#####################################################
  subroutine data_io_csou(csou)
    integer, dimension(3)                 :: n_cube, j_wind, n_wind, f_wind
    complex, dimension(:,:,:),allocatable :: csou
    integer                               :: n_dim_cube, err

    n_dim_cube = 3
    n_cube = (/wgrid%asx%n, wgrid%asy%n       , wgrid%aw%n      /)
    j_wind = (/1          , 1                 , 1               /)
    n_wind = (/wgrid%asx%n, wgrid%asy%n       , wgrid%aw%n      /)
    f_wind = (/0          , 0                 , 0               /)

    csou = 0.
    err = sreed_window(wavelet_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,csou)
    if (err /= 0) call seperr("data_io_csou: error reading souerce header")
  end subroutine

  end module
