 module joint_encode_hlx
! use joint_encode_params
! use joint_encode_adjnull
! use joint_encode_rows
 use joint_encode_prc
 use createhelix_mod
 use cartesian_mod
 use bound_mod
 use regrid_mod
 use adj_mod
 use lapfac_mod
 use polydiv_mod
 use helicon_mod
 use helix_mod


 implicit none
	integer, private, save:: stat	
        type( filter),   private                        :: aa 
        type( filter),   private                        :: bb 
        real ,private,  dimension (:), allocatable      :: tt 
        integer, private                                :: nd 

 contains 
!#####################################################  
        subroutine filter_init_1d_x();  call allocatehelix (aa1,filt); call boundn1 ((/vgrid%ax%n/),(/vgrid%ax%n/), (/ filt /), aa1); aa1%flt = -1.; aa1%lag = 1; end subroutine
        subroutine filter_init_1d_z();  call allocatehelix (aa1,filt); call boundn1 ((/vgrid%az%n/),(/vgrid%az%n/), (/ filt /), aa1); aa1%flt = -1.; aa1%lag = 1; end subroutine
!#####################################################  
    subroutine filter_init()
        integer,dimension(:),allocatable    ::nd,na,center,gap
        real   ,dimension(:),allocatable    ::cube
        integer                             ::stat    

        allocate(na(2),center(2),nd(2),gap(2)) 
        na(1)=2                 ; na(2)=2; 
        nd(1)=vgrid%ax%n        ; nd(2)=vgrid%az%n; 
        center(1)=na(1)/2       ; center(2)=1; 
        gap=0
        aa1=createhelix(nd, center, gap, na)
        if (hlx_conv)then
                call helixcon_init(aa1)		     
        else
                call polyxdiv_init(nmod,aa1)
        endif
filt=3
        call boundn1 ((/vgrid%ax%n, vgrid%az%n/),(/vgrid%ax%n, vgrid%az%n/), (/ filt /), aa1)

        select case (filt)
        case (1)
                aa1%flt = (/ 0,0,-1/) 
        case (2)
                aa1%flt = (/ 0,0,-1/) 
        case (3)
                aa1%flt = (/ 0,0,-1/) 
        case (4)
                aa1%flt = (/ 0,0,-1/) 
        case (5)
                aa1%flt = (/  -1.858735,      0.8680543  /) 
        end select
        deallocate (na,center,nd,gap)
    end subroutine
!#####################################################  
        subroutine boundn1 ( nold, nd, na, aa)
                integer, dimension( :), intent( in) :: nold, nd, na     ! (ndim)
                type( filter)                       :: aa
                integer, dimension( size( nd))      :: nb, ii
                real,    dimension( :), allocatable :: xx, yy
                integer                             :: iy, my, ib, mb, stat
                        nb = nd + 2*na
                        mb = product( nb)                               ! nb is a bigger space to pad into.
                        allocate( xx( mb), yy( mb))		        ! two large spaces, equal size
                        xx = 0.				                ! zeros
                        do ib = 1, mb                                   ! surround the zeros with many ones
                                call line2cart( nb, ib, ii)             ! ii( ib)
                                if ( any( ii <= na  .or.  ii > nb-na)) xx( ib) = 1. 
                        end do 
                        call helixcon_init( aa)		                ! give aa pointer to helicon.lop
                        call regridn( nold, nb, aa)
                        aa%flt = 1.		                        ! put all 1's in filter
                        stat = helixcon_lop( .false., .false., xx, yy)	! apply filter
                        call regridn(   nb, nd, aa)
                        aa%flt = 0.	                                ! remake filter for orig data.
                        my = product( nd)
                        allocate( aa%mis( my))		                ! attach missing designation to y_filter
                        do iy = 1, my                                   ! map from unpadded to padded space
                                call line2cart( nd, iy, ii )
                                call cart2line( nb,     ii+na, ib )	! ib( iy)
                                aa%mis( iy) = ( yy( ib) > 0.)           ! true where inputs missing
                        end do 
                deallocate( xx, yy)
        end subroutine 
!#####################################################  conv or deconv
        integer function hlx_it(adj, add, xx, yy) result(stat)
                logical,intent(in) :: adj,add 
                real, dimension (:)  :: xx, yy 
                call joint_adjnull2 (adj,add, xx,yy)
                if (hlx_conv==1)then
                        if (nsurv > 5)call helixcon_lop2(adj,add,m6,d6); if (nsurv > 4)call helixcon_lop2(adj,add,m5,d5); if (nsurv > 3)call helixcon_lop2(adj,add,m4,d4);                                                           
                        if (nsurv > 2)call helixcon_lop2(adj,add,m3,d3); if (nsurv > 1)call helixcon_lop2(adj,add,m2,d2); if (nsurv > 0)call helixcon_lop2(adj,add,m1,d1);
                else
                        if (nsurv > 5)call polyxdiv_lop2(adj,add,m6,d6); if (nsurv > 4)call polyxdiv_lop2(adj,add,m5,d5); if (nsurv > 3)call polyxdiv_lop2(adj,add,m4,d4);                                                           
                        if (nsurv > 2)call polyxdiv_lop2(adj,add,m3,d3); if (nsurv > 1)call polyxdiv_lop2(adj,add,m2,d2); if (nsurv > 0)call polyxdiv_lop2(adj,add,m1,d1);
                endif
                stat=0
        end function
!#####################################################  conv or deconv
        integer function hlx_it2(adj, add, xx, yy) result(stat)
                logical,intent(in) :: adj,add 
                real, dimension (:)  :: xx, yy 
                call joint_adjnull2 (adj,add, xx,yy)
                if (hlx_conv==1)then
                        if (nsurv > 5)stat= helixcon_lop_2(adj,add,m6,d6); if (nsurv > 4)stat= helixcon_lop_2(adj,add,m5,d5); if (nsurv > 3)stat= helixcon_lop_2(adj,add,m4,d4);                                                           
                        if (nsurv > 2)stat= helixcon_lop_2(adj,add,m3,d3); if (nsurv > 1)stat= helixcon_lop_2(adj,add,m2,d2); if (nsurv > 0)stat= helixcon_lop_2(adj,add,m1,d1);
                else
                        if (nsurv > 5)stat= polyxdiv_lop_2(adj,add,m6,d6); if (nsurv > 4)stat= polyxdiv_lop_2(adj,add,m5,d5); if (nsurv > 3)stat= polyxdiv_lop_2(adj,add,m4,d4);                                                           
                        if (nsurv > 2)stat= polyxdiv_lop_2(adj,add,m3,d3); if (nsurv > 1)stat= polyxdiv_lop_2(adj,add,m2,d2); if (nsurv > 0)stat= polyxdiv_lop_2(adj,add,m1,d1);
                endif
                stat=0
        end function
!#####################################################  
! Convolution, inverse to deconvolution.
        subroutine helixcon_init ( aa_in )
                type( filter)    :: aa_in 
                aa = aa_in
        end subroutine
        function helixcon_lop( adj, add, xx, yy) result(stat)
                integer                   :: stat 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                        call adjnull(adj,add,xx,yy)
                        call helixcon_lop2(adj,add, xx,yy)  
                        stat=0
        end function 
        function helixcon_lop_2( adj, add, xx, yy) result(stat)
                integer                   :: stat,stat1,stat2 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                tmp=0.
                if ( adj) then
                        call helixcon_lop2(F,  F, yy,tmp)     ! tmp = grad y
                        call helixcon_lop2(T,add, xx,tmp)     ! x = x + grad' tmp
                else
                        call helixcon_lop2(F,  F, xx,tmp)     ! tmp = grad x
                        call helixcon_lop2(T,add, yy,tmp)     ! y = y + grad' tmp
                endif
                stat=stat*stat2
        end function 
        subroutine helixcon_lop2(adj,add,xx,yy)
                logical,intent(in) :: adj,add 
                real, dimension (:)  :: xx 
                real, dimension (:)  :: yy 
                integer iy, ix, ia

!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag

                if ( adj) then  ! zero lag
                        xx =	xx + yy
                else
                        yy =	yy + xx
                end if 
                do ia = 1, size( aa%lag)
                        do iy = 1  + aa%lag( ia), size( yy)
                                if ( associated( aa%mis)) then
                                        if ( aa%mis( iy)) cycle
                                end if
                                ix = iy - aa%lag( ia)
                                if ( adj) then
                                        xx(ix) =xx(ix) + yy(iy) * aa%flt(ia)
                                else
                                        yy(iy) =yy(iy) + xx(ix) * aa%flt(ia)
                                end if
                      end do
            end do

!write(0,*)'af'
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag


!stop
        end subroutine 
        subroutine helixcon_close()
        end subroutine 
!#####################################################  
! Helix polynomial division
        subroutine polyxdiv_init ( nd_in,aa_in )
                integer         :: nd_in 
                type( filter)   :: aa_in 
                nd = nd_in 
                aa = aa_in 
                if (allocated(tt)) deallocate(tt); allocate(tt ( nd))
        end subroutine
        function polyxdiv_lop( adj, add, xx, yy) result(stat)
                integer                   :: stat 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                        call adjnull(adj,add,xx,yy)
                        call polyxdiv_lop2(adj,add, xx,yy)  
                stat=0
        end function 
        function polyxdiv_lop_2( adj, add, xx, yy) result(stat)
                integer                   :: stat 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                tmp=0.
                if ( adj) then
                        call polyxdiv_lop2(F,  F, yy,tmp)     ! tmp = grad y
                        call polyxdiv_lop2(T,add, xx,tmp)     ! x = x + grad' tmp
                else
                        call polyxdiv_lop2(F,  F, xx,tmp)     ! tmp = grad x
                        call polyxdiv_lop2(T,add, yy,tmp)     ! y = y + grad' tmp
                endif
                stat=0.
        end function 
        subroutine polyxdiv_lop2(adj,add,xx,yy)
                logical,intent(in) :: adj,add 
                real, dimension (:)  :: xx 
                real, dimension (:)  :: yy 
                integer  ia, ix, iy
                tt = 0.

!write(0,*)'b4'
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag,(size(tt))
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag,(size(tt))
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag,(size(tt))
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag,(size(tt))
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag,(size(tt))
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag,(size(tt))

                if ( adj) then
                        do ix= nd, 1, -1  
                                tt( ix) = yy( ix)
                                do ia = 1, size( aa%lag)
                                        iy = ix + aa%lag( ia)
                                        if ( iy > nd) cycle
                                        tt( ix) = tt( ix) -  aa%flt( ia) * tt( iy)
                                end do
                        end do 
                        xx =	xx + tt
                else
                        do iy= 1, nd  
                                tt( iy) = xx( iy)
                                do ia = 1, size( aa%lag)
                                        ix = iy - aa%lag( ia)
                                        if ( ix < 1) cycle
                                        tt( iy) = tt( iy) -  aa%flt( ia) * tt( ix)
                                end do
                        end do 
                        yy = yy + tt
                end if

!write(0,*)'af'
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag
!write(0,*)sum(xx),sum(yy),aa%flt,aa%lag
        end subroutine 
        subroutine polyxdiv_close()
                deallocate(tt)
        end subroutine 
!#####################################################  
end module 
