module joint_encode_data_mod
  use joint_encode_data_history
  use joint_encode_data_io
  use joint_encode_data_array
  use joint_encode_data_param 
  use joint_encode_data_wavelet
  use joint_encode_data_grid
  use joint_encode_data_compute_mod
  use joint_encode_data_compute_mig
  use joint_encode_data_compute_mig_hx
  use joint_encode_scode_init
  use joint_encode_adjnull !new 
  use joint_encode_init    !new
  use joint_encode_outputs !new 
  use joint_encode_add     !new
  use joint_encode_fft     !new
  use joint_encode_regs    !new
  use joint_encode_prc     !new 
  use joint_encode_hlx     !new
  use joint_encode_nhlx    !new
  use joint_encode_hlx_blk !new
  use joint_encode_utils   !new
  use joint_encode_vnhlx   !new variable non-stationary helix
  use joint_encode_spatio_shift ! new spatio+shift
!
  use joint_encode_data_compute_hess_hx !compute phase encoded hessian

  use cgstep_module
  use solver_sep_mod
  use chain_mod 
!time-shifting
  use joint_shift_mod      !time/depth shifting module
  
  
!hybrid norm
  use solver_reg_nrm_mod   !General Norm
  
  implicit none

contains
!=======================================================================================
  integer function data_init() result(stat)

  call from_param("only_ltshift", only_ltshift , F)
  if(only_ltshift)then
            stat=joint_shift_init()
  else
            call joint_encode_params_init()
            call green_input_history()
            call green_params()                                     !initialize data modeling parameters
            call green_grid()                                       !get greens grid 
            call velocity_grid()                                    !get velocity/refl grid      
            call migration_grid()                                   !get migration grid      
            call data_wavelet_s()                                   !get wavelet
            call data_params()                                      !initialize data modeling parameters
             
            call shift_params_init()                                !time-shift init
            call data_array_alloc()                                 !allocate memory 
            call data_velocity()                                    !get velocity/refl models
!            if (wfld_d_out) then; call wfld_output_history(wfld_d_); call wfld_io_init(wfld_d_); endif; !save down wavefield in time 
!            if (wfld_u_out) then; call wfld_output_history(wfld_u_); call wfld_io_init(wfld_u_); endif; !save down wavefield in time
            if(adj.or.inv)then
                    call data_input_history(); 
                    call data_encode_sum();
                    call data_grid_born();
            else
                    call data_encode_sum();
                    call data_grid_born();
                    call data_output_history(); 
            endif 
            call data_io_init();
            call inv_param(); 
                
            call joint_array_alloc()
            call joint_encode_aux_init() 
            call joint_encode_all()
            if (lobc) call obc_params() 
            if (adj .or. inv) then
                    if (inv) write(*,*)'Initializing aux files for inversion...........' 
                    if (adj) write(*,*)'Initializing aux files for migration...........' 
                    call joint_encode_inout_init()
            else
                    write(*,*)'Initializing aux files for modeling........'
                    call joint_encode_inout_model_init()
            endif       
            if(use_greens)then
                stat = encode_compute_init1(nodes)
                call encode_scode_init()
            else
                if(encode) call joint_scode_init()    
            endif
            if(lhessian)then
                call hessian_params()
                call green_param()
                call hess_output_history()
                call hess_io_init()
            endif
    endif 
    stat = 0
  end function
!=======================================================================================
  integer function data_clos() result(stat); 
        if(only_ltshift)then
                stat = joint_shift_clos()
        else
                call data_array_dealloc(); call joint_encode_array_clos(); stat = 0; 
        endif
  end function
!!=======================================================================================
  integer function data_oper() result(stat) 
  use chain_mod
  use random
  real       ran
  integer :: ii,isurv
  
  if(only_ltshift)then
        stat=joint_shift_oper()
  else
        if(inv) then
                stat = encode_beg()
                write(0,'(A32,I4,2X,A5,I4,2X,A20)')'Starting inversion with mode: ', mode,'and',niter,'iterations'
                if(mode==1)then  
                    call solver_tiny(m=modl0, d=real(fdata0), Fop=data_lop, stepper=cgstepper, niter=niter, verb=T,resd=norm, timer=T)
                elseif(mode==2)then
                    !write(0,*)minval(epsi1),maxval(epsi1),minval(lmbd1),maxval(lmbd1),nmod,size(epsi0)
                    epsi0(nsurv*nmod+1:)=lmbd0    
                    call solver_reg(m=modl0, d=real(fdata0), Fop=data_lop, Aop=spatio_lop, stepper=cgstepper, nAop=nmod*(2*nsurv), niter=niter, eps=epsi0, verb=T, timer=T, Scale=vnhlx_scale, report_in=report_,spatio=spatio,mmov_=mmov_)
                !            call solver_regs2(m=modl0, d=real(fdata0), Fop=data_lop, Aop1=vnhlx_it, Aop2=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=epsi0, lmbd=lmbd0, verb=T, timer=T, Scale=vnhlx_scale, report_in=report_)
                !           call solver_regs2(m=modl0, d=real(fdata0), Fop=data_lop, Aop1=vnhlx_it, Aop2=temp_fb, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=epsi0, lmbd=lmbd0, verb=T, timer=T, Scale=vnhlx_scale, report_in=report_)
                !           call solver_regs_scale(m=modl0, d=real(fdata0), Fop=data_lop, Aop=vnhlx_it, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T, Scale=vnhlx_scale, report_in=report_) 
                !           call solver_regs(m=modl0, d=real(fdata0), Fop=data_lop, Aop=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, eps=maxval(epsi0), verb=T, timer=T)
                elseif(mode==3)then
                    tresh=maxval(abs(data0))/100. 
                    call IRLS_init(tresh)  
                    call IRLS(x=modl0, dat=real(fdata0), oper=data_lop, solv=cgstepper, niter=niter, nfreq=nfreq, wght=weight_IRLS, err=norm)  
                elseif(mode==4)then
                    call solver_regs2(m=modl0, d=real(fdata0), Fop=data_lop, Aop1=grad_oper, Aop2=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=epsi0, lmbd=lmbd0, verb=T, timer=T)
                elseif(mode==5)then
                    !call polydiv_init(nmod,aa)
                    call solver_prc( m=modl0, d=real(fdata0), Fop=data_lop, Sop=int_oper, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=eps, verb=T)
                elseif(mode==6)then
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=int_oper, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T)
                elseif(mode==7)then
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=row_int, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T)
                elseif(mode==8)then
                    hlx_conv=2
                    call filter_init()
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=hlx_it,  stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T)
                elseif(mode==9)then
                    call  ns_init(dips1,eps_dip)
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=nhlx_it, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T)
                elseif(mode==10)then
                    call  ns_init(dips1,eps_dip)
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=nhlx_row1, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=(epsi0*0.+1.), verb=T, timer=T)
                elseif(mode==11)then
                    call  ns_init(dips1,eps_dip)
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=nhlx_chain1, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=(epsi0*0.+1.), verb=T, timer=T)
                elseif(mode==12)then
                    hlx_conv=2
                    call filter_init_blk(dips1)
                    call solver_prc2( m=modl0, d=real(fdata0), Fop=data_lop, Sop=hlx_it_blks,  stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T)
                elseif(mode==13)then 
                    call solver_prc_scale( m=modl0, d=real(fdata0), Fop=data_lop, Sop=int_temp_oper, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T, Scale=int_scale) 
                elseif(mode==14)then
                    !call  ns_init(dips1,eps_dip) 
                    call solver_prc_scale( m=modl0, d=real(fdata0), Fop=data_lop, Sop=nhlx_chain1, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T, Scale=nhlx_scale, report_in=report_) 
                elseif(mode==15)then !spcae_time
                    call solver_prc_scale( m=modl0, d=real(fdata0), Fop=data_lop, Sop=vnhlx_chain1, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T, Scale=vnhlx_scale, report_in=report_) 
                elseif(mode==16)then !time_space
                    call solver_prc_scale( m=modl0, d=real(fdata0), Fop=data_lop, Sop=vnhlx_chain2, stepper=cgstepper, nSop=nmod*nsurv, niter=niter, eps=epsi0, verb=T, timer=T, Scale=vnhlx_scale, report_in=report_) 
                elseif(mode==17)then !regularization ,
                    destroy=T
                    call  ns_init(dips1,eps_dip) 
                    call solver_regs2(m=modl0, d=real(fdata0), Fop=data_lop, Aop1=vnhlx_it, Aop2=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=epsi0, lmbd=lmbd0, verb=T, timer=T)
                elseif(mode==18)then !L1
                    destroy=F
                    call l1_init (quant)
                    call IRLS_init (maxval(cabs(fdata0))/thres)
                    write(0,*)'threshold:',thres,maxval(cabs(fdata0))/thres
                    call solver_irls( m=modl0, d=real(fdata0), Fop=chain_data_prc, Wdiag=l1_nrm, Scale=vnhlx_scale, stepper=cgstepper, niter=niter, nmem=0, nfreq=nfreq, verb=T, timer=T) 
                    !call    IRLS( oper=chain_data_prc, solv=cgstepper, x=modl0, dat=real(fdata0), niter=niter, nfreq=nfreq, wght=weight_IRLS,Scale=vnhlx_scale)
                    !call solver_irls( m=modl0, d=real(fdata0), Fop=chain_data_prc, Wdiag=l1, Scale=vnhlx_scale, stepper=cgstepper, niter=niter, nmem=0, nfreq=nfreq, verb=T, timer=T)
                    !allocate(modl00(size(modl0))); modl00=modl0; !stat=vnhlx_chain2(F,F,modl00,modl0); !deallocate(modl00)    
                elseif(mode==19)then
                !write(0,*)size(lmbd0)
                !write(0,*)size(epsi0),nmod,nmod*nsurv,nmod*nsurv*2
                    !epsi0(nsurv*nmod+1:)=lmbd0    
                    call solver_reg(m=modl0, d=real(fdata0), Fop=data_lop, Aop=spatio_shift_lop, stepper=cgstepper, nAop=nmod*(2*nsurv), niter=niter, eps=epsi0(1:2*nsurv*nmod), verb=T, timer=T, Scale=vnhlx_scale, report_in=report_,spatio=spatio,mmov_=mmov_)
                elseif(mode==20)then
                    call solver_reg_nrm(m=modl0, d=real(fdata0), Fop=data_lop, Aop=spatio_lop, stepper=cgstepper, nAop=nmod*(2*nsurv), niter=niter, eps=epsi0, Scale=vnhlx_scale, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot)
                elseif(mode==21)then
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv)))
                    d0(1:ndat_real*nsurv)=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    call solver_reg_nrm(m=modl0, d=d0, Fop=data_reg_lop, Aop=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, eps=lmbd0, Scale=vnhlx_scale, verb=T, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot)
                    deallocate(d0)    
                elseif(mode==22)then    
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv)))
                    d0(1:ndat_real*nsurv)=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    call solver_tiny(m=modl0, d=d0, Fop=data_reg_lop, stepper=cgstepper, niter=niter, verb=T,resd=norm, timer=T, Scale=vnhlx_scale)
                    deallocate(d0)    
                elseif(mode==23)then    
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv*2)))
                    d0(1:ndat_real*nsurv   )=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    call solver_reg_nrm(m=modl0, d=d0, Fop=data_reg2_lop, Aop=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, eps=epsi0(nsurv*nmod*2+1:), Scale=vnhlx_scale, verb=T, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot)
                    deallocate(d0)
                elseif(mode==24)then    
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv*2)))
                    d0(1:ndat_real*nsurv   )=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    call solver_reg_nrm(m=modl0, d=d0, Fop=data_reg3_lop, Aop=grad_temp_oper, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=gam , Scale=vnhlx_scale, verb=T, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot, &
                                time_in=T, spatio_in=T, report_in=report_, nreg=3, ndat=ndat_real*nsurv, break_reg=T, comb_reg=T)
                    deallocate(d0)
                elseif(mode==25)then
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv*2)))
                    d0(1:ndat_real*nsurv   )=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    call solver_reg_nrm(m=modl0, d=d0, Fop=data_reg3_lop, Aop=vnhlx_chain2, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=gam , Scale=vnhlx_scale, verb=T, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot, &
                                time_in=T, spatio_in=T, report_in=report_, nreg=3, ndat=ndat_real*nsurv, break_reg=T, comb_reg=T)
                    deallocate(d0)
                elseif(mode==26)then    
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv*2)))
                    d0(1:ndat_real*nsurv   )=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    call solver_reg_nrm(m=modl0, d=d0, Fop=data_reg3_lop, Aop=spatio_lop, stepper=cgstepper, nAop=nmod*nsurv*2, niter=niter, epsi=gam , Scale=vnhlx_scale, verb=T, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot, &
                                time_in=T, spatio_in=T, report_in=report_, nreg=3, ndat=ndat_real*nsurv, break_reg=T, comb_reg=T)
                    deallocate(d0)
                elseif(mode==27)then
                    allocate(d0((ndat_real*nsurv)+(nmod*nsurv*2)))
                    d0(1:ndat_real*nsurv   )=real(fdata0)
                    d0(  ndat_real*nsurv+1:)=0.
                    write(0,*)'gam====>>>',gam
                    call solver_reg_nrm(m=modl0, d=d0, Fop=data_reg4_lop, Aop=grad_shift_lop, stepper=cgstepper, nAop=nmod*nsurv, niter=niter, epsi=gam , Scale=vnhlx_scale, verb=T, & 
                                norm_dat_in=norm_d, norm_mod_in=norm_m, qnt_dat_in=quant_d, qnt_mod_in=quant_m,qt_period_in=quant_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot, &
                                time_in=T, spatio_in=T, report_in=report_, nreg=3, ndat=ndat_real*nsurv, break_reg=T, comb_reg=T)
                    deallocate(d0)
                elseif(mode==28)then
                    call solver_reg(m=modl0, d=real(fdata0), Fop=data_lop, Aop=grad_shift_lop, stepper=cgstepper, nAop=nmod*(nsurv), niter=niter, eps=epsi0(1:nsurv*nmod), verb=T, timer=T, Scale=vnhlx_scale, report_in=report_,spatio=spatio,mmov_=mmov_)
                elseif(mode==29)then
                        write(0,*)maxval(epsi0(1:nsurv*nmod )),minval(epsi0(1:nsurv*nmod )), &
                                  maxval(epsi0(1+nsurv*nmod:)),minval(epsi0(1+nsurv*nmod:)) 
                    call solver_reg(m=modl0, d=real(fdata0), Fop=data_lop, Aop=vnhlx_shift_lop, stepper=cgstepper, nAop=2*nmod*(nsurv), niter=niter, eps=epsi0(1:2*nsurv*nmod), verb=T, timer=T, Scale=vnhlx_scale, report_in=report_,spatio=spatio,mmov_=mmov_)
                elseif(mode==30)then
                        write(0,*)maxval(epsi0(1:nsurv*nmod )),minval(epsi0(1:nsurv*nmod )), &
                                  maxval(epsi0(1+nsurv*nmod:)),minval(epsi0(1+nsurv*nmod:)) 
                    call solver_reg(m=modl0, d=real(fdata0), Fop=data_lop, Aop=spatio_shift_lop, stepper=cgstepper, nAop=2*nmod*(nsurv), niter=niter, eps=epsi0(1:2*nsurv*nmod), verb=T, timer=T, Scale=vnhlx_scale, report_in=report_,spatio=spatio,mmov_=mmov_)
                endif
                
                if(exist_file(modl1_))then; do i=1,nsurv; call m_io2('rite', modl0((i-1)*nmod+1:i*nmod),i); enddo; endif
                if(lsft)then; ts0=0;
                        do ishift=1,nshift_cycle
                                write(0,*)'shift cycle',ishift,' of ',nshift_cycle
                                stat= shift_init(ishift,T)
                                stat= shift_oper(adj,F,modl0) 
                        enddo
                if(exist_file(smodl2_))then; do i=1,nsurv; call sm_io2('rite', modl0((i-1)*nmod+1:i*nmod),i); enddo; endif
                return
                endif
        elseif(lhessian)then !only compute hessian
                if (allocated(fdata0))deallocate(fdata0)
                if (allocated( modl0))deallocate( modl0)
                stat = encode_compute_hess_hx(1)
                return
        elseif(only_ltshift)then !only time/depth-shifting
                stat= shift_oper(adj,F,modl0)
                stat= shift_vel()
        else
        write(0,*)'ltshift,adj,lvshift:',ltshift,adj,lvshift
                if(.not. adj.and.ltshift) then 
                      if(verb>6)write(0,*)'Initializing shifting params....' ; stat= shift_init(1,F)
                      if(verb>6)write(0,*)'Shifting the velocity model.....' ; stat= shift_oper(adj,F,vel0)
                      if(verb>6)write(0,*)'Shifting the reflectivity model.' ; stat= shift_oper(adj,F,modl0)
                      if(verb>6)write(0,*)'Modeling the data.....'
                elseif(ltshift)then
                      if(verb>6)write(0,*)'Initializing shifting params.....'; stat= shift_init(1,T) 
                endif
                        
                call joint_adjnull_foper(adj,F,modl0,fdata0)
                if(lsub_offset.and.(.not.lread_data))then
                        do isurv=1,nsurv; 
                            if(allocated(fdata0))deallocate(fdata0) !will read data in subroutine
                            if(verb>5)write(0,'(A20,I5,A15,L6,2X,2(E12.5,2X),2X,E12.5)')'Starting survey', isurv,': adj|mod',adj,sum(modl0((isurv-1)*nmod+1:isurv*nmod));
                            stat = encode_compute_hx(adj,modl0((isurv-1)*nmod+1:isurv*nmod),isurv);
                            if(verb>5)write(0,*)isurv,'sum dat:' , 'sum mod:' ,sum(modl0((isurv-1)*nmod+1:isurv*nmod));
                        enddo
                else        
                        stat=data_lop2(adj,F,modl0,fdata0);
                endif
                if(      adj.and.lsft)then;
                        if(exist_file(modl1_))then; do i=1,nsurv; call m_io2('rite', modl0((i-1)*nmod+1:i*nmod),i); enddo; endif
                        if(lsft)then;
                                do ishift=1,nshift_cycle
                                        write(0,*)'shift cycle',ishift,' of ',nshift_cycle
                                        stat= shift_init(ishift,T) 
                                        stat= shift_oper(T,F,modl0) 
                                enddo
                        if(exist_file(smodl2_))then; do i=1,nsurv; call sm_io2('rite', modl0((i-1)*nmod+1:i*nmod),i); enddo; endif
                        endif       
                return        
                endif
        end if
        call joint_encode_output() 
  endif        
  end function

!#####################################chain data modeling, dip and tempral filtering
  integer function chain_data_prc (adj, add, mod, dat) result(stat)
        logical, intent (in)     :: adj, add
        real, dimension (: )     :: mod, dat
write(0,*)'data_prc-b4 ',sum(dat),sum(mod),adj,add 
                stat=data_lop(adj, add, mod, dat)
                !stat=chain_2 (data_lop, vnhlx_chain2, adj, add, mod, dat)
                !stat=chain_2 (data_lop, vnhlx_chain2, adj, add, mod, dat)
write(0,*)'data_prc-end',sum(dat),sum(mod)
  end function
!!=======================================================================================
  integer function data_reg_lop(adj, add, modl, data) result(stat)
    logical,intent(in)              :: adj, add
    real, dimension(:)              :: modl
    real, dimension(:)              :: data
    call adjnull(adj,add,modl,data)
!    write(0,*)'1st-b4',adj,add,sum(modl),sum(data(1:ndat_real*nsurv ))
        stat=data_lop(adj,add, modl, data(1:ndat_real*nsurv ));
!    write(0,*)'1st-af',adj,add,sum(modl),sum(data(1:ndat_real*nsurv ))
!    write(0,*)'2nd-b4',adj,add,sum(modl),sum(data(ndat_real*nsurv+1:))
        if(adj)then
                stat=vnhlx_it(adj,T, modl,epsi0(1:nmod*nsurv)*data(ndat_real*nsurv+1:));
        else
                stat=vnhlx_it(adj,T, modl*epsi0(1:nmod*nsurv),data(ndat_real*nsurv+1:));
        endif
!     write(0,*)'2nd-af',adj,sum(modl),sum(data(1:ndat_real*nsurv )),sum(data(ndat_real*nsurv+1:))
        stat=0
  end function
!!=======================================================================================
  integer function data_reg2_lop(adj, add, modl, data) result(stat)
    logical,intent(in)              :: adj, add
    real, dimension(:)              :: modl
    real, dimension(:)              :: data
    call adjnull(adj,add,modl,data)
        if(adj)then
                stat=  data_lop(adj,add, modl,                      data(1:ndat_real*nsurv   ));
                stat=spatio_lop(adj,  T, modl,epsi0(1:nsurv*nmod*2)*data(  ndat_real*nsurv+1:));
        else
                stat=  data_lop(adj,add, modl,                      data(1:ndat_real*nsurv   ));
                stat=spatio_lop(adj,  T, modl*epsi0(1:nsurv*nmod*2),data(  ndat_real*nsurv+1:));
        endif
        stat=0
  end function
!!=======================================================================================
  integer function data_reg3_lop(adj, add, modl, data) result(stat)
    logical,intent(in)              :: adj, add
    real, dimension(:)              :: modl
    real, dimension(:)              :: data
    call adjnull(adj,add,modl,data)
        if(adj)then
                stat=       data_lop(adj,add, modl,    data(1:ndat_real*nsurv                                          ));
                stat=       vnhlx_it(adj,  T, modl,eps*data(  ndat_real*nsurv           +1:ndat_real*nsurv+nmod*nsurv  ));
                stat= grad_temp_oper(adj,  T, modl,lmb*data(  ndat_real*nsurv+nmod*nsurv+1:ndat_real*nsurv+nmod*nsurv*2));
        else
                stat=       data_lop(adj,add, modl,    data(1:ndat_real*nsurv                                          ));
                stat=       vnhlx_it(adj,  T, modl*eps,data(  ndat_real*nsurv           +1:ndat_real*nsurv+nmod*nsurv  ));
                stat= grad_temp_oper(adj,  T, modl*lmb,data(  ndat_real*nsurv+nmod*nsurv+1:ndat_real*nsurv+nmod*nsurv*2));
        endif
        stat=0
  end function
!!=======================================================================================
  integer function data_reg4_lop(adj, add, modl, data) result(stat)
    logical,intent(in)              :: adj, add
    real, dimension(:)              :: modl
    real, dimension(:)              :: data
    real, allocatable, dimension(:) :: xx
    integer                         :: isurv
    
    call adjnull(adj,add,modl,data)
        if(adj)then
                stat=        data_lop(adj,add, modl,    data(1:ndat_real*nsurv                                          ));
                stat= vnhlx_shift_lop(adj,  T, modl,eps*data(  ndat_real*nsurv           +1:ndat_real*nsurv+nmod*nsurv  ));
                stat=  grad_shift_lop(adj,  T, modl,lmb*data(  ndat_real*nsurv+nmod*nsurv+1:ndat_real*nsurv+nmod*nsurv*2));
        else
                stat=        data_lop(adj,add, modl,    data(1:ndat_real*nsurv                                          ));
                stat= vnhlx_shift_lop(adj,  T, modl*eps,data(  ndat_real*nsurv           +1:ndat_real*nsurv+nmod*nsurv  ));
                stat=  grad_temp_oper(adj,  T, modl*lmb,data(  ndat_real*nsurv+nmod*nsurv+1:ndat_real*nsurv+nmod*nsurv*2));
        endif
  end function
!!=======================================================================================
!!=======================================================================================
  integer function data_lop(adj, add, modl, data) result(stat)
    logical,intent(in)              :: adj,  add   
    real, dimension(:)              :: modl 
    real, dimension(:)              :: data
        stat=1
        stat=clean_pointers()
        call adjnull(adj,add,modl,data)
if(verb>5)write(0,'(A24,L3,L3,E12.4,2E12.4)')'b4: adj|add|dat|mod',adj,add,sum(data),sum(modl)
        fdata0=cmplx(data);                             !get data from solver 
        call joint_adjnull_foper(adj,F,modl,fdata0)     !initialize pointers 
        stat=data_convert(1);                           !convert to frequency
        stat=data_lop2(adj,add,modl,fdata0);            !run fwd or back
        stat=data_convert(-1)                           !back to time
        data=real(fdata0);                              !send data to solver
        stat=clean_pointers()                           !clean up pointers
        stat=0
if(verb>5)write(0,'(A24,L3,L3,E12.4,2X,E12.4)')'af: adj|add|dat|mod',adj,add,sum(data),sum(modl)

  end function
!!======================================================================================= 
  integer function data_lop2(adj, add, modl, data) result(stat)
  use joint_encode_data_compute_mig
    logical,intent(in)      :: adj,  add 
    real, dimension(:)      :: modl
    complex, dimension(:)   :: data 
    integer                 :: isurv
        stat = 1
        if(verb>5)then 
            write(0,*);write(0,*)'+++++'
            write(0,'(A10,10E12.4)')'min-real',minval(real(data)),minval(real(fd1)),minval(modl),minval(m1)
            write(0,'(A10,10E12.4)')'max-real',maxval(real(data)),maxval(real(fd1)),maxval(modl),maxval(m1)
            write(0,'(A10,10E12.4)')'min-imag',minval(aimag(data)),minval(aimag(fd1)),minval(modl),minval(m1)
            write(0,'(A10,10E12.4)')'max-imag',maxval(aimag(data)),maxval(aimag(fd1)),maxval(modl),maxval(m1)
            write(0,*)'+++++';write(0,*)
        endif
        if (use_greens)then    
            do isurv=1,nsurv;
                if(verb>4)write(0,*)'Starting survey', isurv,': adj|dat|mod',adj,sum(data((isurv-1)*ndat_real+1:isurv*ndat_real)),sum(modl((isurv-1)*nmod+1:isurv*nmod));stat = encode_compute(adj,modl((isurv-1)*nmod+1:isurv*nmod) ,data((isurv-1)*ndat_real+1:isurv*ndat_real),isurv); if(verb>5)write(0,*)'sum dat1:' ,sum(data((isurv-1)*ndat_real+1:isurv*ndat_real)),'sum mod1: ',sum(modl((isurv-1)*nmod+1:isurv*nmod));
            enddo
        else
            if(.not. adj)inv_niter=inv_niter+1
            do isurv=1,nsurv;
                if(verb>5)write(0,'(A20,I5,A15,L6,2X,2(E12.5,2X),2X,E12.5)')'Starting survey', isurv,': adj|dat|mod',adj,sum(data((isurv-1)*ndat_real+1:isurv*ndat_real)),sum(modl((isurv-1)*nmod+1:isurv*nmod));
                if(lsub_offset)then
                        if(allocated(fdata0))deallocate(fdata0) !will read data in subroutine
                        stat = encode_compute_hx(adj,modl((isurv-1)*nmod+1:isurv*nmod),isurv);
                else 
                        stat = encode_compute2  (adj,modl((isurv-1)*nmod+1:isurv*nmod),data((isurv-1)*ndat_real+1:isurv*ndat_real),isurv); 
                endif
                if(verb>5)write(0,*)isurv,'sum dat:' ,sum(data((isurv-1)*ndat_real+1:isurv*ndat_real)),'sum mod:' ,sum(modl((isurv-1)*nmod+1:isurv*nmod));
            enddo
        endif
        !call joint_add_oper(adj,add,modl,data)

        stat=0
  end function
!!=======================================================================================
  integer function encode_compute(adj, modl, data, isurv) result(stat)
    logical,intent(in)                    :: adj 
    complex, dimension(dgrid%agx%n, dgrid%agy%n,dgrid%asx%n,dgrid%asy%n,dgrid%aw%n),target:: data
    real,    dimension(vgrid%ax%n, vgrid%ay%n, vgrid%az%n                         ):: modl
    complex, dimension(:,:,:,:  ),pointer:: dat 
    real,    dimension(:,:,:    ),pointer:: mod
    integer                              :: iiw, iiiw, nb_nw, max_nw, isurv, i
            if(adj)then;allocate(model2( vgrid%ax%n, vgrid%ay%n, vgrid%az%n, ggrid%aw%n)); model2=0.;else;allocate(model2( vgrid%ax%n, vgrid%ay%n, vgrid%az%n, 1)); endif
            do i=1,size(model2,4); model2(:,:,:,i)=modl; enddo
            stat=encode_compute_init_s(isurv)
            nb_nw=nint(1.*ggrid%aw%n/nodes)  
            do iiw= 1,ggrid%aw%n,nodes
                max_nw=min(iiw+nodes-1,ggrid%aw%n)
                stat = encode_compute_init_blk(iiw,max_nw,isurv);
!$OMP parallel do private(inode, iw, iiiw, w, index_aw, mod, dat) schedule(dynamic,1) 
                do iw=iiw, max_nw                                ! freq.  blocks
                    inode=omp_get_thread_num()+1;
                    iiiw=iw-iiw+1
                    w=ggrid%aw%o+(iw-1)*ggrid%aw%d; index_aw=1.5+(w-dgrid%aw%o)/dgrid%aw%d
                    if(sum(cabs(csou_full(:,:,index_aw)))<dgrid%asx%n*dgrid%asy%n*mcsou) cycle; 
                    if(adj)then; mod=>model2(:,:,:,index_aw); else; mod=>model2(:,:,:,1);endif
                    dat=>data(:,:,:,:,index_aw);
                    call encode_compute_data(adj, mod, dat,inode,iiiw,index_aw, isurv)
                    !write(0,'(6I10,8E20.5)')inode,iw,iiw,iiiw,max_nw,index_aw, sum(model2),sum((fdata))
                    stat = encode_report(adj,mod,dat,index_aw,w,inode,isurv)
                    nullify(mod,dat);
                end do
!$OMP end parallel do 
            end do 
            if(adj)modl=sum(model2,4)
    deallocate(model2)
  end function
!!=======================================================================================
  integer function lop(adj, add, modl, data) result(stat) !used for dottest 
    logical,intent(in)    :: adj,  add
    real, dimension(:)    :: modl
    real, dimension(:) :: data
    integer                 :: isurv
        stat = 1
        call adjnull(adj,F,modl,data)
        !call joint_adjnull_foper2(adj,add,modl,data)
        do isurv=1,nsurv
!            if(isurv==1) stat = encode_compute(adj,m1,fd1,1)
!            if(isurv==2) stat = encode_compute(adj,m2,fd2,2)
!            if(isurv==3) stat = encode_compute(adj,m3,fd3,3)
!            if(isurv==4) stat = encode_compute(adj,m4,fd4,4)
!            if(isurv==5) stat = encode_compute(adj,m5,fd5,5)
!            if(isurv==6) stat = encode_compute(adj,m6,fd6,6)
        enddo
      !  call joint_add(adj,add,modl,data)
        stat=0
  end function
!===========================================================================================================================================================================
end module
