module joint_encode_nhlx
 use joint_encode_params
! use joint_encode_adjnull
! use joint_encode_rows
 use chain0_mod
 use joint_encode_prc
 use joint_encode_regs
 use createhelix_mod
 use cartesian_mod
 use bound_mod
 use regrid_mod
 use adj_mod
 use lapfac_mod
 use npolydiv_mod
 use nhelicon_mod
 use wilson_mod

 implicit none
        integer,         private                                ,save   :: stat        
        type( filter),   private                                ,save   :: aa
        type( filter),   private                                ,save   :: bb
        real,            private,  dimension (:), allocatable   ,save   :: tt 
        integer,         private                                ,save   :: nd 
        logical,         private                                ,save   :: lscale
        real,                                     allocatable   ,save   :: scale(:)
        integer,         private                                ,save   :: n1 
        real,            private,                 allocatable   ,save   :: tdip(:)
        real   ,         private                                         :: eps_nhlx
 contains 
!######################################
        subroutine  ns_init(dips,eps_in)
                integer                 ::i,n1_in
                real, dimension(2)      ::u,v
                real                    ::eps_in, eps_nhlx
                real                    :: dips(:)
                if(allocated(ptch)) deallocate(ptch)
                if(allocated(n_nfilt)) deallocate(n_nfilt)
                if(allocated(tdip)) deallocate(tdip)
                
                allocate (ptch(nmod),n_nfilt(nang),tdip(nmod))

                        eps_dip=eps_in

                n1=vgrid%az%n        
                call from_param('scale',lscale,.false.)
                call from_param('destroy',destroy,.false.)

                if (lscale)then !scale according to angle
                        call from_param("osc",osc,oang); call from_param("dsc",dsc,dang); call from_param("nsc",nsc,nang)
                        allocate (scale(nsc))
                        do i=1,nsc
                                write(0,*)scale
                                scale(i)=osc+(i-1)*dsc
                                write(0,*)i,scale
                        enddo
                endif
                ptch=0.

                tdip=dips
                call nhlx_transp(1,tdip) !reange angles to z-1 x-2
                !tdip = atan(tdip)
                where(tdip> 1.570)tdip=1.5708    ! minvalmin-angle :-90/90 
                where(tdip<-1.570)tdip=-1.5708
                if(verb>5)write(0,*)'dipinfo:beg',nang,dang,oang
                if(maxval(tdip)==minval(tdip))then
                        nang=1
                        dang=1
                        oang=minval(tdip)
                if(verb>5)write(0,*)'dipinfo:a',nang,dang,oang
                elseif(minval(tdip)>oang.and.maxval(tdip)<oang+(nang-1)*dang)then
                        oang=minval(tdip)
                        dang=(maxval(tdip)-oang)/(nang-1)
                if(verb>5)write(0,*)'dipinfo:b',nang,dang,oang
                elseif(minval(tdip)>oang)then
                        oang=minval(tdip)
                        if(maxval(tdip)<oang+(nang-1)*dang)then
                                dang=(maxval(tdip)-oang)/(nang-1)
                        endif
                if(verb>5)write(0,*)'dipinfo:c',nang,dang,oang
                endif
                if(verb>5)write(0,*)'dipinfo:end',nang,dang,oang
                do i=1,nmod
                        dip=tdip(i)
                        idip = 1.5+(dip-oang)/dang
                        ptch(i)=idip
                        if(ptch(i)>nang.or.(ptch(i)<0.))write(0,*)dip,idip,ptch(i)
                end do
                where (ptch>float(nang))ptch=nang
                where (ptch<1.         )ptch=1

                n_nfilt=nfilt*3
                call nallocate(n_filt,n_nfilt,ptch)
                do iang=1,nang
                        do ifilt=1,nfilt
                                n_filt%hlx(iang)%lag(         ifilt) = ifilt                      ! early lags (first row)
                                n_filt%hlx(iang)%lag( nfilt  +ifilt) = n1 + ifilt - nfilt         ! late lags  (second row)
                                n_filt%hlx(iang)%lag( nfilt*2+ifilt) = n1 + ifilt                 ! late late lags (second row)
                        end do
                end do
               call nallocate(lapfilt,n_nfilt, ptch)            ! alloc. filter
               lapfilt0 = 1. + eps_dip        ! zero lag coeff. multiply by variance
               
!$OMP parallel do private(dip,u,m,p) schedule (dynamic)
                do iang=1,nang
                        lapfilt%hlx(iang)%lag = (/ 1, n1-1, n1, n1+1/)    ! lag(1)= 1; lag(2)=n1  # one side only
                        dip=oang+(iang-1)*dang
                        if(verb>5)write(0,*)'working on angle',iang,dip,degrad(1,dip),'of', nang
                        write(0,*)'working on angle',iang,dip,degrad(1,dip),'of', nang
                        u(1)= cos(dip); u(2)=-sin(dip); 
        
                        m=(u(1)-u(2))/2.; 
                        p=(u(1)+u(2))/2.;
                        lapfilt%hlx(iang)%flt = 0.;
                        lapfilt%hlx(iang)%flt = (/2*m*p,-p*p,-2*m*p,-m*m/) 
        
                        call wilson_init( 10 * n1 )
                        call wilson_factor( 100, lapfilt0, lapfilt%hlx(iang), n_filt0, n_filt%hlx(iang))
                        call wilson_close()
                        n_filt%hlx(iang)%flt=n_filt%hlx(iang)%flt!/n_filt0
                        if(verb>5)write(0,'(50F10.6)')n_filt%hlx(iang)%flt
                end do
!$OMP end parallel do                        
                if (lscale.and.nang==nsc)then    !scale according to the angle!!!!!! useful for zeroing specific angles
                        do iang=1,nang
                                n_filt%hlx(iang)%flt=n_filt%hlx(iang)%flt*scale(iang)
                        end do
                endif
                call npolydiv_init(nmod,n_filt)

        end subroutine
!#####################################################  transp before filter and back
        subroutine nhlx_transp(fwd,xx,yy)
                integer,intent(in)      :: fwd 
                real, dimension (:)     :: xx, yy 
                real, dimension(:,:), allocatable       :: temp
                optional                        :: yy
                
                if (fwd==1)then
                        allocate (temp(vgrid%ax%n,vgrid%az%n))
                        temp=reshape(xx,(/vgrid%ax%n,vgrid%az%n /)); xx=reshape(transpose(temp),(/ nmod /))
                        if(present(yy))then; temp=reshape(yy,(/vgrid%ax%n,vgrid%az%n /)); yy=reshape(transpose(temp),(/ nmod /)); endif
                else
                        allocate (temp(vgrid%az%n,vgrid%ax%n))
                        temp=reshape(xx,(/vgrid%az%n,vgrid%ax%n /)); xx=reshape(transpose(temp),(/ nmod /))
                        if(present(yy))then; temp=reshape(yy,(/vgrid%az%n,vgrid%ax%n /)); yy=reshape(transpose(temp),(/ nmod /));endif
                endif
                stat=0
                deallocate (temp)
        end subroutine
!#####################################################  conv or deconv
        integer function nhlx_it(adj, add, xx, yy) result(stat)
                logical,intent(in)   :: adj,add 
                real, dimension (:)  :: xx, yy
                integer              :: isurv    
                call joint_adjnull2 (adj,add, xx,yy)
                        do isurv=1,nsurv; call nhlx_transp(1,xx((isurv-1)*nmod+1:isurv*nmod),yy((isurv-1)*nmod+1:isurv*nmod));enddo
                if (destroy)then
                        do isurv=1,nsurv; call filt_nhelicon(adj,add,xx((isurv-1)*nmod+1:isurv*nmod),yy((isurv-1)*nmod+1:isurv*nmod));enddo
                else
                        do isurv=1,nsurv; call filt_npolydiv(adj,add,xx((isurv-1)*nmod+1:isurv*nmod),yy((isurv-1)*nmod+1:isurv*nmod));enddo
                endif
                        do isurv=1,nsurv; call nhlx_transp(-1,xx((isurv-1)*nmod+1:isurv*nmod),yy((isurv-1)*nmod+1:isurv*nmod));enddo
                stat=0
        end function
!#####################################spatial int + temp-smooth ...note:int_temp_oper in prc
        integer function nhlx_row1(adj, add, mod, dat) result(stat)
                logical, intent (in)     :: adj, add
                real, dimension (: )     :: mod, dat
                stat=1; stat=row_1(nhlx_it, int_temp_oper, adj,add,mod,dat) 
        end function
!######################################
        integer function nhlx_row2(adj, add, mod, dat) result(stat)
                logical, intent (in)     :: adj, add
                real, dimension (: )     :: mod, dat
                stat=1; stat=row_2(nhlx_it, int_temp_oper, adj,add,mod,dat) 
        end function
!######################################
        integer function nhlx_row3(adj, add, mod, dat) result(stat)
                logical, intent (in)     :: adj, add
                real, dimension (: )     :: mod, dat
                stat=1; stat=row_2(nhlx_it, int_temp_oper, adj,add,mod,dat) 
        end function
!######################################
        integer function nhlx_chain1(adj, add, mod, dat) result(stat)
                logical, intent (in)            :: adj, add
                real, dimension (: )            :: mod, dat
                real, dimension (size(dat))     :: tmp
                stat=1; call chain0(nhlx_it, int_temp_oper,adj,add,mod,dat,tmp); stat=0
        end function
!######################################
        subroutine filt_npolydiv( adj, add, xx, yy) 
                integer                   :: stat,stat1,stat2 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                tmp=0.
                if ( adj) then
                        stat= npolydiv_lop(.false.,add, yy,tmp)     ! tmp = grad y
                        stat= npolydiv_lop(.true. ,add, xx,tmp)     ! x = x + grad' tmp
                else
                        stat= npolydiv_lop(.false.,add, xx,tmp)     ! tmp = grad x
                        stat= npolydiv_lop(.true. ,add, yy,tmp)     ! y = y + grad' tmp
                endif
                stat=stat*stat2
        end subroutine 
!######################################
        subroutine filt_nhelicon( adj, add, xx, yy) 
                integer                   :: stat,stat1,stat2 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                tmp=0.
                if ( adj) then
                        stat= nhelicon_lop(.false.,add, yy,tmp)     ! tmp = grad y
                        stat= nhelicon_lop(.true. ,T  , xx,tmp)     ! x = x + grad' tmp
                else
                        stat= nhelicon_lop(.false.,add, xx,tmp)     ! tmp = grad x
                        stat= nhelicon_lop(.true. ,T  , yy,tmp)     ! y = y + grad' tmp
                endif
                stat=stat*stat2
        end subroutine 
!#####################################Scaling functions for preconditioner
  integer function filt_scale(iter,niter) result(stat)
        integer, intent (in)    :: iter, niter
        real                        :: eps
        stat=1
                riter=iter
                if((modulo(1.*iter,dsscale)==0.))then        
                        if(lscale)then; eps=sscale(iter);call ns_init(dips1,eps); else; endif
                endif
        stat=0
  end function
!#####################################Scaling functions for nhlx
  integer function nhlx_scale(iter,niter,adj) result(stat)
        integer, intent (in)     :: iter, niter
        logical,intent(in)       :: adj
        stat=1
        if(iter==1.or.(lsscale.and.adj))then; if (modulo(iter,int(dsscale))==1)then; 
            eps_dip=min_eps_dip+(iter-1)*(max_eps_dip-min_eps_dip)/niter; 
            write(0,*)'[Re]Computing spatial-filters with eps_dip:',eps_dip
            call ns_init(dips1,eps_dip); else;            ; endif; 
        endif
        if(iter==1.or.(ltscale.and.adj))then; if (modulo(iter,int(dtscale))==1)then;
            rho_t=max_rho_t-(iter-1)*(max_rho_t-min_rho_t)/niter; 
            write(0,*)'[Re]Computing temporal-filters with rho_t :',rho_t
                                         else; rho_t=rho_t; endif; 
        endif
        stat=0
  end function
! #####################################Scaling functions for nhlx
!  integer function nhlx_scale(iter,niter,adj) result(stat)
!        integer, intent (in)     :: iter, niter
!        logical,intent(in)       :: adj
!        stat=1
!        if(lsscale.and.adj)then; if (modulo(iter,int(dsscale))==1)then;  
!            eps_dip=min_eps_dip+(iter-1)*(max_eps_dip-min_eps_dip)/niter; 
!            write(0,*)'[Re]computing spatial-filters with eps_dip:',eps_dip
!            call ns_init(dips1,eps_dip); else;            ; endif; 
!        endif
!        if(ltscale.and.adj)then; if (modulo(iter,int(dtscale))==1)then;
!            rho_t=max_rho_t-(iter-1)*(max_rho_t-min_rho_t)/niter; 
!            write(0,*)'[Re]computing temporal-filters with rho_t :',rho_t
!                                         else; rho_t=rho_t; endif; 
!        endif
!        stat=0
!  end function
  
end module
