module joint_encode_regs
 use joint_encode_params
 use joint_encode_adjnull
 use joint_encode_add
 use igrad2 
 use igrad1
 use laplac2
 use causint_mod
 use leakint
 use chain_mod
 use joint_encode_rows
! use shift_module
! use joint_encode_hlx
  
 implicit none
        integer, private, save:: stat        

 contains 
!#####################################################  
!#######
  integer function precon_oper(adj, add, modl, data) result(stat)
    logical            :: adj, add
    real, dimension(:):: modl, data
    stat = 1
    call joint_adjnull2 (adj,add,  modl,data)
 !   call joint_precond (adj,add          )
    call joint_add      (adj,add,modl,data) 
    stat = 0
  end function
!####### 
  integer function grad_oper(adj, add, modl, data) result(stat)
    logical,intent(in) :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    call joint_adjnull2         (adj, add,  modl,data)
    call joint_encode_gradient  (adj, add            )
    stat = 0
  end function
!#######
  integer function grad_temp_oper(adj, add, modl, data) result(stat)
    logical,intent(in) :: adj, add
    real, dimension(:) :: modl, data
    stat = 1
    !call joint_adjnull2         (adj, add, modl, data)
    call joint_encode_grad_temp (adj, add, modl, data) 
    stat = 0 
  end function
!#######
  integer function grad_x_temp_oper(adj, add, modl, data) result(stat)
    logical                     :: adj, add
    real, dimension(:)          :: modl, data
    real, dimension(nmod*nsurv) :: tmp1
    stat = 1
    call chain2 (grad_oper, grad_temp_oper, adj,  T,  modl,  data, tmp1)
    stat = 0
  end function
!#######
  subroutine joint_encode_grad_temp(adj, add, modl, data) 
    logical             :: adj, add
    real, dimension(:)  :: modl, data
    integer             :: isurv         
        call adjnull(adj,add,modl,data)
        if (adj) then
                do isurv=1,nsurv-1;
                        modl(((isurv  )*nmod)+1:(isurv+1)*nmod)=modl(((isurv  )*nmod)+1:(isurv+1)*nmod)+data(((isurv-1)*nmod)+1:(isurv  )*nmod)
                        modl(((isurv-1)*nmod)+1:(isurv  )*nmod)=modl(((isurv-1)*nmod)+1:(isurv  )*nmod)-data(((isurv-1)*nmod)+1:(isurv  )*nmod)
                enddo
        else
                do isurv=1,nsurv-1;
                        data(((isurv-1)*nmod)+1:(isurv  )*nmod)=data(((isurv-1)*nmod)+1:(isurv  )*nmod)+modl(((isurv  )*nmod)+1:(isurv+1)*nmod)
                        data(((isurv-1)*nmod)+1:(isurv  )*nmod)=data(((isurv-1)*nmod)+1:(isurv  )*nmod)-modl(((isurv-1)*nmod)+1:(isurv  )*nmod)
                enddo
        end if
  end subroutine
!#####################################
  subroutine joint_encode_grad(adj, modl, data)
    logical                :: adj
    real, dimension(:)     :: modl, data
    integer                :: i

    do i= 1,size(data);  
        if (adj) then
                modl(i+1)=modl(i+1)+data(i  ) 
                modl(i  )=modl(i  )-data(i  ) 
        else 
                data(i  )=data(i  )+modl(i+1) 
                data(i  )=data(i  )-modl(i  ) 
        end if
    end do
    return
  end subroutine
!##################grad
  subroutine joint_encode_gradient(adj, add )
  logical            :: adj, add
                if (nsurv > 5)call grad_base(adj,add,m6,d6);
                if (nsurv > 4)call grad_base(adj,add,m5,d5);                           
                if (nsurv > 3)call grad_base(adj,add,m4,d4);                                                           
                if (nsurv > 2)call grad_base(adj,add,m3,d3);                                                                                             
                if (nsurv > 1)call grad_base(adj,add,m2,d2);
                if (nsurv > 0)call grad_base(adj,add,m1,d1);
  end subroutine
!##################apply link mthd 1 and 2
  subroutine joint_encode_gradient_link(adj)
  logical            :: adj

  end subroutine
!##################apply fact-dz
  subroutine hessian_fact_dz(adj)
  logical            :: adj
                if (nsurv > 5)call fact_dz(adj,m6,d6); 
                if (nsurv > 4)call fact_dz(adj,m5,d5);                            
                if (nsurv > 3)call fact_dz(adj,m4,d4);                                                             
                if (nsurv > 2)call fact_dz(adj,m3,d3);                                                                                              
                if (nsurv > 1)call fact_dz(adj,m2,d2);                                                                                                                               
                if (nsurv > 0)call fact_dz(adj,m1,d1);                                     
  end subroutine
!##################apply damping
  subroutine hessian_damping(adj)
  logical            :: adj
                if (nsurv > 5)call damping(adj,m6,d6); 
                if (nsurv > 4)call damping(adj,m5,d5);                            
                if (nsurv > 3)call damping(adj,m4,d4);                                                             
                if (nsurv > 2)call damping(adj,m3,d3);                                                                                              
                if (nsurv > 1)call damping(adj,m2,d2);                                                                                                                               
                if (nsurv > 0)call damping(adj,m1,d1);                                     
  end subroutine
!#####################################
subroutine grad_base(adj, add, modl, data)
    logical                :: adj, add
    real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
    real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
    integer                :: ix_modl, iy_modl, iz_modl
            if(reg_base==1)then; call  grad_x(adj, add, modl, data);
        elseif(reg_base==2)then; call  lapl_x(adj, add, modl, data); endif
  return
  end subroutine
!##################4d regs
  subroutine grad_4d_xz(adj, modl, data, pick)
    use cartesian
    logical                :: adj
    real, dimension(vgrid%ax%n*vgrid%ay%n*vgrid%az%n) :: modl, data, pick
    real, dimension(vgrid%ax%n*vgrid%ay%n*vgrid%az%n) :: mod , dat , pic
    integer                :: ix_modl, iy_modl, iz_modl
    real                   :: clip1,clip2
        mod=modl
        dat=data
            if(reg_4d==1)then; call grad_4d_x  (adj, mod, dat );
        elseif(reg_4d==2)then; call grad_4d_z  (adj, mod, dat );
        elseif(reg_4d==3)then; call lapl_4d_x  (adj, mod, dat );
        elseif(reg_4d==4)then; call lapl_4d_z  (adj, mod, dat );
        elseif(reg_4d==5)then; call lapl_4d_2d (adj, mod, dat );
        elseif(reg_4d==6)then; call grad_4d_2d (adj, mod, dat );
        elseif(reg_4d==7)then; call grad_4d_2d2(adj, mod, dat );
        elseif(reg_4d==8)then; call grad_4d_z  (adj, mod, dat ); 
                               call grad_4d_x  (adj, modl,data); 
        elseif(reg_4d==9 )then;call grad_4d_2d (adj, mod, dat ); 
                               call grad_4d_x  (adj, modl,data); 
        elseif(reg_4d==10)then;call grad_4d_2d (adj, mod, dat ); 
                               call lapl_4d_x  (adj, modl,data); 
        elseif(reg_4d==11)then;call grad_4d_2d (adj, mod, dat ); 
                               call lapl_4d_2d (adj, modl,data); 
        elseif(reg_4d==12)then;call lapl_4d_2d (adj, mod, dat ); 
                               call grad_4d_x  (adj, modl,data); 
        elseif(reg_4d==13)then;call lapl_4d_2d (adj, mod, dat );  
                               call lapl_4d_x  (adj, modl,data);  
        else
                call seperr('Select reg_4d option between 1(x-grad.), 2(z-gradient), 3(xyz-laplac.), 4(z-laplac.), 5(xyz-grad. square mag.), 6(xyz-grad. mag.), 7(xyz-laplac.)')
        endif
        if(lpick)then
                if(adj)then
                       where(pick>pclip1.and.pick<pclip2);modl=(mod*pick+modl*(1-pick));elsewhere(pick>=pclip2);modl=(mod*pick);elsewhere; modl=mod; end where
                else
                       where(pick>pclip1.and.pick<pclip2);data=(dat*pick+data*(1-pick));elsewhere(pick>=pclip2);data=(dat*pick);elsewhere; data=dat; end where
                endif
        else
                if(adj)then
                       modl=mod; 
                else
                       data=dat;
                endif
        endif
        return
  end subroutine
!#####################################
  subroutine lapl_x(adj, add, modl, data)
        logical                        :: adj, add
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        real, dimension(vgrid%ax%n                      ) :: tmp
        integer                                           :: stat1, stat2
        tmp=0.
        if ( adj) then
                do iz_modl= 1,vgrid%az%n;
                        stat1 = igrad1_lop (  F,    F, data(:,1,iz_modl),tmp)   ! tmp = grad y
                        stat2 = igrad1_lop (  T,  add, modl(:,1,iz_modl),tmp)   ! x = x + grad' tmp
                enddo
        else
                do iz_modl= 1,vgrid%az%n;
                        stat1 = igrad1_lop (  F,    F, modl(:,1,iz_modl),tmp)   ! tmp = grad x
                        stat2 = igrad1_lop (  T,  add, data(:,1,iz_modl),tmp)   ! y = y + grad' tmp
                enddo
        endif
  end subroutine
!#####################################
  subroutine lapl_z(adj, modl, data)
        logical                        :: adj, add
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        real, dimension(vgrid%ax%n                      ) :: tmp
        integer                                           :: stat1, stat2
        if ( adj) then
                do ix_modl= 1,vgrid%ax%n;
                        stat1 = igrad1_lop (  T,  add, modl(ix_modl,1,:), data(ix_modl,1,:))   ! x = x + grad' tmp
                enddo
       else
                do ix_modl= 1,vgrid%ax%n;
                        stat1 = igrad1_lop (  T,  add, modl(ix_modl,1,:), data(ix_modl,1,:))   ! y = y + grad' tmp
                enddo
        endif
  end subroutine
!#####################################
  subroutine lapl_4d_x(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        call laplac2_init  (vgrid%ax%n,1)
        do iz_modl= 1,vgrid%az%n;
                stat=laplac2_lop(adj,T,modl(:,1,iz_modl),data(:,1,iz_modl))
        enddo    
  end subroutine
!#####################################
  subroutine lapl_4d_z(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        call laplac2_init  (vgrid%az%n,1)
        do ix_modl= 1,vgrid%ax%n;
                stat=laplac2_lop(adj,T,modl(ix_modl,1,:),data(ix_modl,1,:))
        enddo    
  end subroutine
!#####################################
  subroutine lapl_4d_2d(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        call laplac2_init  (vgrid%ax%n,vgrid%az%n)
!        stat=laplac2_lop(adj,T,modl,data)
  end subroutine
!#####################################
  subroutine grad_x(adj, add, modl, data)
        logical                :: adj, add
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data

        do iz_modl= 1,vgrid%az%n-1;
                if(adj)then
                        modl(:,1,iz_modl  )=modl(:,1,iz_modl  )-data(:,1,iz_modl)
                        modl(:,1,iz_modl+1)=modl(:,1,iz_modl+1)+data(:,1,iz_modl)
                else
                        data(:,1,iz_modl  )=data(:,1,iz_modl  )+modl(:,1,iz_modl+1)-modl(:,1,iz_modl)
                endif
!                stat=igrad1_lop(adj,T,modl(:,1,iz_modl),data(:,1,iz_modl))
        enddo    
!modl(1,1,:)=modl(2,1,:);modl(vgrid%ax%n,1,:)=modl(vgrid%ax%n-1,1,:);
!data(1,1,:)=data(2,1,:);data(vgrid%ax%n,1,:)=data(vgrid%ax%n-1,1,:);

  end subroutine
!#####################################
  subroutine grad_z(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        do ix_modl= 1,vgrid%ax%n;
                stat=igrad1_lop(adj,T,modl(ix_modl,1,:),data(ix_modl,1,:))
        enddo    
  end subroutine
!#####################################
  subroutine grad_4d_x(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        do iz_modl= 1,vgrid%az%n;
                stat=igrad1_lop(adj,T,modl(:,1,iz_modl),data(:,1,iz_modl))
        enddo    
  end subroutine
!#####################################
  subroutine grad_4d_z(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        do ix_modl= 1,vgrid%ax%n;
                stat=igrad1_lop(adj,T,modl(ix_modl,1,:),data(ix_modl,1,:))
        enddo    
  end subroutine
!#####################################
  subroutine grad_4d_2d(adj, modl, data)
        logical                :: adj
        real, dimension(nmod)  :: modl
        real, dimension(nmod)  :: data
        real, dimension(nmod*2):: d2
        call laplac2_init  (vgrid%ax%n,vgrid%az%n)
        if(adj)then
                d2(     1:nmod  )=data
                d2(nmod+1:nmod*2)=data
                stat=igrad2_lop(T,T,modl,d2)
        else
                stat=igrad2_lop(F,T,modl,d2)
                data=d2(1:nmod)+d2(nmod+1:nmod*2)
        endif
  end subroutine
!#####################################
  subroutine grad_4d_2d2(adj, modl, data)
        logical                :: adj
        real, dimension(nmod)  :: modl
        real, dimension(nmod)  :: data
        real, dimension(nmod*2):: d2
        call laplac2_init  (vgrid%ax%n,vgrid%az%n)
        if(adj)then
                d2(     1:nmod  )=data**2
                d2(nmod+1:nmod*2)=data**2
                stat=igrad2_lop(T,T,modl,d2)
        else
                stat=igrad2_lop(F,T,modl,d2)
                data=d2(1:nmod)**2+d2(nmod+1:nmod*2)**2
        endif
  end subroutine
!!#####################################
  subroutine fact_dz(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        integer                :: ix_modl, iy_modl, iz_modl
        do iz_modl= 1,vgrid%az%n;
                if(adj)then
                        modl(:,:,iz_modl)=modl(:,:,iz_modl)*1/(iz_modl*dz_fact)
                else    
                        data(:,:,iz_modl)=data(:,:,iz_modl)*1/(iz_modl*dz_fact)
                endif
        enddo    
  end subroutine
!#####################################
  subroutine damping(adj, modl, data)
        logical                :: adj
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: modl
        real, dimension(vgrid%ax%n,vgrid%ay%n,vgrid%az%n) :: data
        integer                :: ix_modl, iy_modl, iz_modl

        if(adj)then
                modl=modl+data
        else    
                data=data+modl
        endif
  end subroutine
!#####################################
  integer function row_grad(adj, add, mod, dat) result(stat)
        logical, intent (in)     :: adj, add
        real, dimension (: )     :: mod, dat
        stat=row_1(grad_oper, grad_temp_oper, adj,add,mod,dat)
  end function
!### forward+back
!######################################spatial (forward * back)  row ====>>>d = (eA'A)m and m = (eA'A)d -space
  integer function temp_fb( adj, add, mod, dat) result(stat) 
        integer                   :: stat1,stat2 
        logical,intent(in)        :: adj,add 
        real,dimension(:)         :: mod,dat 
                stat= row_fb (grad_temp_oper, adj, add, mod, dat)
  end function


  end module
!#####################################damping the differences
  subroutine damp_diff_oper(adj, add, modl, data)
        logical                	:: adj, add
        real, dimension(:) 	:: modl
        real, dimension(:) 	:: data

!	call joint_encode_grad_temp(adj, add, modl, data)
        if(adj)then
                modl=modl+data
        else    
                data=data+modl
        endif
  end subroutine
