module joint_encode_rows
 use joint_encode_params

 implicit none
contains
!#####################################
  integer function row_1 (oper1, oper2, adj, add, mod, dat) result(stat) 
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
                function oper2 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper2
        end interface
        logical, intent (in)            :: adj, add
        real, dimension (: )            :: mod, dat
        real, dimension (size(dat))     ::  xx, yy
        integer                         ::  stat1, stat2
        xx=mod; yy=dat
        if (adj) then
                stat1 = oper1 (T, add, mod, dat)
                stat2 = oper2 (T, add,  xx,  yy)
                mod=epsi1*mod+lmbd1*xx
        else
                stat2 = oper2 (F, add, mod, dat)
                stat1 = oper1 (F, add,  xx,  yy)
                dat=epsi1*dat+lmbd1*yy
        end if
        stat=stat1+stat2
  end function row_1
!#####################################
  integer function row_2 (oper1, oper2, adj, add, mod, dat) result(stat)
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
                function oper2 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper2
        end interface
        logical, intent (in)     :: adj, add
        real, dimension (: )     :: mod, dat
        integer                  ::  stat1, stat2
        if (adj) then
                stat1 = oper1 (T, add, mod, dat)
                stat2 = oper2 (T,   T, mod, dat)
        else
                stat2 = oper2 (F, add, mod, dat)
                stat1 = oper1 (F,   T, mod, dat)
        end if
        stat=stat1+stat2
  end function row_2
!#####################################
  integer function row_3 (oper1, oper2, adj, add, mod, dat) result(stat)
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
                function oper2 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper2
        end interface
        logical, intent (in)     :: adj, add
        real, dimension (: )     :: mod, dat
        integer                  ::  stat1, stat2

        if(verb>7)write(0,'(A8,2L6,8E15.5)')'af--',adj,add,maxval(epsi0),minval(epsi0),maxval(lmbd0),minval(lmbd0),sum(mod),sum(dat)
        if (adj) then
                stat2 = oper2 (T, add, mod, lmbd1*dat)
                stat1 = oper1 (T,   T, mod, epsi1*dat)
        else
                stat1 = oper1 (F, add, epsi1*mod, dat)
                stat2 = oper2 (F,   T, lmbd1*mod, dat) 
        end if
        if(verb>7)write(0,'(A8,2L6,8E15.5)')'af--',adj,add,maxval(epsi0),minval(epsi0),maxval(lmbd0),minval(lmbd0),sum(mod),sum(dat)
        stat=stat1+stat2
  end function row_3
!#####################################
  integer function row_me_3 (oper1, oper2, oper3, adj, add, mod, dat) result(stat)
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
                function oper2 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper2
                function oper3 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper3
        end interface
        logical, intent (in)     :: adj, add
        real, dimension (: )     :: mod, dat
        integer                  :: stat1, stat2,stat3
        if (adj) then
                stat1 = oper1 (T, add, mod, epsi1*dat)
                stat2 = oper2 (T,   T, mod, epsi1*dat)
                stat3 = oper3 (T,   T, mod, epsi1*dat)
        else
                stat1 = oper1 (F, add, epsi1*mod, dat)
                stat2 = oper2 (F,   T, epsi1*mod, dat)
                stat3 = oper3 (F,   T, epsi1*mod, dat)
        end if
        stat=stat1+stat2+stat3
  end function row_me_3
!#####################################!run operator forward*back
  integer function row_fb (oper1, adj, add, mod, dat) result(stat) 
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
        end interface
        logical, intent (in)            :: adj, add
        real, dimension (: )            :: mod, dat
        real, dimension (size(dat))     ::  tmp
        integer                         ::  stat1, stat2
        tmp=0.
        if (adj) then
                stat2 = oper1 (T, add,tmp,dat)
                stat1 = oper1 (F, add,tmp,mod)
        else
                stat1 = oper1 (F, add,mod,tmp)
                stat2 = oper1 (T, add,dat,tmp)
        end if
        stat=stat1+stat2
  end function row_fb
!#####################################!run operator back*forward
  integer function row_bf (oper1, adj, add, mod, dat) result(stat) 
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
        end interface
        logical, intent (in)            :: adj, add
        real, dimension (: )            :: mod, dat
        real, dimension (size(dat))     ::  tmp
        integer                         ::  stat1, stat2
        tmp=0.
        if (adj) then
                stat2 = oper1 (F, add,dat,tmp)
                stat1 = oper1 (T, add,mod,tmp)
        else
                stat1 = oper1 (T, add,tmp,mod)
                stat2 = oper1 (F, add,tmp,dat)
        end if
        stat=stat1+stat2
  end function row_bf
!#####################################! chain two operators together    d = (AB)m and m = (AB)d
  integer function chain_2 (oper1, oper2, adj, add, mod, dat) result(stat)
        interface
                function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
                function oper2 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper2
        end interface
        logical, intent (in)                  :: adj, add
        real, dimension (:)                   :: mod, dat
        real, dimension (size(mod))           :: tmp1
        integer                               :: stat1, stat2
        tmp1=0.
        if (adj) then
                stat1 = oper1 (T,   F, tmp1, dat)
                stat2 = oper2 (T, add, mod, tmp1)
        else
                stat2 = oper2 (F,   F, mod, tmp1)
                stat1 = oper1 (F, add, tmp1, dat)
        end if
        stat=(stat1+stat2)/2
  end function chain_2
!#####################################!run operator forward + back d=(A+A)m and m=(A+A)d
  integer function row_sum_fb (oper1, adj, mod, dat) result (stat)
    interface
        function oper1 (adj, add, mod, dat) result (stat); integer :: stat; logical, intent (in) :: adj, add; real, dimension (:)  :: mod, dat; end function oper1
    end interface
    logical, intent (in)          :: adj
    real, dimension (:)           :: mod, dat
    real, dimension (size (mod))  :: temp1, temp2
    integer                       :: stat1, stat2

    if (adj) then
       stat1 = oper1 (T, F, temp1, dat)
       stat2 = oper1 (F, F, dat, temp2)
       mod = mod + temp1 + temp2
    else
       stat2 = oper1 (F, F, mod, temp2)
       stat1 = oper1 (T, F, temp1, mod)
       dat = dat + temp2 + temp1
    end if
    stat=0.
  end function row_sum_fb





end module 
