module map_mod
  use global_params_mod
  implicit none

contains

!#################################################
  subroutine k2_map(progrid)
    type(pro_grid) :: progrid   

    if(allocated(kx2)) deallocate(kx2); allocate(kx2(progrid%ax%n))
    if(allocated(ky2)) deallocate(ky2); allocate(ky2(progrid%ay%n))
    call k_map(kx2, progrid%ax%n, progrid%ax%d)
    call k_map(ky2, progrid%ay%n, progrid%ay%d)
  end subroutine
!#################################################
  subroutine k2map(progrid,kx2_in,ky2_in)
        type(pro_grid)                  :: progrid   
        real, dimension(:), allocatable :: kx2_in           !wavenumber kx*kx
        real, dimension(:), allocatable :: ky2_in           !wavenumber ky*ky

        if(allocated(kx2_in)) deallocate(kx2_in); allocate(kx2_in(progrid%ax%n))
        if(allocated(ky2_in)) deallocate(ky2_in); allocate(ky2_in(progrid%ay%n))

        call k_map(kx2_in, progrid%ax%n, progrid%ax%d)
        call k_map(ky2_in, progrid%ay%n, progrid%ay%d)
  end subroutine

!###############################################
  subroutine k_map(k,n,d)
    real,  dimension(:) :: k
    integer, intent(in) :: n
    real,    intent(in) :: d
    integer             :: i,m
    k=0.
    if(n/=1) then
       m = n/2
       do i=1,m+1; k(i)=i-  1; end do
       do i=m+2,n; k(i)=i-n-1; end do
    else
       k=0. !! make sure k=0 if n=1
    end if

    k = k / (n*d)      !! 1/length
    k = 2 * PI * k     !! radians
    k = k * k          !! k squared
  end subroutine
!###############################################
  subroutine k_map2(k,n,d)
    real,  dimension(:) :: k
    integer, intent(in) :: n
    real,    intent(in) :: d
    integer             :: i,m
    real                :: kd,ko

    k=0.
    kd = 1./(n*d)
    ko = 1./(2*d)-kd !k does not increase uniformly!

    do ik=1, n
       k(ik) = modulo(ko+(ik-1)*kd,1./d)-ko
    end do
    k = 2*pi * k
    k = k * k          !! k squared
  end subroutine
!###############################################
  subroutine data_radians_calc()
  !!.. convert frequencies to radians
  if (.not.radian) then
     dgrid%aw%d = dgrid%aw%d * 2.0 * PI
     dgrid%aw%o = dgrid%aw%o * 2.0 * PI
  end if
  end subroutine

  subroutine gre_radians_calc()
  !!.. convert frequencies to radians
  if (.not.radian) then
     ggrid%aw%d = ggrid%aw%d * 2.0 * PI
     ggrid%aw%o = ggrid%aw%o * 2.0 * PI
  end if
  end subroutine

  subroutine wav_radians_calc()
  !!.. convert frequencies to radians
  if (.not.radian) then
     wgrid%aw%d = wgrid%aw%d * 2.0 * PI
     wgrid%aw%o = wgrid%aw%o * 2.0 * PI
  end if
  end subroutine


  end module
