module nhelix_mod
! Define nonstationary filter type
	use helix_mod
implicit none
	type nfilter	! nd is data length.
		logical, dimension(:), pointer :: mis ! (nd) boundary conditions
		integer, dimension(:), pointer :: pch ! (nd) patches
		type( filter), dimension(:), pointer :: hlx ! (np) filters
	end type
!Gboyega ( 04/01/2009)
!Several nhelix with difference ranges
	type vnfilter	! nd is data length.
		type( nfilter), dimension(:), pointer :: nhlx ! (np) nfilters
		integer, dimension(:), pointer :: var ! (nv) variances
	end type
  contains
! allocate a filter
  subroutine nallocate( aa, nh, pch)
	type( nfilter) :: aa
	integer, dimension(:), intent( in) :: nh, pch
	integer :: ip, np, nd
	np = size( nh)
	allocate( aa%hlx( np))
	do ip = 1, np
		call allocatehelix( aa%hlx( ip), nh( ip))
	end do
	nd = size( pch)
	allocate( aa%pch( nd))
	aa%pch = pch
	nullify( aa%mis) ! set null pointer for mis.
  end subroutine
! destroy a filter
  subroutine ndeallocate( aa)
	type( nfilter) :: aa
	integer :: ip
	do ip = 1, size( aa%hlx)
		if(associated(aa%hlx( ip)%flt))call deallocatehelix( aa%hlx( ip))
	end do
	if(associated(aa%hlx))deallocate( aa%hlx)
	if(associated(aa%pch))deallocate( aa%pch)
	if ( associated( aa%mis)) then ! if logicals were allocated
		deallocate( aa%mis) ! free them
	end if
  end subroutine

! allocate a filter
  subroutine vnallocate( aa, nh, pch, var, nv)
	type( vnfilter) :: aa
	integer, dimension(:), intent( in) :: nh, pch, var
	integer :: ip, iv, np, nd, nv
	np = size( nh)
	nd = size( pch)
	allocate(aa%nhlx(nv))
	do iv=1,nv
		allocate(aa%nhlx(iv)%hlx(np))
		allocate(aa%nhlx(iv)%pch(nd))
		allocate(aa%var(nd))
		do ip = 1, np
			call allocatehelix(aa%nhlx(iv)%hlx(ip), nh( ip))
		end do
		nullify( aa%nhlx(iv)%mis) ! set null pointer for mis.
		aa%nhlx(iv)%pch = pch
		aa%var = var
	end do
  end subroutine
! destroy a filter
  subroutine vndeallocate( aa)
	type( vnfilter) :: aa
	integer :: ip,iv
	do iv = 1, size( aa%nhlx)
		do ip = 1, size( aa%nhlx(iv)%hlx)
			if(associated(aa%nhlx(iv)%hlx(ip)%flt))call deallocatehelix( aa%nhlx(iv)%hlx(ip))
		end do
		if(associated(aa%nhlx(iv)%hlx))deallocate( aa%nhlx(iv)%hlx)
		if(associated(aa%nhlx(iv)%pch))deallocate( aa%nhlx(iv)%pch)

		if ( associated( aa%nhlx(iv)%mis)) then ! if logicals were allocated
			deallocate( aa%nhlx(iv)%mis) ! free them
		end if
	enddo
  end subroutine
end module
