!npef  - find non stationary prediction error filter
!=head1 SYNOPSIS
!C<call find_pef(dd,aa,rr,niter,eps,nh)>
!=head1 INPUT PARAMETERS
!=item dd  -  C<real(:)>
!      Input data
!=item niter - integer
!      Number of itterations
!=item rr   - type(nfilter)
!      Filter (aa) regularizer
!=item eps  -  real
!      Epsilon to apply to filter estimation problem
!      larger mean more regularized
!=over 4
!=item  aa - type(filter)
!       output filter
! Find a non-stationary prediction-error filter (helix magic)
!L<pef>,L<nhconest>,L<solver>,L<cgstep>,L<nhelix>
!B<geef90>
module npef_mod
! Estimate non-stationary PEF
  use nhconest_mod
  use cgstep_mod1
  use npolydiv_mod 
  use solver_sep_mod,  only: solver_prec
  implicit none 
  contains
  subroutine find_pef( dd, aa, rr, niter, eps, nh)
    integer,          intent( in)    :: niter, nh       ! number of iterations
    real,             intent( in)    :: eps	        ! epsilon
    type( nfilter)                   :: aa              ! estimated PEF output.
    type( nfilter),   intent( in)    :: rr              ! roughening filter.
    real,    dimension(:), pointer   :: dd              ! input data
    integer                          :: ip, ih, np, nr  ! filter lengths
    real, dimension (:), allocatable :: flt	        ! np*na filter coefs
    np = size( aa%hlx)			                ! data   length
    nr = np*nh
    allocate( flt( nr))
    call nhconest_init( dd, aa, nh)
    call npolydiv_init( nr, rr) 
    call solver_prec(nhconest_lop, cgstep, x=flt, dat=-dd, niter=niter,prec=npolydiv_lop, nprec=nr, eps=eps)
    call   cgstep_close()
    call npolydiv_close()
    call nhconest_close()
    do ip = 1, np
      do ih = 1, size( aa%hlx(ip)%lag)
        aa%hlx( ip)%flt( ih) = flt( (ip-1)*nh + ih)
      end do
    end do 
    deallocate( flt)
  end subroutine 
end module 