!One-Way Modeling /Migration....
module oneway_mod
  use history_mod 
  use io_mod
  use array_mod
  use param_mod 
  use wavelet_mod
  use myfft_mod
  use map_mod
  use grid_mod
  use vel_mod
  use ctaper_mod
  use ssf_mod
  use ssf_mod_ffd
  implicit none

contains
!=======================================================================================
  integer function oneway_init() result(stat)
        call          mig_param()                           !initialize migration parameters 
        call      wavelet_input()                           !read wavelet history
        call          data_grid()                           !get data grid     
        call      velocity_grid()                           !get velocity grid     
        call     migration_grid()                           !get velocity grid     
        call mig_output_history()                           !output migration history
        call    mig_array_alloc()                           !allocate memory
        call       mig_velocity()                           !get velocity
        call        mig_io_init()                           !mig to zero
        if (wfld_d_out) call wfld_output_history(wfld_d_)   !output down wavefield history
        if (wfld_u_out) call wfld_output_history(wfld_u_)   !output up wavefield history
        if (wfld_d_out) call wfld_io_init(wfld_d_)          !down wavefield to zero
        if (wfld_u_out) call wfld_io_init(wfld_u_)          !up   wavefield to zero
    stat = 0
  end function
!=======================================================================================
  integer function oneway_clos() result(stat)
        call array_dealloc()
        stat = 0
  end function
!!=======================================================================================
  integer function oneway_oper() result(stat)
        stat=1
        if(loop=='s'.and.lloyd)then
                write(0,*)'Parallelize over source.....and using lloyds..fastest!!!!'
                stat=oneway_oper_slloyd()
        else
                write(0,*)'Parallelize over source.....'
                stat=oneway_oper_s()
        endif
        stat=0
  end function
!===========================================================================================================================================================================
!parallel over source
  integer function oneway_oper_s() result(stat)
        icount=0;    stackd=1 
        call omp_set_num_threads(nodes); write(*,*) 'using', nodes, 'of' , nnode
!$OMP parallel do private(iw, inode, w, index_aw, phase_shaping, data, isy, sy, index_sx, isx, sx, index_sy, pgrid, ogrid, kx2, ky2, tape,wfld_slice_d, wfld_slice_u, tmp_wfld1_d, tmp_wfld1_u, tmp_wfld2_d, tmp_wfld2_u, tmp_mig, tmp_mig2, iz, vref, vel_slice, ind_vref, iv, iy, ix, ihy, ihx, hy, hx, jhy, jhx, iyd, ixd, iyu, ixu ) schedule(dynamic,1) 
        do isx=1, dgrid%asx%n;    ! shot loop
                sx = dgrid%asx%o + (isx -1) * dgrid%asx%d; index_sx=1.5 + (sx-vgrid%ax%o)/vgrid%ax%d
                if (hole.and.isx>ih_sx1.and.isx<ih_sx2)then; write(0,*)'Hole: isx,', isx,index_sx; cycle; endif
        do isy=1, dgrid%asy%n;
                sy = dgrid%asy%o + (isy -1) * dgrid%asy%d; index_sy=1.5 + (sy-vgrid%ay%o)/vgrid%ay%d
                if (hole.and.isy>ih_sy1.and.isy<ih_sy2) cycle
                call mig_pro_grid(index_sx,index_sy, pgrid, ogrid); call k2map(pgrid,kx2,ky2);
                call mig_array_alloc2(wfld_slice_d, wfld_slice_u, tmp_wfld1_d, tmp_wfld1_u, tmp_wfld2_d, tmp_wfld2_u, tmp_mig, tmp_mig2, pgrid, ogrid,tape)
                stat = ssf_init2(vgrid%az%d)
                inode=omp_get_thread_num(); 
                write(* ,'(A19,I5,A1,I5,A6,I5,A1,I5,A9,I5)')'Shot [isx, isy]:  (', isx ,',' ,isy, ') of (', dgrid%asx%n ,',',dgrid%asy%n,') @ node ', inode
                write(10,'(A19,I5,A1,I5,A6,I5,A1,I5,A9,I5)')'Shot [isx, isy]:  (', isx  ,',',isy, ') of (', dgrid%asx%n ,',',dgrid%asy%n,') @ node ', inode 
                if (allocated(data)) deallocate(data); allocate(data(dgrid%agx%n,dgrid%agy%n,dgrid%aw%n,1)); data=0.; tmp_mig2=0.;
                if (allocated(vel_slice)) deallocate(vel_slice); allocate(vel_slice(pgrid%ax%n,pgrid%ay%n)); vel_slice=0.
                if (allocated(vref     )) deallocate(vref     ); allocate(vref    (vnum)); vref=0.
                !$OMP CRITICAL
                        call data_1w_io("reed", data, isx,isy)
                !$OMP END CRITICAL
                if(hole)then
                        if(ih_gx1<ih_gx2)then; if(ih_gx1>1.and.ih_gx2<dgrid%agx%n) data(ih_gx1:ih_gx2,:,:,1)=0.; write(0,*)isx,ih_gx1,ih_gx2,ih_sx1,ih_sx2;endif
                        if(ih_gy1<ih_gy2)then; if(ih_gy1>1.and.ih_gy2<dgrid%agy%n) data(:,ih_gy1:ih_gy2,:,1)=0.; endif
                endif
                do iw=1, dgrid%aw%n                                                                                                     ! freq. count.
                        w = dgrid%aw%o + (iw-1)*dgrid%aw%d; index_aw=1.5 + (w-dgrid%aw%o)/ggrid%aw%d
                        if(abs(csou(iw)) <= mcsou*.005 )cycle;     
                        if (igrid%ay%n== 1 .or. igrid%ax%n==1) then; phase_shaping = cmplx(cos(PI/4.), -sin(PI/4.)) / sqrt(w); else; phase_shaping = cmplx(0., - 1. * w); endif; !!..2D ==45 deg. rot. || 3D == 90 deg. rotation i^(-1)
                        wfld_slice_d=0.; wfld_slice_u=0.;tmp_mig=0
                        do iz=imin_z,pgrid%az%n;   ! depth loop
                                if (iz /=imin_z) then
                                        tmp_wfld1_d=wfld_slice_d; wfld_slice_d=0.
                                        tmp_wfld1_u=wfld_slice_u; wfld_slice_u=0.
                                        if(lloyd)then
                                                vref(:)=vel_ref(:,iz,1); vel_slice = vel(pgrid%index_xmin:pgrid%index_xmax, pgrid%index_ymin:pgrid%index_ymax,iz);
                                                do iv=nref(iz),1,-1                                                                      !vel. loop   lloyds
                                                        tmp_wfld2_d=tmp_wfld1_d; tmp_wfld2_u=tmp_wfld1_u;
                                                        call ssf_oper3(F, sign, w, vref(iv), vel_slice, tmp_wfld2_d, kx2, ky2); call ssf_oper3(F,-sign, w, vref(iv), vel_slice, tmp_wfld2_u, kx2, ky2)
                                                        where(vel_slice>=vel_ref(iv,iz,2)); wfld_slice_d = tmp_wfld2_d; wfld_slice_u = tmp_wfld2_u; endwhere
                                                enddo
                                        else
                                                call  vel_alloc(vref, vel_slice, ind_vref, pgrid     ); vel_slice = vel(pgrid%index_xmin:pgrid%index_xmax, pgrid%index_ymin:pgrid%index_ymax,iz);
                                                call   velocity(vref, vel_slice, ind_vref, nref(iz), vnum);
                                                do iv=1, nref(iz)                                                                      !vel. loop linear interp 
                                                        tmp_wfld2_d=tmp_wfld1_d; tmp_wfld2_u=tmp_wfld1_u;
                                                        call ssf_oper3(F, sign, w, vref(iv), vel_slice, tmp_wfld2_d, kx2, ky2); call ssf_oper3(F,-sign, w, vref(iv), vel_slice, tmp_wfld2_u, kx2, ky2)
                                                        where (ind_vref == iv            ); wfld_slice_d = tmp_wfld2_d*tape; wfld_slice_u = tmp_wfld2_u*tape; end where
                                                enddo
                                        endif
                                else
                                         if (spike)then; wfld_slice_d(pgrid%index_datpro_sx,pgrid%index_datpro_sy)=1.; else; wfld_slice_d(pgrid%index_datpro_sx,pgrid%index_datpro_sy)=csou(iw); endif  
                                         wfld_slice_u(pgrid%index_datpro_gxmin:pgrid%index_datpro_gxmax,pgrid%index_datpro_gymin:pgrid%index_datpro_gymax)=data(:,:,iw,1)
                                endif
                                do  iy=1,ogrid%ay%n;  
                                do  ix=1,ogrid%ax%n;  
                                        do ihy=1,ogrid%ahy%n; 
                                                hy=ogrid%ahy%o + (ihy-1)*ogrid%ahy%d; jhy = nint(hy/pgrid%ay%d) 
                                                iyd = (iy + ogrid%index_pro_ymin-1)-jhy; iyu = (iy + ogrid%index_pro_ymin-1)+jhy; if (iyd < 1 .or. iyd > pgrid%ay%n .or. iyu < 1 .or. iyu > pgrid%ay%n ) cycle 
                                        do ihx=1,ogrid%ahx%n;
                                                hx=ogrid%ahx%o + (ihx-1)*ogrid%ahx%d; jhx = nint(hx/pgrid%ax%d) 
                                                ixd = (ix + ogrid%index_pro_xmin-1)-jhx; ixu = (ix + ogrid%index_pro_xmin-1)+jhx; if (ixd < 1 .or. ixd > pgrid%ax%n .or. ixu < 1 .or. ixu > pgrid%ax%n ) cycle
                                                tmp_mig(ihx,ihy,ix,iy,iz)=tmp_mig(ihx,ihy,ix,iy,iz)+(wfld_slice_u(ixu,iyu)*conjg(wfld_slice_d(ixd,iyd))*phase_shaping);
                                        enddo
                                        enddo
                                enddo
                                enddo
!                                if(wfld_d_out .and. sx == wfld_d_sx .and. sy == wfld_d_sy ) & 
!                                        wfld_out_d(ogrid%index_img_xmin:ogrid%index_img_xmax,ogrid%index_img_ymin:ogrid%index_img_ymax,iz, iw, 1)=wfld_slice_d(ogrid%index_pro_xmin:ogrid%index_pro_xmax,ogrid%index_pro_ymin:ogrid%index_pro_ymax)
!                                if(wfld_u_out .and. sx == wfld_u_sx .and. sy == wfld_u_sy ) & 
!                                        wfld_out_u(ogrid%index_img_xmin:ogrid%index_img_xmax,ogrid%index_img_ymin:ogrid%index_img_ymax,iz, iw, 1)=wfld_slice_u(ogrid%index_pro_xmin:ogrid%index_pro_xmax,ogrid%index_pro_ymin:ogrid%index_pro_ymax)

                        enddo !end depth
                        if    (nostack_w)then;
                        else
                                if(verb>1)write(*,'(A13,I5,2x,A13,F10.4,2x,A13,F10.4)')'Adding freq:', iw, 'sum_tmpmig2:', sum(tmp_mig2), 'sum tmp_mig:',sum(tmp_mig); 
                                tmp_mig2=tmp_mig2 + tmp_mig  
                        endif
                enddo           !end freq.
                !$OMP CRITICAL
                if (nostack_s)then;
                else;
                        if(verb>0)write(*,*)'Adding shot:', isx,isy, 'sum_mig:', sum(mig), 'sum tmp_mig:',sum(tmp_mig2)
                        do  iy=1,ogrid%ay%n;  
                        do  ix=1,ogrid%ax%n;  
                               mig(:,:,ix+ogrid%index_img_xmin-1,iy+ogrid%index_img_ymin-1,:) = & 
                               mig(:,:,ix+ogrid%index_img_xmin-1,iy+ogrid%index_img_ymin-1,:) + tmp_mig2(:,:,ix,iy,:);
                        enddo
                        enddo
                        call mig_io(mig); 
                endif
                !$OMP END CRITICAL
        deallocate (tmp_mig2)
        enddo !end shot y 
        enddo !end shot x
!$OMP end parallel do
        if(wfld_d_out) call wfld_out(wfld_out_d,wfld_d_); 
        if(wfld_u_out) call wfld_out(wfld_out_u,wfld_u_);
       stat=0;
  end function
!===========================================================================================================================================================================
!parallel over source and only lloyds
  integer function oneway_oper_slloyd() result(stat)
        icount=0;    stackd=1
!$OMP parallel do private(iw, inode, w, index_aw,           &
!$OMP                           tape,                       & 
!$OMP                           phase_shaping, data,        & 
!$OMP                           wfld_slice_d, wfld_slice_u, &
!$OMP                           tmp_wfld1_d, tmp_wfld1_u,   &
!$OMP                           tmp_wfld2_d, tmp_wfld2_u,   &
!$OMP                           tmp_mig, tmp_mig2,          &
!$OMP                           vref, vel_slice,            &
!$OMP                           ind_vref, iv,               &
!$OMP                           isy, sy, index_sx,      &
!$OMP                           isx, sx, index_sy,      &
!$OMP                           pgrid, ogrid,           &
!$OMP                           kx2, ky2,               &
!$OMP                           iz, iy, ix,     &
!$OMP                           ihy, ihx, hy,   &
!$OMP                           hx, jhy, jhx,   &
!$OMP                           iyd, ixd,       &
!$OMP                           iyu, ixu ) schedule(dynamic,1) ordered
        do isx=1, dgrid%asx%n;    ! shot loop
                sx = dgrid%asx%o + (isx -1) * dgrid%asx%d; index_sx=1.5 + (sx-vgrid%ax%o)/vgrid%ax%d
                if (hole.and.isx>ih_sx1.and.isx<ih_sx2)then; write(0,*)'Hole: isx,', isx,index_sx; cycle; endif; 
        do isy=1, dgrid%asy%n;
                sy = dgrid%asy%o + (isy -1) * dgrid%asy%d; index_sy=1.5 + (sy-vgrid%ay%o)/vgrid%ay%d
                if (hole.and.isy>ih_sy1.and.isy<ih_sy2) cycle
                call mig_pro_grid(index_sx,index_sy, pgrid, ogrid); call k2map(pgrid,kx2,ky2);
                call mig_array_alloc2(wfld_slice_d, wfld_slice_u, tmp_wfld1_d, tmp_wfld1_u, tmp_wfld2_d, tmp_wfld2_u, tmp_mig, tmp_mig2, pgrid, ogrid,tape)
                call ctaper2d(tape,vgrid,nxpad,nypad,2)
                stat = ssf_init2(vgrid%az%d); if(ffd) stat  = ssf_ffd_init(sign,kx2, ky2, vgrid%az%d, vgrid%ax%d, vgrid%ay%d)
                inode=omp_get_thread_num(); 
                write(* ,'(A19,I5,A1,I5,A6,I5,A1,I5,A9,I5)')'Shot [isx, isy]:  (', isx ,',' ,isy, ') of (', dgrid%asx%n ,',',dgrid%asy%n,') @ node ', inode
                write(10,'(A19,I5,A1,I5,A6,I5,A1,I5,A9,I5)')'Shot [isx, isy]:  (', isx  ,',',isy, ') of (', dgrid%asx%n ,',',dgrid%asy%n,') @ node ', inode 
                if (allocated(data )) deallocate(data); allocate(data(dgrid%agx%n,dgrid%agy%n,dgrid%aw%n,1)); data=0.; tmp_mig2=0.;
                if (allocated(vel_slice)) deallocate(vel_slice); allocate(vel_slice(pgrid%ax%n,pgrid%ay%n)); vel_slice=0.
                if (allocated(vref     )) deallocate(vref     ); allocate(vref    (vnum)); vref=0.
                !$OMP ORDERED
                        call data_1w_io("reed", data, isx,isy)
                !$OMP END ORDERED 
                write(0,*)sum(data)
                if(cabs(sum(data))==0.) cycle
                if(hole)then
                        if(ih_gx1<ih_gx2)then; if(ih_gx1>1.and.ih_gx2<dgrid%agx%n) data(ih_gx1:ih_gx2,:,:,1)=0.; write(0,*)isx,ih_gx1,ih_gx2,ih_sx1,ih_sx2;endif
                        if(ih_gy1<ih_gy2)then; if(ih_gy1>1.and.ih_gy2<dgrid%agy%n) data(:,ih_gy1:ih_gy2,:,1)=0.; endif
                endif
                do iw=1, dgrid%aw%n                                                                                                     ! freq. count.
                        write(0,*)iw,dgrid%aw%n
                        w = dgrid%aw%o + (iw-1)*dgrid%aw%d; index_aw=1.5 + (w-dgrid%aw%o)/ggrid%aw%d
                        if(abs(csou(iw)) <= mcsou*.005 )cycle;     
                        if(phase)then;
                            if (igrid%ay%n== 1 .or. igrid%ax%n==1) then; phase_shaping = cmplx(cos(PI/4.), -sin(PI/4.)) / sqrt(w); else; phase_shaping = cmplx(0., - 1. * w); endif; !!..2D ==45 deg. rot. || 3D == 90 deg. rotation i^(-1)
                        else
                            phase_shaping=1.
                        endif
                        wfld_slice_d=0.; wfld_slice_u=0.;tmp_mig=0
                        if (spike)then; wfld_slice_d(pgrid%index_datpro_sx,pgrid%index_datpro_sy)=1.; else; wfld_slice_d(pgrid%index_datpro_sx,pgrid%index_datpro_sy)=csou(iw); endif  

                                        wfld_slice_u(pgrid%index_datpro_gxmin:pgrid%index_datpro_gxmax,pgrid%index_datpro_gymin:pgrid%index_datpro_gymax)=data(:,:,iw,1)
                        if(ffd)then                        
                            do iz=imin_z+1,pgrid%az%n;   ! depth loop
                                    tmp_wfld1_d=wfld_slice_d; wfld_slice_d=0.
                                    tmp_wfld1_u=wfld_slice_u; wfld_slice_u=0.
                                    vel_slice = vel(pgrid%index_xmin:pgrid%index_xmax, pgrid%index_ymin:pgrid%index_ymax,iz); vref(1)=minval(vel_slice)
                                    tmp_wfld2_d=tmp_wfld1_d; call ssf_oper3(F, sign, w, vref(1), vel_slice, tmp_wfld2_d, kx2, ky2); wfld_slice_d = tmp_wfld2_d*tape; 
                                    tmp_wfld2_u=tmp_wfld1_u; call ssf_oper3(F,-sign, w, vref(1), vel_slice, tmp_wfld2_u, kx2, ky2); wfld_slice_u = tmp_wfld2_u*tape; 
                                    do  iy=1,ogrid%ay%n;  
                                    do  ix=1,ogrid%ax%n;  
                                            do ihy=1,ogrid%ahy%n; 
                                                    hy=ogrid%ahy%o + (ihy-1)*ogrid%ahy%d; jhy = nint(hy/pgrid%ay%d) 
                                                    iyd = (iy + ogrid%index_pro_ymin-1)-jhy; iyu = (iy + ogrid%index_pro_ymin-1)+jhy; if (iyd < 1 .or. iyd > pgrid%ay%n .or. iyu < 1 .or. iyu > pgrid%ay%n ) cycle 
                                            do ihx=1,ogrid%ahx%n;
                                                    hx=ogrid%ahx%o + (ihx-1)*ogrid%ahx%d; jhx = nint(hx/pgrid%ax%d) 
                                                    ixd = (ix + ogrid%index_pro_xmin-1)-jhx; ixu = (ix + ogrid%index_pro_xmin-1)+jhx; if (ixd < 1 .or. ixd > pgrid%ax%n .or. ixu < 1 .or. ixu > pgrid%ax%n ) cycle
                                                    tmp_mig(ihx,ihy,ix,iy,iz)=tmp_mig(ihx,ihy,ix,iy,iz)+(wfld_slice_u(ixu,iyu)*conjg(wfld_slice_d(ixd,iyd))*phase_shaping);
                                            enddo
                                            enddo
                                    enddo
                                    enddo
                            enddo !end depth
                        else
                            do iz=imin_z+1,pgrid%az%n;   ! depth loop
                                    tmp_wfld1_d=wfld_slice_d; wfld_slice_d=0.
                                    tmp_wfld1_u=wfld_slice_u; wfld_slice_u=0.
                                    vref(:)=vel_ref(:,iz,1); vel_slice = vel(pgrid%index_xmin:pgrid%index_xmax, pgrid%index_ymin:pgrid%index_ymax,iz);
                                    do iv=nref(iz),1,-1                                                                      !vel. loop   lloyds 
                                        tmp_wfld2_d=tmp_wfld1_d; call ssf_oper3(F, sign, w, vref(iv), vel_slice, tmp_wfld2_d, kx2, ky2); 
                                        tmp_wfld2_u=tmp_wfld1_u; call ssf_oper3(F,-sign, w, vref(iv), vel_slice, tmp_wfld2_u, kx2, ky2)
                                        !stop
                                        where(vel_slice>=vel_ref(iv,iz,2)); wfld_slice_d = tmp_wfld2_d*tape; wfld_slice_u = tmp_wfld2_u*tape; endwhere
                                    enddo
                                    do  iy=1,ogrid%ay%n;  
                                    do  ix=1,ogrid%ax%n;  
                                            do ihy=1,ogrid%ahy%n; 
                                                    hy=ogrid%ahy%o + (ihy-1)*ogrid%ahy%d; jhy = nint(hy/pgrid%ay%d) 
                                                    iyd = (iy + ogrid%index_pro_ymin-1)-jhy; iyu = (iy + ogrid%index_pro_ymin-1)+jhy; if (iyd < 1 .or. iyd > pgrid%ay%n .or. iyu < 1 .or. iyu > pgrid%ay%n ) cycle 
                                            do ihx=1,ogrid%ahx%n;
                                                    hx=ogrid%ahx%o + (ihx-1)*ogrid%ahx%d; jhx = nint(hx/pgrid%ax%d) 
                                                    ixd = (ix + ogrid%index_pro_xmin-1)-jhx; ixu = (ix + ogrid%index_pro_xmin-1)+jhx; if (ixd < 1 .or. ixd > pgrid%ax%n .or. ixu < 1 .or. ixu > pgrid%ax%n ) cycle
                                                    tmp_mig(ihx,ihy,ix,iy,iz)=tmp_mig(ihx,ihy,ix,iy,iz)+(wfld_slice_u(ixu,iyu)*conjg(wfld_slice_d(ixd,iyd))*phase_shaping);
                                            enddo
                                            enddo
                                    enddo
                                    enddo
                            enddo !end depth
                        endif
                        if    (nostack_w)then;
                        else
                                if(verb>1.and.inode==0)write(*,'(A13,I5,2x,A13,F10.4,2x,A13,F10.4)')'Adding freq:', iw, 'sum_tmpmig2:', sum(tmp_mig2), 'sum tmp_mig:',sum(tmp_mig); 
                                tmp_mig2=tmp_mig2 + tmp_mig  
                        endif
                enddo           !end freq.
                !$OMP CRITICAL
                if (nostack_s)then;
                else;
                        if(verb>0)write(*,*)'Adding shot:', isx,isy, 'sum_mig:', sum(mig), 'sum tmp_mig:',sum(tmp_mig2)
                        do  iy=1,ogrid%ay%n;  
                        do  ix=1,ogrid%ax%n;  
                               mig(:,:,ix+ogrid%index_img_xmin-1,iy+ogrid%index_img_ymin-1,:) = & 
                               mig(:,:,ix+ogrid%index_img_xmin-1,iy+ogrid%index_img_ymin-1,:) + tmp_mig2(:,:,ix,iy,:);
                        enddo
                        enddo
                        call mig_io(mig); 
                endif
                !$OMP END CRITICAL
        deallocate (tmp_mig2)
        enddo !end shot y 
        enddo !end shot x
!$OMP end parallel do
        if(wfld_d_out) call wfld_out(wfld_out_d,wfld_d_); 
        if(wfld_u_out) call wfld_out(wfld_out_u,wfld_u_);
       stat=0;
  end function
!===========================================================================================================================================================================

  end module
