module rms_mod
  real, allocatable             :: tri(:,:,:),rrr(:,:,:),file1(:,:,:), file2(:,:,:)
  real   ,private               :: dx, dy, dz, ox, oy, oz
  integer,private               :: ix, iy, iz, nx, ny, nz 
  integer,private               :: iix, iiy, iiz
  integer,private               :: inode
  real   ,private               :: x, y, z,xx, yy, zz
  real                          :: rms1,rms2,rms12,nrms12
  integer                       :: nhz,nhx,nhy
  integer                       :: ntri
  real                          :: dhx,dhy,dhz
  character(len=4),private      :: mode
  real            ,private      :: eps
  logical                       :: ltri,stat1
  
  contains
      
        subroutine rms_init(mode_in,nz_in,nx_in,ny_in,nhz_in,nhx_in,nhy_in,ltri_in,inode_in)
                integer                 :: nz_in,nx_in,ny_in, nhz_in,nhx_in,nhy_in,inode_in
                character(len=4)        ::mode_in
                logical                 :: ltri_in
                mode=mode_in
                eps=eps_in
                nz=nz_in
                nx=nx_in
                ny=ny_in
                nhz=min(nz,nhz_in); if (nz==1)nhz=0.
                nhx=min(nx,nhx_in); if (nx==1)nhx=0.
                nhy=min(ny,nhy_in); if (ny==1)nhy=0.
                ltri=ltri_in
                inode=inode_in
                allocate(  rrr(nz, nx, ny)); rrr=0.
                allocate(  tri(2*nhz+1,2*nhx+1,2*nhy+1)); tri=0.; 

!prepare 3D prismatic taper
                if(ltri)then
                        dhx=1./nhx; dhz=1./nhz; dhy=1./nhy
                        do iz=1,2*nhz+1
                                if(nx>1)then
                                        do ix=1,nhx+1
                                              tri(iz,        ix,:)=(ix-1)*dhx
                                              tri(iz,2*nhx+2-ix,:)=(ix-1)*dhx
                                        enddo
                                endif
                                if(ny>1)then
                                        do iy=1,nhy+1
                                              tri(iz,:,        iy)=tri(iz,:,        iy)*(iy-1)*dhy
                                              tri(iz,:,2*nhy+2-iy)=tri(iz,:,2*nhy+2-iy)*(iy-1)*dhy
                                        enddo
                                endif
                        enddo
                        
                        if(nz>1)then
                                do iz=1,nhz+1
                                         tri(        iz,:,:)=tri(        iz,:,:)*(iz-1)*dhz
                                         tri(2*nhz+2-iz,:,:)=tri(2*nhz+2-iz,:,:)*(iz-1)*dhz
                                 enddo
                        endif
                else            
                        tri=1.
                endif        
                        tri=1.
                ntri=sum(tri)
                stop
        end subroutine

        subroutine rms_clos()
                deallocate(tri,rrr)
        end subroutine

	function rms_lop ( adj, add, xx, yy, rr) result(stat)
		integer            :: stat
		logical,intent(in) :: adj,add
		real,dimension(:)  :: xx,yy,rr 
		        stat=1
			call rms_lop2(adj,add,xx,yy); 
			rr=reshape(rrr,(/nx*ny*nz/))
			stat=0
	end function
	
	subroutine rms_lop2(adj,add,xx,yy)
        	real,dimension(nz,nx,ny)  :: xx,yy
        	logical                   :: add,adj
                integer, external         :: omp_get_thread_num
        	
call omp_set_num_threads(node)
!$OMP parallel private(ix,iy,iz,rms1,rms2,inode)
               inode=omp_get_thread_num()+1;
!$OMP do schedule (static)
                do iy=nhy+1,ny-nhy
                do ix=nhx+1,nx-nhx
                do iz=1+nhz,nz-nhz
                        rms1=sum((tri*xx(max(iz-nhz,1):min(iz+nhz,nz),ix-nhx:ix+nhx,iy-nhy:iy+nhy))**2)
                        rms2=sum((tri*yy(max(iz-nhz,1):min(iz+nhz,nz),ix-nhx:ix+nhx,iy-nhy:iy+nhy))**2)
                        if (rms1==0.and.rms2==0.)then
                                rrr(iz,ix,iy)=0.
                        else
                                rrr(iz,ix,iy)=2.*(sqrt(rms1)-sqrt(rms2))/(sqrt(rms1)+sqrt(rms2))
                        endif
                end do;
                end do;
                end do;
!$OMP end do
!$OMP end parallel

!fill-in the end-effects
                do ix=1,nhx
                        rrr(1:n1,   ix  ,1:n3)=rrr(1:n1,   nhx+1,1:n3)
                        rrr(1:n1,n2-ix+1,1:n3)=rrr(1:n1,n2-nhx  ,1:n3)
                end do
                do iz=1,nhz
                        rrr(   iz  ,1:n2,1:n3)=rrr(   nhz+1,1:n2,1:n3)
                        rrr(n1-iz+1,1:n2,1:n3)=rrr(n1-nhz  ,1:n2,1:n3)
                end do
                do iy=1,nhy
                        rrr(1:n1,1:n2,   iy  )=rrr(1:n1,1:n2,   nhy+1)
                        rrr(1:n1,1:n2,n3-iy+1)=rrr(1:n1,1:n2,n3-nhy  )
                end do
        end subroutine	

end module

