module ssf_mod
  use  myfft_mod
  use fftw_plans
  implicit none

  real, dimension(:), allocatable, private,save  :: kx2
  real, dimension(:), allocatable, private,save  :: ky2 
  real                           , private, save :: dz 
contains
!=========================================================================================
  integer function ssf_init(kx2_in, ky2_in, dz_in) result(stat)
    real, dimension(:), allocatable     :: kx2_in
    real, dimension(:), allocatable     :: ky2_in
    real                                :: dz_in
    real                                :: maxang_in
    stat = 1
    if(allocated(kx2)) deallocate(kx2); allocate(kx2(size(kx2_in)))
    if(allocated(ky2)) deallocate(ky2); allocate(ky2(size(ky2_in)))
    kx2     = kx2_in 
    ky2     = ky2_in
    dz      = dz_in
    stat = 0
  end function
!=========================================================================================
  integer function ssf_clos() result(stat)
    if(allocated(kx2)) deallocate(kx2)
    if(allocated(ky2)) deallocate(ky2)
    dz   = 0.
    stat = 0
  end function
!=========================================================================================
  integer function ssf_oper(adj, sign, w, vref, vmod_slice, wfld_slice) result(stat)
    logical                              :: adj
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:), allocatable :: vmod_slice
    complex, dimension(:,:), allocatable :: wfld_slice

    stat = 1
    if (adj) then
       call FFT2DX(wfld_slice, 1)
       call ssr_phase_shift(sign, w, vref, wfld_slice)
       call FFT2DX(wfld_slice, -1)
       call ssr_split_step( sign, w, vref, vmod_slice, wfld_slice)
    else
       call ssr_split_step( sign, w, vref, vmod_slice, wfld_slice)
       call FFT2DX(wfld_slice, 1)
       call ssr_phase_shift(sign, w, vref, wfld_slice)
       call FFT2DX(wfld_slice, -1)
    end if
    stat = 0 
  end function
!=========================================================================================
  subroutine ssr_phase_shift(sign, w, vref, wfld_slice)
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    complex, dimension(:,:), allocatable :: wfld_slice
    real                                 :: wov, wov2, kz, kz2
    integer                              :: ikx, iky

    do iky=1, size(ky2)
       do ikx=1, size(kx2)
         kz2 = w**2 / vref**2 - kx2(ikx) - ky2(iky);                     !phase shift
         if (kz2 >= 0) then                                                  !evanescence
             kz=sqrt(kz2)
             wfld_slice (ikx, iky)= wfld_slice (ikx, iky)*exp(cmplx(0,sign*dz*kz));
         else
             wfld_slice (ikx, iky)= 0.
         endif
       enddo
    enddo                                                                           !end wavenumber loop
  end subroutine
!=========================================================================================
  subroutine ssr_split_step(sign, w, vref, vmod_slice, wfld_slice)
    integer                             :: sign
    real                                :: w
    real                                :: vref
    real   , dimension(:,:), allocatable:: vmod_slice
    complex, dimension(:,:), allocatable:: wfld_slice 
    integer                             :: ix, iy
    real                                :: split

    do iy=1, size(ky2)
       do ix=1, size(kx2)
          split = sign * w * dz * (1./vmod_slice(ix,iy) - 1./vref) 
          wfld_slice(ix, iy) = wfld_slice(ix, iy) * cmplx(cos(split), sin(split))
       end do
    end do
  end subroutine
!=========================================================================================
  integer function ssf_init2(dz_in) result(stat)
    real                                :: dz_in
    stat = 1;  dz      = dz_in; stat = 0
  end function
!=========================================================================================
 integer function ssf_oper2(adj, sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2) result(stat)
    logical                              :: adj
    integer                              :: sign,ix
    real                                 :: w
    real                                 :: vref 
    real                                 :: vzref
    real   , dimension(:,:), pointer :: vmod_slice
    complex, dimension(:,:), pointer :: wfld_slice
    real, dimension(:), allocatable      :: kx_2           !wavenumber kx*kx
    real, dimension(:), allocatable      :: ky_2           !wavenumber ky*ky 

    stat = 1
    if (adj) then
       call FFT2DX(wfld_slice,  1)
       call ssr_phase_shift2(sign, w, vref, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, -1)
       call ssr_split_step2( sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
    else
       call ssr_split_step2( sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, 1)
       call ssr_phase_shift2(sign, w, vref, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, -1)
    end if
    stat = 0
  end function

!=====================================================================
  subroutine ssr_phase_shift2(sign, w, vref, wfld_slice,kx_2,ky_2)
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    complex, dimension(:,:), pointer :: wfld_slice
    real                                 :: wov, wov2, kz, kz2
    integer                              :: ikx, iky
    real, dimension(:), allocatable      :: kx_2           
    real, dimension(:), allocatable      :: ky_2           

    do iky=1, size(ky_2)
       do ikx=1, size(kx_2)
         kz2 = w**2 / vref**2 - kx_2(ikx) - ky_2(iky);                   !phase shift
         if (kz2 >= 0) then                                              !evanescence
             kz=sqrt(kz2)
             wfld_slice (ikx, iky)= wfld_slice (ikx, iky)*exp(cmplx(0,sign*dz*kz));
         else
             wfld_slice (ikx, iky)= 0.
         endif
       enddo
    enddo                                                                 !end wavenumber loop
  end subroutine

!=============================================================================
  subroutine ssr_split_step2(sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
    integer                             :: sign
    real                                :: w
    real                                :: vref
    real   , dimension(:,:), pointer:: vmod_slice
    complex, dimension(:,:), pointer:: wfld_slice 
    real, dimension(:), allocatable      :: kx_2           
    real, dimension(:), allocatable      :: ky_2           

    integer                             :: ix, iy
    real                                :: split

    do iy=1, size(ky_2)
       do ix=1, size(kx_2)
          split = sign * w * dz * (1./vmod_slice(ix,iy) - 1./vref) 
          wfld_slice(ix, iy) = wfld_slice(ix, iy) * cmplx(cos(split), sin(split))
       end do
    end do
  end subroutine
!=============================================================================
 subroutine ssf_oper3(adj, sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2) 
    logical                              :: adj
    integer                              :: sign,ix, stat
    real                                 :: w
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:)              :: vmod_slice
    complex, dimension(:,:)              :: wfld_slice
    real, dimension(:)                   :: kx_2           !wavenumber kx*kx
    real, dimension(:)                   :: ky_2           !wavenumber ky*ky

    stat = 1
    if (adj) then
       call FFT2DX(wfld_slice, 1)
       call ssr_phase_shift3(sign, w, vref, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, -1) 
       call ssr_split_step3( sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
    else
       call ssr_split_step3( sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, 1)
       call ssr_phase_shift3(sign, w, vref, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, -1)
    end if
    stat = 0
  end subroutine
!=====================================================================
!
  subroutine ssr_phase_shift3(sign, w, vref, wfld_slice,kx_2,ky_2) 
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    complex, dimension(:,:)              :: wfld_slice
    real                                 :: wov, wov2, kz, kz2
    integer                              :: ikx, iky
    real, dimension(:)                   :: kx_2
    real, dimension(:)                   :: ky_2

    do iky=1, size(ky_2)
       do ikx=1, size(kx_2)
         kz2 = w**2 / vref**2 - kx_2(ikx) - ky_2(iky);                   !phase shift
         if (kz2 >= 0) then                                              !evanescence
             kz=sqrt(kz2)
             wfld_slice (ikx, iky)= wfld_slice (ikx, iky)*exp(cmplx(0,sign*dz*kz));
         else
             wfld_slice (ikx, iky)= 0.
         endif
       enddo
    enddo                                                                 !end wavenumber loop
  end subroutine
!=====================================================================
!
  subroutine ssr_split_step3(sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
    integer                             :: sign
    real                                :: w
    real                                :: vref
    real   , dimension(:,:)             :: vmod_slice
    complex, dimension(:,:)             :: wfld_slice 
    real, dimension(:)                  :: kx_2           
    real, dimension(:)                  :: ky_2           
    integer                             :: ix, iy
    real                                :: split

    do iy=1, size(ky2)
       do ix=1, size(kx2)
          split = sign * w * dz * (1./vmod_slice(ix,iy) - 1./vref) 
          wfld_slice(ix, iy) = wfld_slice(ix, iy) * cmplx(cos(split), sin(split))
       end do
    end do
  end subroutine
!=============================================================================
 subroutine ssf_oper4(sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2) 
    logical                              :: adj
    integer                              :: sign,ix, stat
    real                                 :: w
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:)              :: vmod_slice
    complex, dimension(:,:)              :: wfld_slice
    real, dimension(:)                   :: kx_2           !wavenumber kx*kx
    real, dimension(:)                   :: ky_2           !wavenumber ky*ky

    stat = 1
       call FFT2DX(wfld_slice, 1)
       call ssr_phase_shift3(sign, w, vref, wfld_slice,kx_2,ky_2)
       call FFT2DX(wfld_slice, -1) 
       call ssr_split_step3( sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
    stat = 0
  end subroutine
!=============================================================================
 subroutine ssf_oper_fftw(sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2,plan_2dxF,plan_2dxI) 
    logical                              :: adj
    integer                              :: sign,ix, stat
    real                                 :: w
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:)              :: vmod_slice
    complex, dimension(:,:)              :: wfld_slice
    real, dimension(:)                   :: kx_2           !wavenumber kx*kx
    real, dimension(:)                   :: ky_2           !wavenumber ky*ky
    integer*8                            :: plan_2dxF,plan_2dxI

    stat = 1
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, size(kx_2), size(ky_2))
       call ssr_phase_shift3(sign, w, vref, wfld_slice,kx_2,ky_2)
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, size(kx_2), size(ky_2))
       call ssr_split_step3( sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
    stat = 0
  end subroutine
!=====================================================================
  subroutine ssr_phase_shift4(sign, w, vref, wfld_slice,kx_2,ky_2) 
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    complex, dimension(:,:)              :: wfld_slice
    real                                 :: wov, wov2, kz, kz2
    integer                              :: ikx, iky
    real, dimension(:)                   :: kx_2
    real, dimension(:)                   :: ky_2
    real                                 :: kzz2(size(kx_2),size(ky_2))

    do iky=1, size(ky_2)
       do ikx=1, size(kx_2)
         kzz2(ikx,iky) = w**2 / vref**2 - kx_2(ikx) - ky_2(iky);          !phase shift
       enddo
    enddo                                                                 !end wavenumber loop

    where (kzz2 >= 0)                                                     !evanescence
        wfld_slice= wfld_slice*exp(cmplx(0,sign*dz*sqrt(kzz2)));
    elsewhere
        wfld_slice = 0.
    end where

  end subroutine

end module














