module ssf_mod_ffd
  use  ssf_mod
  implicit none

  real, dimension(:), allocatable, private,save  :: kx2
  real, dimension(:), allocatable, private,save  :: ky2 
  real                           , private,save  :: dz,dx,dy 
  !integer                        , private,save  :: sign
  integer                        , private,save  :: nx,ny
contains
!=========================================================================================
  integer function ssf_ffd_init(sign_in,kx2_in, ky2_in, dz_in, dx_in, dy_in) result(stat)
    real, dimension(:), allocatable     :: kx2_in
    real, dimension(:), allocatable     :: ky2_in
    real                                :: dz_in, dx_in, dy_in
    real                                :: maxang_in
    integer                             :: sign_in
    stat = 1
!    if(allocated(kx2)) deallocate(kx2); allocate(kx2(size(kx2_in)))
!    if(allocated(ky2)) deallocate(ky2); allocate(ky2(size(ky2_in)))
!    kx2     = kx2_in 
!    ky2     = ky2_in
    dz      = dz_in
    dx      = dx_in
    dy      = dy_in
    nx      = size(kx2_in)
    ny      = size(ky2_in)
    !sign = sign_in

    stat = 0
  end function
!=========================================================================================
 integer function ssf_ffd_oper(adj, sign, w, vref, vmod_slice, wfld_slice, kx_2, ky_2) result(stat)
    logical                              :: adj
    integer                              :: ix,iy
    integer                              :: down
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:), pointer     :: vmod_slice
    complex, dimension(:,:), pointer     :: wfld_slice
    real, dimension(:), allocatable      :: kx_2           !wavenumber kx*kx
    real, dimension(:), allocatable      :: ky_2           !wavenumber ky*ky
    !sign=down
    vref=minval(vmod_slice)    
    stat = 1

    if (adj) then
        stat = ssf_oper2(.true., sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2)
        do iy=1,size(ky_2); call fd_forffd(wfld_slice(:,iy), vmod_slice(:,iy), vref, w,sign); enddo
        if(size(ky_2)>1)then;do ix=1,size(kx_2); call fd_forffd(wfld_slice(ix,:), vmod_slice(ix,:), vref, w,sign); enddo;endif
    else
        do iy=1,size(ky_2); call fd_forffd(wfld_slice(:,iy), vmod_slice(:,iy), vref, w,sign); enddo
        if(size(ky_2)>1)then; do ix=1,size(kx_2); call fd_forffd(wfld_slice(ix,:), vmod_slice(ix,:), vref, w,sign); enddo;endif
        stat = ssf_oper2(.false., sign, w, vref, vmod_slice, wfld_slice,kx_2,ky_2) 
    end if
    stat = 0
  end function 
!=========================================================================================
subroutine fd_forffd(u,v,v0,w,sign)
        complex         :: u(:)
        real            :: v(:)
        real            :: v0 
        integer         :: iw
        integer         :: sign
        real            :: w
        real            :: tx,a,b
        complex, pointer:: uj(:),ra(:),rb(:),rc(:)
        integer         :: ix
        complex         :: url,urr,ror,rol,dd
        real            :: vw,v0w
        complex         :: uu(size(u))
        integer         :: n

        n=size(u)
        a=0.5   ; b=0.25
        !a=0.4765; b=0.3725
        tx=0.11

        allocate(uj(n),ra(n),rb(n),rc(n))
        v0w=v0/(w)
        ra=0.;rb=0;rc=0.;
        do ix=1,n; uj(ix)=u(ix); end do
        do ix=1,n; 
                if (ix==1 ) then; url=0.0; else; url=uj(ix-1); end if
                if (ix==n) then; urr=0.0; else; urr=uj(ix+1); end if
                vw=v(ix)/(w)
                ror=cmplx(tx+b*(vw*vw+v0w*vw+v0w*v0w)/(dx*dx),sign*a*(vw-v0w)*dz/(2.0*dx*dx)) 
                rol=conjg(ror)
                u(ix)=uj(ix) +ror*(url+urr-2.0*uj(ix))
                ra(ix)=1.0-2.0*rol
                if (ix <n) rb(ix) =rol
                if (ix >1) rc(ix-1)=rol
        end do
        call thr(ra,rb,rc,u)
        deallocate(uj,ra,rb,rc)
end subroutine fd_forffd
!=========================================================================================
!tri-diagonal solver
        subroutine thr(a,b,c,u)
        complex :: a(:),b(:),c(:),u(:)
        integer :: n,ix
        b(1)=b(1)/a(1)
        u(1)=u(1)/a(1)
        n=size(u)
        do ix=2,n
                a(ix)=a(ix)-b(ix-1)*c(ix-1)
                u(ix)=(u(ix)-c(ix-1)*u(ix-1))/a(ix)
                if(ix.le.n-1) then
                        b(ix)=b(ix)/a(ix)
                endif
        enddo
        do ix=n-1,1,-1
                u(ix)=u(ix)-b(ix)*u(ix+1)
        enddo
end subroutine thr
!=========================================================================================
end module
