module ssf_zhang_mod
  use  myfft_mod
  implicit none

  real, dimension(:), allocatable, private,save  :: kx2
  real, dimension(:), allocatable, private,save  :: ky2 
  real                           , private, save :: dz 
contains
!=========================================================================================
  integer function ssf_init_zz(dz_in) result(stat)
    real                                :: dz_in
    stat = 1;  dz      = dz_in; stat = 0
  end function
!=====================================================================
 integer function ssf_oper_z1(w, vref, vmod_slice, wfld_slice, kx_2, ky_2) result(stat)
    integer                              :: sign,ix,ikx,iky
    real                                 :: w, kz2
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:), allocatable :: vmod_slice,vz,kz0
    complex, dimension(:,:), allocatable :: wfld_slice,wfld_slice0,wfld_slice1,wfld_slice2 
    real, dimension(:), allocatable      :: kx_2                    !wavenumber kx*kx
    real, dimension(:), allocatable      :: ky_2                    !wavenumber ky*ky
    allocate(wfld_slice0(size(wfld_slice,1),size(wfld_slice,2)),kz0(size(wfld_slice,1),size(wfld_slice,2)))
    stat = 1
    wfld_slice0=wfld_slice
    wfld_slice0=wfld_slice0*(1/vmod_slice-1/vref)
    call FFT2DX(wfld_slice0, 1); 
    do iky=1, size(ky_2)
        do ikx=1, size(kx_2)
            kz2 = w**2 / vref**2 - kx_2(ikx) - ky_2(iky);                   !phase shift
            if (kz2 >= 0) then                                              !evanescence
                kz0(ikx, iky)=sqrt(kz2)
                wfld_slice (ikx, iky)=-0.5*1/cmplx(0,1)*kz0(ikx, iky)*(wfld_slice(ikx, iky)-w/kz0(ikx,iky)*wfld_slice0(ikx, iky));
            else
                wfld_slice (ikx, iky)= 0.
            endif
       enddo
    enddo                                   
    call FFT2DX(wfld_slice,-1); 
    deallocate(kz0,wfld_slice0)
    stat = 0

 end function
!=====================================================================
 integer function ssf_oper_zz(adj, sign, w, vref, vz, vmod_slice, wfld_slice, kx_2, ky_2) result(stat)
    logical                              :: adj
    integer                              :: sign,ix
    real                                 :: w
    real                                 :: vref
    real                                 :: vzref
    real   , dimension(:,:), allocatable :: vmod_slice,vz
    complex, dimension(:,:), allocatable :: wfld_slice
    complex, dimension(:,:), allocatable :: wfld_slice0,wfld_slice1,wfld_slice2 
    real, dimension(:), allocatable      :: kx_2                    !wavenumber kx*kx
    real, dimension(:), allocatable      :: ky_2                    !wavenumber ky*ky
    allocate(wfld_slice0(size(wfld_slice,1),size(wfld_slice,2)),wfld_slice1(size(wfld_slice,1),size(wfld_slice,2)),wfld_slice2(size(wfld_slice,1),size(wfld_slice,2)))
    stat = 1
    if (adj) then
       call FFT2DX(wfld_slice, 1); 
       wfld_slice0(:,1)=wfld_slice(:,1);
       call ssr_phase_shift_zz(sign, w, vref, wfld_slice0,kx_2,ky_2)!0th order
       wfld_slice1(:,1)=wfld_slice0(:,1)*(kx_2/w**2);                         !1st order
       wfld_slice2(:,1)=wfld_slice1(:,1)*(kx_2/w**2);                         !2nd order
       call FFT2DX(wfld_slice0, -1); 
       call FFT2DX(wfld_slice1, -1); 
       call FFT2DX(wfld_slice2, -1); 
       call ssr_split_step_zz( sign, w, vref, vz, vmod_slice, wfld_slice,wfld_slice0,wfld_slice1,wfld_slice2,kx_2,ky_2)
    else
       call ssr_split_step_zz( sign, w, vref, vz, vmod_slice, wfld_slice,wfld_slice0,wfld_slice1,wfld_slice2,kx_2,ky_2)
       call FFT2DX(wfld_slice, 1)
       wfld_slice0(:,1)=wfld_slice(:,1);                                      !0th order
       wfld_slice1(:,1)=wfld_slice0(:,1)*(kx_2/w**2);                         !1st order
       wfld_slice2(:,1)=wfld_slice1(:,1)*(kx_2/w**2);                         !2nd order
       call ssr_phase_shift_zz(sign, w, vref, wfld_slice0,kx_2,ky_2)
       call FFT2DX(wfld_slice0, -1)
       call FFT2DX(wfld_slice1, -1)
       call FFT2DX(wfld_slice2, -1)
    end if
    deallocate(wfld_slice0,wfld_slice1,wfld_slice2 )
    stat = 0

  end function
!=====================================================================
  subroutine ssr_phase_shift_zz(sign, w, vref, wfld_slice,kx_2,ky_2)
    integer                              :: sign
    real                                 :: w
    real                                 :: vref
    complex, dimension(:,:), allocatable :: wfld_slice 
    real                                 :: wov, wov2, kz, kz2
    integer                              :: ikx, iky
    real, dimension(:), allocatable      :: kx_2           
    real, dimension(:), allocatable      :: ky_2           

    do iky=1, size(ky_2)
       do ikx=1, size(kx_2)
         kz2 = w**2 / vref**2 - kx_2(ikx) - ky_2(iky);                   !phase shift
         if (kz2 >= 0) then                                              !evanescence
             kz=sqrt(kz2)
             wfld_slice (ikx, iky)= wfld_slice (ikx, iky)*exp(cmplx(0,sign*dz*kz));
         else
             wfld_slice (ikx, iky)= 0.
         endif
       enddo
    enddo                                                                 !end wavenumber loop
  end subroutine
!=====================================================================
  subroutine ssr_split_step_zz(sign, w, vref, vz, vmod_slice, wfld_slice, wfld_slice0, wfld_slice1, wfld_slice2, kx_2,ky_2)
    integer                             :: sign
    real                                :: w
    real                                :: vref
    real   , dimension(:,:), allocatable:: vmod_slice, vz
    complex, dimension(:,:), allocatable:: wfld_slice,wfld_slice0,wfld_slice1,wfld_slice2 
    real, dimension(:), allocatable     :: kx_2           
    real, dimension(:), allocatable     :: ky_2           
    integer                             :: ix, iy
    real                                :: split

    wfld_slice0=                                               wfld_slice0 !0th order
    wfld_slice1=-1/2*vz*dz*vmod_slice                         *wfld_slice1 !1st order
    wfld_slice2=+1/2*vz*dz*vmod_slice**2*(vz*dz/4-vmod_slice)*wfld_slice2 !2nd order
    wfld_slice=(wfld_slice0+wfld_slice1+wfld_slice2)*exp(-vz*dz/(2*vref))               !sum 

    do iy=1, size(ky2)
       do ix=1, size(kx2)
          split = sign * w * dz * (1./vmod_slice(ix,iy) - 1./vref) 
          wfld_slice(ix, iy) = wfld_slice(ix, iy) * cmplx(cos(split), sin(split))
       end do
    end do
  end subroutine

end module
!=====================================================================














