module type_mod
  use sep

!--------------------------------------------------
  !!.. define type for axes
  type axis 
    integer :: n   ! total number
    real    :: d   ! spacing
    real    :: o   ! origin
    character(len=10):: l='blank'    !! label (i.e. axis name)
  end type

!---------------------------------------------------
  !!.. define propagation grid
  type pro_grid
    type(axis) :: ax, ay, az
    real       :: sx, sy, sz
    real       :: gxmin, gxmax
    real       :: gymin, gymax
    real       :: sxmin, sxmax
    real       :: symin, symax
    real       :: pxmin, pxmax
    real       :: pymin, pymax  
    real       :: jx, jy             ! sampling for the propagation grid : dpx = jx * dgx
    integer    :: index_gxmin   ! index of the first receiver gx on the velocity grid (for regular data)
    integer    :: index_gymin   ! index of the first receiver gy on the velocity grid (for regular data)
    integer    :: index_gxmax   ! index of the last receiver gx on the velocity grid (for regular data)
    integer    :: index_gymax   ! index of the last receiver gy on the velocity grid (for regular data)
    integer    :: index_datpro_sx    ! index of sx on the propagation grid
    integer    :: index_datpro_sy    ! index of sy on the propagation grid
    integer    :: index_datpro_sx_end! index of last sx in encoded shots on the propagation grid
    integer    :: index_datpro_sy_end! index of last sy in encoded shots on the propagation grid
    integer    :: index_datpro_gxmin   ! index of the first receiver gx on the propagation grid (for regular data)
    integer    :: index_datpro_gymin   ! index of the first receiver gy on the propagation grid (for regular data)
    integer    :: index_datpro_gxmax   ! index of the last receiver gx on the propagation grid (for regular data)
    integer    :: index_datpro_gymax   ! index of the last receiver gy on the propagation grid (for regular data)
    integer, allocatable, dimension(:,:) :: index_datpro_hrec  ! the index file for unsorted receiver
    integer, allocatable, dimension(:)   :: index_datpro_hsou  ! the index file for unsorted source
    integer    :: nxpad_beg, nxpad_end
    integer    :: nypad_beg, nypad_end
    integer    :: index_xmin, index_ymin 
    integer    :: index_xmax, index_ymax 
  end type

!---------------------------------------------------
  !!.. define data grid
  type dat_grid
    type(axis) :: at       !time
    type(axis) :: aw       !freq      
    type(axis) :: agx, agy, agz ! for regular data
    type(axis) :: asx, asy, asz ! for regular data
    real       :: sx, sy   ! for unsorted data
    real       :: sz, gz   ! not used yet
    integer    :: ntr      ! total traces per shot
    integer    :: nsh      ! total shot number
    logical    :: hdr_rec  ! flag indicating whether the receiver header is presented
    logical    :: hdr_sou  ! flag indicating whether the source header is presented
    real, allocatable, dimension(:,:) :: header_rec ! header file--receiver locations, header_rec(2,ntraces)
    real, allocatable, dimension(:)   :: header_sou ! header filei--source locations, header_sou(2), x, y
  end type
!---------------------------------------------------
  !!.. define wavelet grid
  type wav_grid
    type(axis) :: at       !time
    type(axis) :: aw       !freq      
    type(axis) :: agx, agy ! for regular data
    type(axis) :: asx, asy ! for regular data
  end type
!----------------------------------------------------
  type vel_grid
    type(axis) :: ax, ay, az
    integer    :: index_provel_xmin  ! first x index of the input velocity slice on the whole velocity grid
    integer    :: index_provel_ymin  ! first y index of the input velocity slice on the whole velocity grid
    integer    :: index_provel_xmax  ! last x index of the input velocity slice on the whole velocity grid
    integer    :: index_provel_ymax  ! last y index of the input velocity slice on the whole velocity grid
  end type


!----------------------------------------------------
  !!.. define imaging grid
  type img_grid
    type(axis) :: ax, ay, az
    type(axis) :: ahx, ahy
    type(axis) :: aax, aay !angle
    real       :: image_xmin
    real       :: image_ymin
    real       :: image_xmax
    real       :: image_ymax
    real       :: image_dx
    real       :: image_dy
  end type

!-----------------------------------------------------
  type out_grid
    type(axis) :: ax, ay, ahx, ahy, az
    integer    :: index_img_xmin  ! first x index of the output image on the image grid
    integer    :: index_img_ymin  ! first y index of the output image on the image grid
    integer    :: index_img_zmin  ! first z index of the output image on the image grid
    integer    :: index_img_xmax  ! last x index of the output image on the image grid
    integer    :: index_img_ymax  ! last y index of the output image on the image grid
    integer    :: index_img_zmax  ! last z index of the output image on the image grid
    integer    :: index_pro_xmin  ! first x index of the output image on the propagation grid
    integer    :: index_pro_ymin  ! first y index of the output image on the propagation grid
    integer    :: index_pro_zmin  ! first z index of the output image on the propagation grid
    integer    :: index_pro_xmax  ! last x index of the output image on the propagation grid
    integer    :: index_pro_ymax  ! last y index of the output image on the propagation grid
    integer    :: index_pro_zmax  ! last z index of the output image on the propagation grid
  end type

!--------------------------------------------------------
  type gre_grid
    type(axis) :: ax_modl, ay_modl, az_modl
    type(axis) :: ax_surf, ay_surf, aw
    real       :: surface_xmin       ! xmin for the surface location
    real       :: surface_xmax       ! xmax for the surface location
    real       :: surface_ymin       ! ymin for the surface location
    real       :: surface_ymax       ! ymax for the surface location
    integer    :: index_surpro_xmin  ! first x index of the surface grid on the propagation grid
    integer    :: index_surpro_xmax  ! last x index of the surface grid on the propagation grid
    integer    :: index_surpro_ymin
    integer    :: index_surpro_ymax
    integer    :: index_modpro_modx  ! x index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_mody  ! y index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_modz  ! z index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_minx  ! x index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_miny  ! y index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_minz  ! z index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_maxx  ! x index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_maxy  ! y index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_maxz  ! z index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_mod_minx, index_mod_miny, index_mod_minz  
    integer    :: index_mod_maxx, index_mod_maxy, index_mod_maxz 
    integer    :: index_pro_minx, index_pro_miny, index_pro_minz  
    integer    :: index_pro_maxx, index_pro_maxy, index_pro_maxz  
    real       :: xmin_modl, xmax_modl
    real       :: ymin_modl, ymax_modl
    real       :: zmin_modl, zmax_modl
  end type 
!---------------------------------------------------------------
  type hes_grid
    type(axis)                       :: ax_modl, ay_modl, az_modl ! defines the main diagonal of Hessian, or the points in the model space
    type(axis)                       :: ax_offd, ay_offd, az_offd ! defines the off diagonal of Hessian
    type(axis)                       :: ahx_modl, ahy_modl        ! defines the subsurface offset for the main diagonal of Hessian
    type(axis)                       :: ahx_offd, ahy_offd        ! defines the subsurface offset for the off diagonal of Hessian
    type(axis)                       :: aax_modl, aax_offd
    real                             :: sx, sy                    ! source location for current shot
    real                             :: xmin_modl, xmax_modl
    real                             :: ymin_modl, ymax_modl
    real                             :: zmin_modl, zmax_modl
    integer                          :: index_datgre_sx
    integer                          :: index_datgre_sy
    integer                          :: index_datgre_gxmin
    integer                          :: index_datgre_gymin
    integer, dimension(:,:), pointer :: index_datgre_rec     ! index of receiver locations on the Green's surface grid for each shot
    integer, dimension(:,:), pointer :: mask_rec_loc         ! receiver location mask, 1 for where receivers are located, 0 for not.
  end type
!----------------------------------------------------------------
  type block
    integer                            :: ns
    integer                            :: nb
    integer                            :: nb_res
    real                               :: ds
    real   , allocatable, dimension(:) :: os
    integer, allocatable, dimension(:) :: blen
    integer, allocatable, dimension(:) :: indx
    logical                            :: bres
  end type

!----------------------------------------------------
  !!.. define reflectivity grid
  type ref_grid
    type(axis) :: ax, ay, az
    type(axis) :: ahx, ahy
    real       :: ref_xmin
    real       :: ref_ymin
    real       :: ref_xmax
    real       :: ref_ymax
    real       :: ref_dx
    real       :: ref_dy
  end type

end module


