module util_
  use sep
  use type_mod
  implicit none

  type, public :: myaxis
     integer   :: n
     real      :: o,d
     character(len=128) :: l
  end type myaxis

  interface grab_param
     module procedure grab_param_real
     module procedure grab_param_logical
     module procedure grab_param_integer
  end interface

contains
  !----------------------------------------------------------------
  integer function sig(m) result(s)
    real            :: m
    if (m>0.) then
       s=+1
    else if (m .eq. 0.) then
       s=0
    else 
       s=-1
    end if
  end function sig
  !----------------------------------------------------------------
  subroutine houaxis(file,a,i)
    character(len=*), intent(in)  :: file
    integer,          intent(in)  :: i
    type(axis),       intent(in)  :: a
    character(len=128)            :: buf

    write(buf, "(a,i1)") 'n'    ,i; call to_history(trim(buf),     a%n, trim(file))
    write(buf, "(a,i1)") 'o'    ,i; call to_history(trim(buf),     a%o, trim(file))
    write(buf, "(a,i1)") 'd'    ,i; call to_history(trim(buf),     a%d, trim(file))
    write(buf, "(a,i1)") 'label',i; call to_history(trim(buf),trim(a%l),trim(file))
  end subroutine houaxis
!----------------------------------------------------------------
  subroutine hinaxis2(filetag, ax, ndim)
    character(len=*)           :: filetag
    type(axis)                 :: ax
    integer                    :: ndim
    character(len=128)         :: buffer

    write(buffer, "(a,i1,i1)") "n", ndim;    call from_aux(filetag, buffer, ax%n)
    write(buffer, "(a,i1,i1)") "d", ndim;    call from_aux(filetag, buffer, ax%d)
    write(buffer, "(a,i1,i1)") "o", ndim;    call from_aux(filetag, buffer, ax%o)

    return
  end subroutine
!----------------------------------------------------------------
  !! load axis info from a header file
  subroutine hinaxis(file,a,i)
    character(len=*), intent(in)  :: file
    integer,          intent(in)  :: i
    type(axis),       intent(out) :: a
    character(len=128)            :: buf

    write(buf,"(a,i1)") 'n',i; call from_aux(trim(file),trim(buf),a%n,1 )
    write(buf,"(a,i1)") 'o',i; call from_aux(trim(file),trim(buf),a%o,0.)
    write(buf,"(a,i1)") 'd',i; call from_aux(trim(file),trim(buf),a%d,1.)
  end subroutine hinaxis
!-----------------------------------------------------------------
  subroutine houaxis2(filetag, ax, ndim)
    character(len=*)   :: filetag
    type(axis)         :: ax
    integer            :: ndim
    character(len=128) :: buffer
    write(buffer, "(a,i1)") "n", ndim
    call auxputch(buffer, "i", ax%n, filetag)
    write(buffer, "(a,i1)") "d", ndim
    call auxputch(buffer, "r", ax%d, filetag)
    write(buffer, "(a,i1)") "o", ndim
    call auxputch(buffer, "r", ax%o, filetag)
    return
  end subroutine
  !----------------------------------------------------------------
  subroutine pinaxis(axis)
    type(myaxis), intent(out) :: axis
    character (len=128)      :: buffer

    write (buffer, "(a,a)" ) trim(axis%l),'n'
    call from_param(trim(buffer),axis%n,1)

    write (buffer, "(a,a)" ) trim(axis%l),'o'
    call from_param(buffer,axis%o,0.)

    write (buffer, "(a,a)" ) trim(axis%l),'d'
    call from_param(buffer,axis%d,1.)
  end subroutine  pinaxis
  !----------------------------------------------------------------
  subroutine ainaxis(file,ax,i)
    character(len=128)       :: file
    integer    , intent(in)  :: i
    type(axis)         :: ax
    character (len=128)      :: buffer
    write (buffer, "(a,i1)" ) 'n',i; call from_aux(file,buffer,ax%n)
    write (buffer, "(a,i1)" ) 'o',i; call from_aux(file,buffer,ax%o)
    write (buffer, "(a,i1)" ) 'd',i; call from_aux(file,buffer,ax%d)
 !   write (buffer, "(a,i1)" ) 'l',i; call from_aux(file,buffer,ax%l)
 !    write(0,*)file
    !  write(0,*)buffer
  end subroutine ainaxis
  !----------------------------------------------------------------

  subroutine printaxis(axis)
    type(myaxis), intent(out) :: axis
    write(0,*) axis%n,axis%o,axis%d,trim(axis%l)
    axis=axis
  end subroutine printaxis

  !----------------------------------------------------------------

  subroutine grab_param_real(pname,pval,pdefault)
    real,               intent(inout) :: pval
    character(len=128), intent(in)    :: pname
    real,               intent(in)    :: pdefault

    call from_param(pname,pval,pdefault)
    write(0,*) trim(pname),pval
  end subroutine grab_param_real


  subroutine grab_param_logical(pname,pval,pdefault)
    logical,            intent(inout) :: pval
    character(len=128), intent(in)    :: pname
    logical,            intent(in)    :: pdefault

    call from_param(pname,pval,pdefault)
    write(0,*) trim(pname),pval
  end subroutine grab_param_logical


  subroutine grab_param_integer(pname,pval,pdefault)
    integer,            intent(inout) :: pval
    character(len=128), intent(in)    :: pname
    integer,            intent(in)    :: pdefault

    call from_param(pname,pval,pdefault)
    write(0,*) trim(pname),pval
  end subroutine grab_param_integer

end module util_



