module wavelet_mod
  use global_params_mod
  use ricker_mod
  use io_mod
  use map_mod
  implicit none

contains

!========================================================================================== 
 subroutine wavelet_input()
    integer             :: iw, niw
    real                :: w 
    integer             :: num_match
    character(len=1024) :: string

    wavelet_="wavelet"; twavelet_="twavelet"

    call from_param("nt",dgrid%at%n,1001);call from_param("dt",dgrid%at%d,0.004); call from_param("ot",dgrid%at%o,0.); call from_param("fund",fund,20.); call from_param("flip",lflip,.true.) 
    call from_param("tmax", tmax, dgrid%at%o + dgrid%at%d*(dgrid%at%n-1))
    call from_param("jw", jw , 1);   

!read the source wavelet from file    
    if (.not. exist_file(wavelet_)) then
        call ricker(wavelet_, dgrid%at, fund, floor(dgrid%at%n*.1), lflip);call hinaxis(wavelet_,dgrid%aw,3); 
    else
        call auxin(wavelet_)                                              ;call hinaxis(wavelet_,dgrid%aw,3); 

    endif
    if (.not. radian) call data_radians_calc()

    allocate(csou(dgrid%aw%n),wavelet(1))
    do iw=1,dgrid%aw%n; call data_io_wavelet(wavelet, iw); csou(iw)=wavelet(1); enddo
    mcsou=maxval(abs(csou))
!=compute the useful freq.!!!!!!!!!!!!!!!!!!!!!!!! !only take values >=0.5% of max value in wavelet
    iw=0.;
    do niw=1,dgrid%aw%n-1
           if (abs(csou(niw))<=mcsou*.005 .and. niw <= 0.05*dgrid%aw%n) iw=iw+1;
           if (abs(csou(niw))<=mcsou*.005 .and. niw >= 0.15*dgrid%aw%n) exit;
    enddo
    
    unw=niw-iw; ggrid%aw%o=dgrid%aw%o+(iw-1)*dgrid%aw%d; ggrid%aw%d=dgrid%aw%d*jw; ggrid%aw%n=(niw-iw)/jw; 
    if(ldata.or.allfreq)ggrid%aw=dgrid%aw

    write(0,'(A65)')'========================Summary wavelet================================='
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','Data  Freq. ax','|',dgrid%aw, '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','Green Freq. ax','|',ggrid%aw, '|'
    write(0,'(A65)')'------------------------------------------------------------------'
    write(*,'(A2,A15,A2,F10.3,A8)')'|','Min. Freq.','|',mcsou,'radians'
    write(*,'(A2,A15,A2,F10.3,A8)')'|','Max. Freq.','|',mcsou+(ggrid%aw%n-1)*ggrid%aw%d,'radians'
    write(0,'(A65)')'End wavelet============================================================='
    deallocate(wavelet)
  end subroutine
!###############################################
 subroutine data_io_wavelet(wavelet, index_ws)
    complex, dimension(:), allocatable  :: wavelet
    integer                             :: index_ws
    integer, dimension(1)               :: n_cube, j_wind, n_wind, f_wind
    integer                             :: n_dim_cube, err

    n_dim_cube = 1
    n_cube = dgrid%aw%n
    j_wind = 1 
    n_wind = 1
    f_wind = index_ws-1

    wavelet = 0.
    err = sreed_window(wavelet_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,wavelet)
    if (err .lt. 0) call seperr("io_wavelet: trouble reading source wavelet")
  end subroutine

  !==========================================#
 subroutine green_wavelet()
    integer             :: iw, niw
    real                :: w 
    real                ::max_t
    character(len=1024) :: string

    wavelet_="wavelet"; twavelet_="twavelet"

    call from_param("nt", dgrid%at%n, 1001 ); call from_param("dt", dgrid%at%d, 0.004); call from_param("ot", dgrid%at%o, 0.   ); call from_param("fund",fund   , 20.    )
    call from_param("flip", lflip , .true. ); call from_param("tmax", tmax, dgrid%at%o + dgrid%at%d*(dgrid%at%n-1));              call from_param("jw", jw , 1);   

!read the source wavelet from file    
    if (.not. exist_file(wavelet_)) then
        call ricker(wavelet_, dgrid%at, fund, floor(dgrid%at%n*.1), lflip);call hinaxis(wavelet_,dgrid%aw,3); 
    else
        call auxin(wavelet_)                                              ;call hinaxis(wavelet_,dgrid%aw,3); 
    endif

    allocate(csou(dgrid%aw%n),wavelet(1))
    do iw=1,dgrid%aw%n
            call data_io_wavelet(wavelet, iw)
            csou(iw)=wavelet(1)
    end do
    mcsou=maxval(abs(csou))
!=compute the useful freq.!!!!!!!!!!!!!!!!!!!!!!!!
!only take values >=0.5% of max value in wavelet
    iw=0.;
    !do niw=floor(dgrid%aw%n*0.1),dgrid%aw%n-1
    do niw=1,dgrid%aw%n-1
           !if (abs(csou(niw))<=mcsou*.001) then; iw=iw+1; if (iw>=floor(dgrid%aw%n*0.02)) exit; endif
           if (abs(csou(niw))<=mcsou*.005 .and. niw <= 0.05*dgrid%aw%n) iw=iw+1;
           if (abs(csou(niw))<=mcsou*.005 .and. niw >= 0.15*dgrid%aw%n) exit;
    enddo
    
    unw=niw-iw
    ggrid%aw%o=dgrid%aw%o+(iw)*dgrid%aw%d; ggrid%aw%d=dgrid%aw%d*jw; ggrid%aw%n=(niw-iw)/jw; 
    deallocate(wavelet)
!initialize the data time axis.
        max_t=1/dgrid%aw%d
        dgrid%at%n=dgrid%aw%n
        dgrid%at%d=(max_t-dgrid%at%o)/(dgrid%at%n-1)
        if(ldata.or.allfreq)ggrid%aw=dgrid%aw

    write(0,'(A65)')'========================Summary wavelet================================='
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','Data  Freq. ax','|',dgrid%aw, '|'
    write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','Green Freq. ax','|',ggrid%aw, '|'
    write(0,'(A65)')'------------------------------------------------------------------'
    write(*,'(A2,A15,A2,F10.3,A8)')'|','Min. Freq.','|',mcsou,'radians'
    write(*,'(A2,A15,A2,F10.3,A8)')'|','Max. Freq.','|',mcsou+(ggrid%aw%n-1)*ggrid%aw%d,'radians'
    write(0,'(A65)')'End wavelet============================================================='
  end subroutine

  !==========================================#
 subroutine wavelet_hdr()
     complex, dimension(:), allocatable  :: wavelet
    integer                             :: index_ws
    integer, dimension(1)               :: n_cube, j_wind, n_wind, f_wind
    integer                             :: n_dim_cube, err

    n_dim_cube = 1
    n_cube = dgrid%aw%n
    j_wind = 1 
    n_wind = 1
    f_wind = index_ws-1

    wavelet = 0.
    err = sreed_window(wavelet_,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,wavelet)
    if (err .lt. 0) call seperr("io_wavelet: trouble reading source wavelet")
  end subroutine

  !==========================================#
 subroutine wavelet_s()
  use init_mod

    integer             :: iw, niw
    real                :: w 
    real                ::max_t,ocsou
    character(len=1024) :: string

        wavelet_="wavelet"; twavelet_="twavelet"; 

        call from_param("nt", dgrid%at%n, 1001 ); call from_param("dt", dgrid%at%d, 0.004); call from_param("ot", dgrid%at%o, 0.   ); call from_param("fund",fund   , 20.    )
        call from_param("flip", lflip , .true. ); call from_param("tmax", tmax, dgrid%at%o + dgrid%at%d*(dgrid%at%n-1));              call from_param("jw", jw , 1);   

!read the source wavelet from file    
        if (.not. exist_file(wavelet_)) then
                call ricker(wavelet_, dgrid%at, fund, floor(dgrid%at%n*.1), lflip);call hinaxis(wavelet_,dgrid%aw,3); 
        else
                call auxin(wavelet_);
                call hinaxis(wavelet_,wgrid%asx,1); call hinaxis(wavelet_,wgrid%asy,2); call hinaxis(wavelet_,wgrid%aw ,3); 
                dgrid%aw=wgrid%aw
                dgrid%asx=wgrid%asx
                dgrid%asy=wgrid%asy
        endif

        allocate(csou_full(wgrid%asx%n,wgrid%asy%n,wgrid%aw%n)); call data_io_csou(csou_full); 
        mcsou=maxval(abs(csou_full))
         ocsou=minval(abs(csou_full))
        allocate(csou(wgrid%aw%n)); csou(:)=csou_full(1,1,:);

!=compute the useful freq.!!!!!!!!!!!!!!!!!!!!!!!!!only take values >=0.5% of max value in wavelet
        iw=0.;
        do niw=1,dgrid%aw%n-1
           if (abs(csou(niw))<=mcsou*.005 .and. niw <= 0.05*dgrid%aw%n) iw=iw+1;
           if (abs(csou(niw))<=mcsou*.005 .and. niw >= 0.15*dgrid%aw%n) exit;
        enddo

        unw=niw-iw; ggrid%aw%o=dgrid%aw%o+(iw)*dgrid%aw%d; ggrid%aw%d=dgrid%aw%d*jw; ggrid%aw%n=(niw-iw)/jw; 
!initialize the data time axis.
        max_t=1/dgrid%aw%d
        dgrid%at%n=dgrid%aw%n
        dgrid%at%d=(max_t-dgrid%at%o)/(dgrid%at%n-1)
        if(ldata.or.allfreq)ggrid%aw=dgrid%aw

! apply phase shifts
        if (.not. radian) call wav_radians_calc()
        call wfld_z1_encode(csou_full)

        write(0,'(A65)')'========================Summary wavelet================================='
        write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','Data  Freq. ax','|',dgrid%aw, '|'
        write(0,'(A2,A15,A2,I10,2(F10.3,2X),A10,A2)') '|','Green Freq. ax','|',ggrid%aw, '|'
        write(0,'(A65)')'------------------------------------------------------------------'
        write(*,'(A2,A15,A2,F10.3,A8)')'|','Min. Freq.','|',ocsou,'radians'
        write(*,'(A2,A15,A2,F10.3,A8)')'|','Max. Freq.','|',ocsou+(ggrid%aw%n-1)*ggrid%aw%d,'radians'
        write(0,'(A65)')'End wavelet============================================================='
  end subroutine
  end module
